/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.algorithms;

import java.util.ArrayList;
import java.util.HashMap;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.dragom.io.LanguagePrinter;
import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.Label;
import de.uni_due.inf.ti.graph.Node;

/**
 * Algorithm to compute and print a type graph language.
 * 
 * @author Dennis Nolte
 */
public class LanguageAlgorithm {

		private class BitEdge {
			int src;
			int tar;
			Label label;
			
			public BitEdge(int src, int tar,Label label) {
				super();
				this.src = src;
				this.tar = tar;
				this.label = label;
			}
			public int getSrc() {
				return src;
			}
			public int getTar() {
				return tar;
			}
			public Label getLabel() {
				return label;
			}
			
			public String printEdgeBit() {
				return "(" + src + "," + tar + ")"; 
			}
			public boolean equals(BitEdge edge) {
				return ((this.src == edge.getSrc()) && (this.tar == edge.getTar()));
			}
		}
		
		
		HashMap<Node,ArrayList<ArrayList<Integer>>> bitNodesCombinations;
		ArrayList<ArrayList<ArrayList<Integer>>> bitNodeSelections;
		
		HashMap<Edge,ArrayList<ArrayList<BitEdge>>> bitEdgesCombinations;
		ArrayList<ArrayList<ArrayList<BitEdge>>> bitEdgeSelections;
		
		HashMap<Integer,ArrayList<AnnotatedTypeGraph>> isoGraphs;
		
		AnnotatedTypeGraph typeGraph;
		
		public LanguageAlgorithm()
		{
			this(new AnnotatedTypeGraph());
		}
		
		public LanguageAlgorithm(AnnotatedTypeGraph typeGraph)
		{
			this.typeGraph = typeGraph;
			resetAllLists();
		}

		public AnnotatedTypeGraph getTypeGraph() {
			return typeGraph;
		}

		public void setTypeGraph(AnnotatedTypeGraph typeGraph) {
			this.typeGraph = typeGraph;
		}
		
		public void computeLanguage() {
			resetAllLists();
        	LanguagePrinter printer = new LanguagePrinter();
        	printer.createFolderTree(typeGraph.getName());
        	printer.createTMPFolder();
        	printer.printGraph(typeGraph, typeGraph.getName(), typeGraph.getName());
			
        	printer.setPrintAnnotations(false);
        	//MAIN NODE CALCULATION
        	initializeBitNodeSelections();

			//MAIN EDGE CALCULATION
			//Build the graphs (and print them)
			int graphNumber = 0;
			for(ArrayList<ArrayList<Integer>> selection : bitNodeSelections){
				
				initializeBitEdgeSelections(selection);
				
				for(ArrayList<ArrayList<BitEdge>> edgeSelection : bitEdgeSelections) {
					
					AnnotatedTypeGraph graph = new AnnotatedTypeGraph();				
					HashMap<Integer,Node> nodeMap = new HashMap<Integer,Node>();
					for(ArrayList<Integer> nodeSet : selection){
						for(int j : nodeSet){
							nodeMap.put(j, graph.addStandardAnnotatedNode());
						}
					}
					
					for(ArrayList<BitEdge> edgeSet : edgeSelection){
						for(BitEdge k: edgeSet){
							graph.addStandardAnnotatedEdge(k.getLabel(), nodeMap.get(k.getSrc()), nodeMap.get(k.getTar()));
						}
					}
					
					if(foundNewGraph(graph)){
						printer.printGraph(graph,typeGraph.getName(),graphNumber + "");
						graphNumber++;
					}
				}
					
				//Delete the gathered edge-knowledge
				this.bitEdgesCombinations = new HashMap<Edge,ArrayList<ArrayList<BitEdge>>>();
				this.bitEdgeSelections = new ArrayList<ArrayList<ArrayList<BitEdge>>>();
				
			} //Next node choice
			//END CALCULATION
			printer.generateLanguageHTMLDoc(typeGraph.getName(), graphNumber-1);
			System.out.println("Finished Print");
		}

		private void resetAllLists() {
			this.bitNodesCombinations = new HashMap<Node,ArrayList<ArrayList<Integer>>>();
			this.bitNodeSelections = new ArrayList<ArrayList<ArrayList<Integer>>>();
			this.bitEdgesCombinations = new HashMap<Edge,ArrayList<ArrayList<BitEdge>>>();
			this.bitEdgeSelections = new ArrayList<ArrayList<ArrayList<BitEdge>>>();
			this.isoGraphs = new HashMap<Integer,ArrayList<AnnotatedTypeGraph>>();
		}

		private boolean foundNewGraph(AnnotatedTypeGraph graph) {
			int isoKey = graph.getIsoHash();
			if(isoGraphs.containsKey(isoKey)){
				ArrayList<AnnotatedTypeGraph> isoCandidates = isoGraphs.get(isoKey);
				for(AnnotatedTypeGraph candidate : isoCandidates){
					if(graph.isIsomorphic(candidate)){
						return false;
					}
				}
				isoCandidates.add(graph);
				isoGraphs.put(isoKey, isoCandidates);
			}
			else{ 
				ArrayList<AnnotatedTypeGraph> newList = new ArrayList<AnnotatedTypeGraph>();
				newList.add(graph);
				isoGraphs.put(isoKey, newList);
			}
			return true;
		}

		private void initializeBitEdgeSelections(ArrayList<ArrayList<Integer>> selection) {
			ArrayList<BitEdge> selectableEdges = new ArrayList<BitEdge>();
			for(Edge e: typeGraph.getEdges()){
				for(int src : selection.get(e.getSource().getId()))
				{
					for(int tar : selection.get(e.getTarget().getId()))
					{
						for(int i=0; i<typeGraph.getMaxBound(e);i++){
							selectableEdges.add(new BitEdge(src,tar,e.getLabel()));
						}
					}
				}
				bitEdgesCombinations.put(e, transformIntoEdgePowerset(selectableEdges));
				selectableEdges = new ArrayList<BitEdge>();
			}
			filterEdgePowerSetElements();
			if(isCurrentChoiceValid()) {
				//printCurrentEdgePowersets(selection);
				
				ArrayList<ArrayList<ArrayList<BitEdge>>> trimmedBitEdgesCombinations = new ArrayList<ArrayList<ArrayList<BitEdge>>>();
				for(Edge e: typeGraph.getEdges()){
					trimmedBitEdgesCombinations.add(bitEdgesCombinations.get(e));
				}
				generateEdgeChoices(trimmedBitEdgesCombinations, new ArrayList<ArrayList<BitEdge>>());
			
				//printFinalEdgeChoices();	
			}
		}

		/**
		 * Initializes bitNodeSelections as the filtered powerset combinations of all possible bit node clusters
		 */
		private void initializeBitNodeSelections() {
			int id = 0;
			ArrayList<Integer> selectableNodes = new ArrayList<Integer>();
			for(Node n: typeGraph.getNodes()){
				for(int i=0; i<typeGraph.getMaxBound(n);i++){
					selectableNodes.add(id);
					id++;
				}
				bitNodesCombinations.put(n, transformIntoNodePowerset(selectableNodes));
				selectableNodes = new ArrayList<Integer>();
			}
			filterNodePowerSetElements();
			
			ArrayList<ArrayList<ArrayList<Integer>>> trimmedBitNodesCombinations = new ArrayList<ArrayList<ArrayList<Integer>>>();
			for(Node n: typeGraph.getNodes()){
				trimmedBitNodesCombinations.add(bitNodesCombinations.get(n));
			}
			generateNodeChoices(trimmedBitNodesCombinations, new ArrayList<ArrayList<Integer>>());
		}
	
		//-----------------------------------------------------------------------------------------------------------
		//--------------------------------------Node Transformation Methods------------------------------------------
		//-----------------------------------------------------------------------------------------------------------
		
		public ArrayList<ArrayList<Integer>> transformIntoNodePowerset(ArrayList<Integer> elements) {
			ArrayList<ArrayList<Integer>> powerSet = new ArrayList<ArrayList<Integer>>();
			ArrayList<Integer> powerSetElement = new ArrayList<Integer>();
			ArrayList<Integer> newElement = new ArrayList<Integer>();
			powerSet.add(powerSetElement); //adding the empty Element
			
			for(int i : elements){
				for(ArrayList<Integer> currentElement: new ArrayList<ArrayList<Integer>>(powerSet)){
					newElement = new ArrayList<Integer>(currentElement);
					newElement.add(i);
					powerSet.add(newElement);
				}
			}
			return powerSet;	
		}
		
		public void filterNodePowerSetElements() {
			for(Node n: typeGraph.getNodes()) {
				for(ArrayList<Integer> element : new ArrayList<ArrayList<Integer>>(bitNodesCombinations.get(n))) {
					int minBound = typeGraph.getMinBound(n);
					if(element.size() < minBound){
						bitNodesCombinations.get(n).remove(element);
					}
				}
			}
		}
		
		public void generateNodeChoices(ArrayList<ArrayList<ArrayList<Integer>>> listOfPowersets, ArrayList<ArrayList<Integer>> chosenSets) {
			if(!listOfPowersets.isEmpty()) {
				ArrayList<ArrayList<Integer>> nextPowerset = listOfPowersets.remove(0);
				for(ArrayList<Integer> set : nextPowerset){
					ArrayList<ArrayList<Integer>> resultSet = new ArrayList<ArrayList<Integer>>(chosenSets);
					resultSet.add(set);
					generateNodeChoices(new ArrayList<ArrayList<ArrayList<Integer>>>(listOfPowersets), resultSet);
				}
			}
			else { bitNodeSelections.add(chosenSets); }	
		}

		//-----------------------------------------------------------------------------------------------------------
		//--------------------------------------Edge Transformation Methods------------------------------------------
		//-----------------------------------------------------------------------------------------------------------
		
		//Diese Methodik ist absolut nicht performant da die powerset konstruktion ab 24 Elementen in die Knie gezwungen wird
		//Alternative Graph aufzahlung vll mittels SMT Solver bentigt.
		public ArrayList<ArrayList<BitEdge>> transformIntoEdgePowerset(ArrayList<BitEdge> elements) {
			ArrayList<ArrayList<BitEdge>> powerSet = new ArrayList<ArrayList<BitEdge>>();
			ArrayList<BitEdge> powerSetElement = new ArrayList<BitEdge>();
			ArrayList<BitEdge> newElement = new ArrayList<BitEdge>();
			powerSet.add(powerSetElement); //adding the empty Element
			
			for(BitEdge e : elements){
				ArrayList<ArrayList<BitEdge>> nextIteration = new ArrayList<ArrayList<BitEdge>>(powerSet);

				for(ArrayList<BitEdge> currentElement: nextIteration){
					newElement = new ArrayList<BitEdge>(currentElement);
					newElement.add(e);
					powerSet.add(newElement);
				}
			}
			return powerSet;	
		}
		
		public void filterEdgePowerSetElements() {
			for(Edge e: typeGraph.getEdges()) {
				for(ArrayList<BitEdge> element : new ArrayList<ArrayList<BitEdge>>(bitEdgesCombinations.get(e))) {
					if(element.size() < typeGraph.getMinBound(e) || element.size() > typeGraph.getMaxBound(e)){
						bitEdgesCombinations.get(e).remove(element);
					}
				}
			}
			boolean foundNewHighlander = true;
			ArrayList<ArrayList<BitEdge>> highlanderList = new ArrayList<ArrayList<BitEdge>>();
			for(Edge e: typeGraph.getEdges()) {
				highlanderList = new ArrayList<ArrayList<BitEdge>>();
				for(ArrayList<BitEdge> highlanderCandidate : bitEdgesCombinations.get(e)) {
					if(!highlanderList.isEmpty()){
						foundNewHighlander = true;
						for(ArrayList<BitEdge> highlander : new ArrayList<ArrayList<BitEdge>>(highlanderList)) {
							if(highlander.size() == highlanderCandidate.size() && foundNewHighlander) {
								foundNewHighlander = false;
								for(int k=0; k< highlander.size();k++){
									if(!highlander.get(k).equals(highlanderCandidate.get(k))) {
										foundNewHighlander = true;
									}
								}
							}
						}
						
						if(foundNewHighlander) {
							highlanderList.add(highlanderCandidate);
						}
					}
					else{ highlanderList.add(highlanderCandidate);	}				
				}				
				bitEdgesCombinations.put(e, highlanderList);
			}
		}
		
		public void generateEdgeChoices(ArrayList<ArrayList<ArrayList<BitEdge>>> listOfPowersets, ArrayList<ArrayList<BitEdge>> chosenSets) {
			if(!listOfPowersets.isEmpty()) {
				ArrayList<ArrayList<BitEdge>> nextPowerset = listOfPowersets.remove(0);
				for(ArrayList<BitEdge> set : nextPowerset){
					ArrayList<ArrayList<BitEdge>> resultSet = new ArrayList<ArrayList<BitEdge>>(chosenSets);
					resultSet.add(set);
					generateEdgeChoices(new ArrayList<ArrayList<ArrayList<BitEdge>>>(listOfPowersets), resultSet);
				}
			}
			else { bitEdgeSelections.add(chosenSets); }	
		}
		
		private boolean isCurrentChoiceValid(){

			for(Edge e: typeGraph.getEdges())
			{
				bitEdgesCombinations.get(e);
				if(bitEdgesCombinations.get(e).isEmpty()) {
					return false;
				}
			}
			return true;
		}
		//-----------------------------------------------------------------------------------------------------------
		//--------------------------------------Test Print Methods---------------------------------------------------
		//-----------------------------------------------------------------------------------------------------------
		
		public void printCurrentPowersets(){
			for(Node n: typeGraph.getNodes()) {
				System.out.print("Node with id: " + n.getId() + " with minBound: " + typeGraph.getMinBound(n) + " and maxBound: " + typeGraph.getMaxBound(n) + " has PowersetNodes: [ ");
				ArrayList<ArrayList<Integer>> selectableSets = bitNodesCombinations.get(n);
				for(ArrayList<Integer> set: selectableSets){
					System.out.print("[ ");
					for(int i: set){
						System.out.print(i + " ");
					}
					System.out.print("] ");
				}
				System.out.println("]");
			}
			System.out.println("---------------------------");
		}
		
		/**
		 * Prints the bitNodeSelections List (which contains all possible node selection possibilities)
		 */
		public void printFinalNodeChoices(){
			int i = 0;
			for(ArrayList<ArrayList<Integer>> choiceSet : bitNodeSelections){ 
				i= 0;
				System.out.print("[ ");
					for(ArrayList<Integer> choice : choiceSet){
						System.out.print("Node" + i + ": [ ");
							for(int id: choice){
								System.out.print(id + " ");
							}
						System.out.print("] ");
						i++;
					}
				System.out.println("]");
			}
			System.out.println("---------------------------");
		}
		
		public void printFinalEdgeChoices(){
			int i = 0;
			for(ArrayList<ArrayList<BitEdge>> choiceSet : bitEdgeSelections){ 
				i= 0;
				System.out.print("[ ");
					for(ArrayList<BitEdge> choice : choiceSet){
						System.out.print("Edge" + i + ": [ ");
							for(BitEdge e: choice){
								System.out.print(e.printEdgeBit() + " ");
							}
						System.out.print("] ");
						i++;
					}
				System.out.println("]");
			}
			System.out.println("<<---------------------------<<");
		}
		
		public void printCurrentEdgePowersets(ArrayList<ArrayList<Integer>> selection){
			int currid = 0;
			System.out.println("The current selection of nodes is:");
			for(ArrayList<Integer> nodeSet : selection) {
					System.out.print("Node" + currid + " [ ");
					for(int i: nodeSet){
						System.out.print(i + " ");
					}
					System.out.print("] ");	
					currid++;
			}
			System.out.println("\n\n there are the following edge powersets");
			for(Edge e: typeGraph.getEdges())
			{
				ArrayList<ArrayList<BitEdge>> selectableSets = bitEdgesCombinations.get(e);
				System.out.print("[ ");
				for(ArrayList<BitEdge> set: selectableSets){
					System.out.print("[ ");
					for(BitEdge i: set){
						System.out.print(i.printEdgeBit() + " ");
					}
					System.out.print("] ");
				}
				System.out.println("]");
			}
			
			System.out.println(">>--------------------------->>");
		}
		
		
		
}
