/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.algorithms;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import com.microsoft.z3.BoolExpr;
import com.microsoft.z3.Context;
import com.microsoft.z3.IntExpr;
import com.microsoft.z3.Model;
import com.microsoft.z3.Solver;
import com.microsoft.z3.Status;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.dragom.data.AnnotationBounds;
import de.uni_due.inf.ti.dragom.data.Materialization;
import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.GraphElement;
import de.uni_due.inf.ti.graph.Morphism;
import de.uni_due.inf.ti.graph.Node;

/**
 * Algorithm to compute the rewritable materialization.
 * 
 * @author Dennis Nolte
 */
public class MaterializationConstructionAlgorithm {

		private Morphism baseMorphism;
		private List<Integer> validMultiplicityIndexes;
		private List<Node> deadNodes;
		
		private AnnotatedTypeGraph leftHandSideGraph;
		private AnnotatedTypeGraph abstractGraph;
		
		/**
		 * Constructor for the Materialization algorithm
		 * @param baseMorphism The match of the left-hand side into the host-graph
		 * @param validMultiplicityIndexes List of integers with valid multiplicity indexes for which the base morphism is semi legal
		 * @param deadNodes List of nodes which are obsolete wrt the rule
		 */
		public MaterializationConstructionAlgorithm(Morphism baseMorphism, List<Integer> validMultiplicityIndexes, List<Node> deadNodes) {
			
			this.baseMorphism = baseMorphism;
			this.leftHandSideGraph = (AnnotatedTypeGraph) baseMorphism.getDomain();
			this.abstractGraph = (AnnotatedTypeGraph) baseMorphism.getCodomain();
			this.validMultiplicityIndexes = validMultiplicityIndexes;
			this.deadNodes = deadNodes;
		}
		
		/**
		 * Materializazion algorithm main function
		 * @return Returns the Materialization instance of this algorithm
		 */
		public Materialization computeMaterialization(){
						
			AnnotatedTypeGraph materialization = new AnnotatedTypeGraph();
			Morphism leftEmbedding = new Morphism(leftHandSideGraph,materialization);		
			Morphism rightEmbedding = new Morphism(materialization,abstractGraph);
			
			//Construct the plain graph structure
			List<Node> leftNodes = leftHandSideGraph.getNodes();
			List<Edge> leftEdges = leftHandSideGraph.getEdgeList();
			
			List<Node> abstractNodes = abstractGraph.getNodes();
			List<Edge> abstractEdges = abstractGraph.getEdgeList();
			
			Node newMatNode;
			Edge newMatEdge;
			List<Node> sourceImage;
			List<Node> targetImage;
			
			List<GraphElement> formerLeft = new LinkedList<GraphElement>();
			List<Node> dyingNodes = new LinkedList<Node>();
			
			//Adding the left-hand side nodes and updating the left and right embedding
			for(Node lNode: leftNodes) {
				newMatNode = materialization.addNode();
				formerLeft.add(newMatNode);
				leftEmbedding.checkAndPut(lNode, newMatNode);
				rightEmbedding.checkAndPut(newMatNode, baseMorphism.get(lNode));
				if(deadNodes.contains(lNode)) dyingNodes.add(newMatNode);
			}
			//Adding the abstract graph nodes and updating the right embedding
			for(Node aNode: abstractNodes) {
				newMatNode = materialization.addNode();
				rightEmbedding.checkAndPut(newMatNode, aNode);
			}
			//Adding the left-hand side edge and updating both embeddings
			for(Edge lEdge: leftEdges) {
				newMatEdge = materialization.addEdge(lEdge.getLabel(), leftEmbedding.get(lEdge.getSource()), leftEmbedding.get(lEdge.getTarget()));
				formerLeft.add(newMatEdge);
				leftEmbedding.checkAndPut(lEdge, newMatEdge);
				rightEmbedding.checkAndPut(newMatEdge, baseMorphism.get(lEdge));
			}
			//Adding the abstract edges alongside their materializations and updating the right embedding
			for(Edge aEdge: abstractEdges) {

				sourceImage = new LinkedList<Node>();
				sourceImage.add(aEdge.getSource());
				
				targetImage = new LinkedList<Node>();
				targetImage.add(aEdge.getTarget());

				for(Node matSrcNode: rightEmbedding.getPreimageOfNodes(sourceImage)) {
					
					for(Node matTarNode: rightEmbedding.getPreimageOfNodes(targetImage)) {
						
						//Don't add materialized edges that would unsatisfy the dangling edge condition
						if(!(dyingNodes.contains(matSrcNode) || dyingNodes.contains(matTarNode))) {
							newMatEdge = materialization.addEdge(aEdge.getLabel(), matSrcNode, matTarNode);
							rightEmbedding.checkAndPut(newMatEdge, aEdge);
						}
					}
				}
				
			}
			
			//Add the multiplicities
			HashMap<GraphElement,IntExpr> lowerBound = new HashMap<GraphElement,IntExpr>();
			HashMap<GraphElement,IntExpr> upperBound = new HashMap<GraphElement,IntExpr>();
			List<GraphElement> matAllElements = new LinkedList<GraphElement>();
			List<Node> imageNode;
			List<Edge> imageEdge;
			
			Context ctx = new Context();
			Solver s;
			int limit = abstractGraph.getLimit();
			
			// Prepare SMT variables
			for(Node mNode: materialization.getNodes()) {
				lowerBound.put(mNode, ctx.mkIntConst("NL"+mNode.getId()));
				upperBound.put(mNode, ctx.mkIntConst("NU"+mNode.getId()));
				matAllElements.add(mNode);
			}
			int edgeId = 0;
			for(Edge mEdge: materialization.getEdgeList()) {
				lowerBound.put(mEdge, ctx.mkIntConst("EL"+edgeId));
				upperBound.put(mEdge, ctx.mkIntConst("EU"+edgeId));
				edgeId++;
				matAllElements.add(mEdge);
			}
			
			// Prepare constrain that there are no negative values and bounds are respected
			BoolExpr[] cfAllGreaterEqualZeroAndValidBound = new BoolExpr[matAllElements.size()*3];
			int item = 0;
			for(GraphElement g: matAllElements) {
				cfAllGreaterEqualZeroAndValidBound[item] = ctx.mkGe(lowerBound.get(g), ctx.mkInt(0));
				cfAllGreaterEqualZeroAndValidBound[item+1] = ctx.mkGe(upperBound.get(g), ctx.mkInt(0));
				cfAllGreaterEqualZeroAndValidBound[item+2] = ctx.mkLe(lowerBound.get(g), upperBound.get(g));
				item = item+3;
			}
			
			// Preserve left-hand side standard annotations
			BoolExpr[] cfFormerLeftEqualsStandard = new BoolExpr[formerLeft.size()*2];
			item = 0;
			for(GraphElement g: matAllElements) {
				if(formerLeft.contains(g)) {
					cfFormerLeftEqualsStandard[item] = ctx.mkEq(lowerBound.get(g), ctx.mkInt(1));
					cfFormerLeftEqualsStandard[item+1] = ctx.mkEq(upperBound.get(g), ctx.mkInt(1));
					item = item+2;
				}
			}
			
			List<BoolExpr> formerSolutions = new LinkedList<BoolExpr>();
			List<Integer> prunedValidMultiplicityIndexes = pruneValidMultiplicityIndexes();
			
			// Check for all valid multiplicities...
			int formerSelectedMultiplicityIndex = abstractGraph.getActiveMultiplicityIndex();
			for(int index: prunedValidMultiplicityIndexes) {
				abstractGraph.selectMultiplicity(index);
			
				s = null;
				s = ctx.mkSolver();
				s.add(ctx.mkAnd(cfAllGreaterEqualZeroAndValidBound));
				s.add(ctx.mkAnd(cfFormerLeftEqualsStandard));
				for(BoolExpr sol: formerSolutions) {
					s.add(sol);
				}
	
				//...iterating over abstract node preimages... 
				for(Node aNode: abstractNodes) {
					imageNode = new LinkedList<Node>();
					imageNode.add(aNode);
					int min = abstractGraph.getAnnotationBounds(aNode).getMin();
					int max = abstractGraph.getAnnotationBounds(aNode).getMax();
					List<Node> leftPreImageNode = new LinkedList<Node>();
					
					//... setting the upper bound constraint...
					if(max == limit) {
						for(Node preImageN: rightEmbedding.getPreimageOfNodes(imageNode)) {
							if(!formerLeft.contains(preImageN)) {
								s.add(ctx.mkEq(upperBound.get(preImageN), ctx.mkInt(limit)));
							}
							else {
								leftPreImageNode.add(preImageN);
							}
						}
					}
					else {
						IntExpr[] summands = new IntExpr[rightEmbedding.getPreimageOfNodes(imageNode).size()];
						item = 0;
						for(Node preImageN: rightEmbedding.getPreimageOfNodes(imageNode)) {
							summands[item] = upperBound.get(preImageN);
							item++;
							if(formerLeft.contains(preImageN)) leftPreImageNode.add(preImageN);
						}
						s.add(ctx.mkEq(ctx.mkAdd(summands), ctx.mkInt(max)));
					}
					
					//... setting the lower Bound constraint...
					if(leftPreImageNode.size() >= min) {
						for(Node preImageN: rightEmbedding.getPreimageOfNodes(imageNode)) {
							if(!formerLeft.contains(preImageN)) s.add(ctx.mkEq(lowerBound.get(preImageN), ctx.mkInt(0)));
						}
					}
					else {
						IntExpr[] summands = new IntExpr[rightEmbedding.getPreimageOfNodes(imageNode).size()];
						item = 0;
						for(Node preImageN: rightEmbedding.getPreimageOfNodes(imageNode)) {
							summands[item] = lowerBound.get(preImageN);
							item++;
						}
						s.add(ctx.mkEq(ctx.mkAdd(summands), ctx.mkInt(min)));
					}	
				}
				
				//...iterating over abstract edge preimages... 
				for(Edge aEdge: abstractEdges) {
					imageEdge = new LinkedList<Edge>();
					imageEdge.add(aEdge);
					int min = abstractGraph.getAnnotationBounds(aEdge).getMin();
					int max = abstractGraph.getAnnotationBounds(aEdge).getMax();
					List<Edge> leftPreImageEdge = new LinkedList<Edge>();
					
					//... setting the upper bound constraint...
					if(max == limit) {
						for(Edge preImageE: rightEmbedding.getPreimageOfEdges(imageEdge)) {
							if(!formerLeft.contains(preImageE)) {
								s.add(ctx.mkEq(upperBound.get(preImageE), ctx.mkInt(limit)));
							}
							else {
								leftPreImageEdge.add(preImageE);
							}
						}
					}
					else {
						IntExpr[] summands = new IntExpr[rightEmbedding.getPreimageOfEdges(imageEdge).size()];
						item = 0;
						for(Edge preImageE: rightEmbedding.getPreimageOfEdges(imageEdge)) {
							summands[item] = upperBound.get(preImageE);
							item++;
							if(formerLeft.contains(preImageE)) leftPreImageEdge.add(preImageE);
						}
						s.add(ctx.mkEq(ctx.mkAdd(summands), ctx.mkInt(max)));
					}
					
					//... setting the lower Bound constraint...
					if(leftPreImageEdge.size() >= min) {
						for(Edge preImageE: rightEmbedding.getPreimageOfEdges(imageEdge)) {
							if(!formerLeft.contains(preImageE)) s.add(ctx.mkEq(lowerBound.get(preImageE), ctx.mkInt(0)));
						}
					}
					else {
						IntExpr[] summands = new IntExpr[rightEmbedding.getPreimageOfEdges(imageEdge).size()];
						item = 0;
						for(Edge preImageE: rightEmbedding.getPreimageOfEdges(imageEdge)) {
							summands[item] = lowerBound.get(preImageE);
							item++;
						}
						s.add(ctx.mkEq(ctx.mkAdd(summands), ctx.mkInt(min)));
					}	
				}
							
				//...finally generate the models
				while(s.check() == Status.SATISFIABLE) {
					Model m = s.getModel();
					HashMap<GraphElement, AnnotationBounds> multiplicityResult = new HashMap<GraphElement,AnnotationBounds>();
					BoolExpr[] newSolution = new BoolExpr[matAllElements.size()*2];
					item = 0;
					for(GraphElement modelElement: matAllElements) { 
						multiplicityResult.put(modelElement, new AnnotationBounds(Integer.parseInt(m.getConstInterp(lowerBound.get(modelElement)).toString()),Integer.parseInt(m.getConstInterp(upperBound.get(modelElement)).toString())));
						newSolution[item] = ctx.mkEq(lowerBound.get(modelElement), m.getConstInterp(lowerBound.get(modelElement)));
						newSolution[item+1] = ctx.mkEq(upperBound.get(modelElement), m.getConstInterp(upperBound.get(modelElement)));
						item = item+2;
					}
					materialization.addMultiplicity(multiplicityResult);
					s.add(ctx.mkNot(ctx.mkAnd(newSolution)));
					formerSolutions.add(ctx.mkNot(ctx.mkAnd(newSolution)));
				}
			}
			materialization.selectMultiplicity(0);
			materialization.setLimit(limit);
			materialization.setName(abstractGraph.getName());
			abstractGraph.selectMultiplicity(formerSelectedMultiplicityIndex);
//			IntExpr x = ctx.mkIntConst("x");
//			IntExpr y = ctx.mkIntConst("y");
//			s.add(ctx.mkGe(x, ctx.mkInt(0)));
//			s.add(ctx.mkGe(y, ctx.mkInt(0)));
//			s.add(ctx.mkLe(ctx.mkAdd(x,y), ctx.mkInt(5)));
//			
//			while(s.check() == Status.SATISFIABLE) {
//				Model m = s.getModel();
//				System.out.println(m.getConstInterp(x) + " and " + m.getConstInterp(y));
//				s.add(ctx.mkNot(ctx.mkAnd(ctx.mkEq(x, m.getConstInterp(x)),ctx.mkEq(y, m.getConstInterp(y)))));
//			}
//			System.out.println(s.toString());
			
			ctx.close();
			return new Materialization(leftEmbedding, rightEmbedding);
		}

		
		private List<Integer> pruneValidMultiplicityIndexes() {
			List<Integer> result = new LinkedList<Integer>();
			boolean newAnnot = false;
			
			for(int firstIndex=0;firstIndex < validMultiplicityIndexes.size();firstIndex++) {
				
				newAnnot = (firstIndex==validMultiplicityIndexes.size()-1);
				for(int secondIndex=firstIndex+1;secondIndex < validMultiplicityIndexes.size();secondIndex++) {
					HashMap<GraphElement,AnnotationBounds> firstAnnotation = abstractGraph.getMultiplicities().get(validMultiplicityIndexes.get(firstIndex));
					HashMap<GraphElement,AnnotationBounds> secondAnnotation = abstractGraph.getMultiplicities().get(validMultiplicityIndexes.get(secondIndex));
					newAnnot = false;
					
					for(GraphElement element: firstAnnotation.keySet()) {
						if((firstAnnotation.get(element).getMin() < secondAnnotation.get(element).getMin()) || (secondAnnotation.get(element).getMax() < firstAnnotation.get(element).getMax())) {
							newAnnot = true;
							break;
						}
					}	
					if(!newAnnot) break;
				}
				
				if(newAnnot) result.add(validMultiplicityIndexes.get(firstIndex));
			}		
			return result;
		}		
		
}
