/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.algorithms;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.dragom.data.AnnotationBounds;
import de.uni_due.inf.ti.dragom.data.Postcondition;
import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.Graph;
import de.uni_due.inf.ti.graph.GraphElement;
import de.uni_due.inf.ti.graph.Morphism;
import de.uni_due.inf.ti.graph.Node;
import de.uni_due.inf.ti.graph.Rule;
import de.uni_due.inf.ti.graph.Transition;

/**
 * Algorithm to compute the postcondition of a graph language wrt. a GTS.
 * 
 * @author Dennis Nolte
 */
public class PostconditionConstructionAlgorithm {

		private Rule rule;
		private Morphism match;
		
		/**
		 * Constructor for the Postcondition algorithm
		 * @param rule The rule to be applied
		 * @param match The match from the left-hand side into the rewritable materialization
		 */
		public PostconditionConstructionAlgorithm(Rule rule, Morphism match) {			
			this.rule = rule;
			this.match = match;
		}
		
		/**
		 * Postcondition algorithm main function
		 * @return Returns the Postcondition instance of this algorithm
		 */
		public Postcondition computePostcondition(){
			
			Transition spoStep = rule.applyToMatch(match);
			AnnotatedTypeGraph result = new AnnotatedTypeGraph();
			AnnotatedTypeGraph materialization = (AnnotatedTypeGraph) spoStep.getSource();
			Graph purePost = spoStep.getTarget();
			Set<Node> rhsNodes = new HashSet<Node>();
			Set<Edge> rhsEdges = new HashSet<Edge>();
			HashMap<GraphElement,GraphElement> pureToResult = new HashMap<GraphElement,GraphElement>(); 
			HashMap<GraphElement,GraphElement> resultToMat = new HashMap<GraphElement,GraphElement>();
			
			Map<Node, Node> matToPureNodes = spoStep.getMorphism().getNodeMap();
			Map<Edge, Edge> matToPureEdges = spoStep.getMorphism().getEdgeMap();
			
			Set<GraphElement> interfaceElement = new HashSet<GraphElement>();

			//Extract interface elements to identify right-hand-side elements
			Morphism corr = rule.getCorrespondence();

			for(Node in: corr.getNodeMap().keySet()) {
				interfaceElement.add(match.get(in));
			}
			for(Edge en: corr.getEdgeMap().keySet()) {
				interfaceElement.add(match.get(en));
			}
			
			
			//Build the result graph without any annotations and build the mapping from pure to result
			for(Node np: purePost.getNodes()) {
				pureToResult.put(np, result.addNode());
			}
			for(Edge ep: purePost.getEdges()) {
				pureToResult.put(ep, result.addEdge(ep.getLabel(),(Node) pureToResult.get(ep.getSource()),(Node) pureToResult.get(ep.getTarget())));
			}
			
			//Build the mapping from the result graph to the materialization to inherit the annotations and filter non-Right elements
			Set<GraphElement> nonRight = new HashSet<GraphElement>();
			for(Node nm: matToPureNodes.keySet()) {
				GraphElement target = pureToResult.get(matToPureNodes.get(nm));
				resultToMat.put(target, nm);
				if(!interfaceElement.contains(nm)) nonRight.add(target);
			}
			for(Edge em: matToPureEdges.keySet()) {
				GraphElement target = pureToResult.get(matToPureEdges.get(em));
				resultToMat.put(target, em);
				if(!interfaceElement.contains(em)) nonRight.add(target);
			}
			
			//Fill the sets of right Elements
			for(Node nr: result.getNodes()) {
				if(!nonRight.contains(nr)) rhsNodes.add(nr);
			}
			for(Edge er: result.getEdges()) {
				if(!nonRight.contains(er)) rhsEdges.add(er);
			}
			
			//Inherit annotations
			HashMap<GraphElement,AnnotationBounds> newAnnot;
			for(HashMap<GraphElement,AnnotationBounds> matAnnot: materialization.getMultiplicities()) {
				newAnnot = new HashMap<GraphElement,AnnotationBounds>();
				
				for(Node n: result.getNodes()) {
					if(nonRight.contains(n)) {
						newAnnot.put(n, matAnnot.get((resultToMat.get(n))));
					}
					else {
						newAnnot.put(n, new AnnotationBounds(1, 1));
					}
				}
				for(Edge e: result.getEdges()) {
					if(nonRight.contains(e)) {
						newAnnot.put(e, matAnnot.get((resultToMat.get(e))));
					}
					else {
						newAnnot.put(e, new AnnotationBounds(1, 1));
					}
				}
				result.addMultiplicity(newAnnot);
			}
			result.selectMultiplicity(0);
			result.setLimit(materialization.getLimit());
			result.setName("Post"+materialization.getName());
			
			return new Postcondition(result, rhsNodes, rhsEdges);
		}
		
}
