/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.data;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;

import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.Graph;
import de.uni_due.inf.ti.graph.GraphElement;
import de.uni_due.inf.ti.graph.Label;
import de.uni_due.inf.ti.graph.Node;

/**
 * Class of multiply annotated type graphs.
 * 
 * @author Dennis Nolte
 */
public class AnnotatedTypeGraph extends Graph{
	
	private int limit;
	private HashMap<GraphElement,AnnotationBounds> multiplicity;
	private List<HashMap<GraphElement,AnnotationBounds>> multiplicities;
	private int activeMultiplicityIndex;

    /**
     * Creates a new multiply annotated type graph.
     */
	public AnnotatedTypeGraph(){
		super();
		multiplicity = new HashMap<GraphElement,AnnotationBounds>();
		multiplicities = new LinkedList<HashMap<GraphElement,AnnotationBounds>>();
		limit = 2;
		activeMultiplicityIndex = 0;
	}

	/**
	 * Creates a new multiply annotated type graph with a specified name.
	 * @param name 
	 * 		The name of the annotated type graph
	 */
	public AnnotatedTypeGraph(String name) {
		super(name);
		multiplicity = new HashMap<GraphElement,AnnotationBounds>();
		multiplicities = new LinkedList<HashMap<GraphElement,AnnotationBounds>>();
		limit = 2;
		activeMultiplicityIndex = 0;
	}
	
    /**
     * Creates a new multiply annotated type graph which is a copy of another graph.
     * 
     * @param orig          
     *      the graph of which the new graph must be a copy
     * @throws NullPointerException
     *      if {@code orig} is {@code null}
     */
    public AnnotatedTypeGraph(AnnotatedTypeGraph orig)
    {  
    	super();
        limit = orig.getLimit();
        multiplicity = new HashMap<GraphElement,AnnotationBounds>();
        multiplicities = new LinkedList<HashMap<GraphElement,AnnotationBounds>>();
        nodes = new LinkedList<>();
        edges = new HashSet<>();

        nodeMap = new HashMap<>();
        HashMap<Node,Node> nodeBackRef = new HashMap<>();
        HashMap<Edge,Edge> edgeBackRef = new HashMap<>();
        
        for (Node node : orig.getNodes()) {
            Node newNode = new Node(this, node.getId());
            nodes.add(newNode);
            multiplicity.put(newNode, orig.getAnnotationBounds(node));
            nodeMap.put(node, newNode);
            nodeBackRef.put(newNode, node);
        }
        for (Edge edge : orig.getEdges()) {
            List<Node> ans = new ArrayList<>();
            for (Node n : edge.getNodes()) {
                assert (n.getGraph() == orig);
                Node an = nodeMap.get(n);
                assert an != null : "@AssumeAssertion(nullness)"; // by construction
                ans.add(an);
            }
            Edge newEdge = new Edge(this, edge.getLabel(), ans);
            edges.add(newEdge);
            multiplicity.put(newEdge, orig.getAnnotationBounds(edge));
            edgeBackRef.put(newEdge, edge);
        }   
        
        nextNodeId = orig.nextNodeId;
        isoHash = orig.isoHash;
        setName(orig.getName());
        setAttribute(orig.getAttribute());
        activeMultiplicityIndex = orig.getActiveMultiplicityIndex();
        
        List<HashMap<GraphElement,AnnotationBounds>> origMultiplicities = orig.getMultiplicities();
        for(int i=0; i<origMultiplicities.size();i++){
        	if(i != activeMultiplicityIndex) {
        		HashMap<GraphElement,AnnotationBounds> multiplicityCopy = new HashMap<>();
        		for(Node x: nodes){
        			multiplicityCopy.put(x, origMultiplicities.get(i).get(nodeBackRef.get(x)));
        		}
        		for(Edge y: edges){
        			multiplicityCopy.put(y, origMultiplicities.get(i).get(edgeBackRef.get(y)));
        		}
        		multiplicities.add(multiplicityCopy);
        	}
        	else{
        		multiplicities.add(getCurrentMultiplicity());        		
        	}
        }
    }

    /**
     * Creates a new annotated type graph which is a standard annotated copy of a graph.
     * 
     * @param orig          
     *      the graph of which the new graph must be a copy
     * @throws NullPointerException
     *      if {@code orig} is {@code null}
     */
    public AnnotatedTypeGraph(Graph orig)
    {  
    	super();
        limit = 2;
        multiplicity = new HashMap<GraphElement,AnnotationBounds>();
        multiplicities = new LinkedList<HashMap<GraphElement,AnnotationBounds>>();
        nodes = new LinkedList<>();
        edges = new HashSet<>();

        nodeMap = new HashMap<>();
        
        for (Node node : orig.getNodes()) {
            Node newNode = new Node(this, node.getId());
            nodes.add(newNode);
            multiplicity.put(newNode, new AnnotationBounds());
            nodeMap.put(node, newNode);
        }
        for (Edge edge : orig.getEdges()) {
            List<Node> ans = new ArrayList<>();
            for (Node n : edge.getNodes()) {
                assert (n.getGraph() == orig);
                Node an = nodeMap.get(n);
                assert an != null : "@AssumeAssertion(nullness)"; // by construction
                ans.add(an);
            }
            Edge newEdge = new Edge(this, edge.getLabel(), ans);
            edges.add(newEdge);
            multiplicity.put(newEdge, new AnnotationBounds());
        }   
        
        nextNodeId = orig.getNextNodeId();
        isoHash = orig.getIsoHash();
        setName(orig.getName());
        setAttribute(orig.getAttribute());
        activeMultiplicityIndex = 0;
        multiplicities.add(getCurrentMultiplicity());
    }
    
    /**
     * Returns a copy of this graph
     */
    public AnnotatedTypeGraph clone()
    {
        return new AnnotatedTypeGraph(this);
    }

	//----------------------------------------------------------------------------------------
	//Add node section
	//----------------------------------------------------------------------------------------
	
    /**
     * Adds a fresh standard annotated node
     * @return The node that has been added
     */
	public Node addStandardAnnotatedNode(){
		
		Node newNode = addNode();
		multiplicity.put(newNode, new AnnotationBounds());
		return newNode;
	}
	
	/**
	 * Adds a fresh node with annotation bounds
	 * @param bounds The bounds to be set for the new node
	 * @return The node that has been added
	 */
	public Node addAnnotatedNode(AnnotationBounds bounds){
		
		Node newNode = addNode();
		multiplicity.put(newNode, bounds);
		return newNode;
	}

	/**
	 * Adds a fresh node with annotation bounds
	 * @param min The minimal bound to be set for the new node
	 * @param max The maximal bound to be set for the new node
	 * @return The node that has been added
	 */
	public Node addAnnotatedNode(int min, int max){
		
		Node newNode = addNode();
		multiplicity.put(newNode, new AnnotationBounds(min,max));
		return newNode;		
	}
	
	//----------------------------------------------------------------------------------------
	//Add edge section
	//----------------------------------------------------------------------------------------
	
	/**
	 * Adds a standard annotated edge to the graph
	 * @param label The edge label to be used
	 * @param nodes The nodes incident to this edge
	 * @return The added standard annotated edge
	 */
	public Edge addStandardAnnotatedEdge(Label label, Node... nodes){
		
		Edge newEdge = addEdge(label,nodes);
		multiplicity.put(newEdge, new AnnotationBounds());
		return newEdge;
	}
	
	/**
	 * Adds an annotated edge to the graph
	 * @param label The edge label to be used
	 * @param bounds The annotation bounds of the edge
	 * @param nodes The nodes incident to this edge
	 * @return The added annotated edge
	 */
	public Edge addAnnotatedEdge(Label label, AnnotationBounds bounds, Node... nodes){
		
		Edge newEdge = addEdge(label,nodes);
		multiplicity.put(newEdge, bounds);
		return newEdge;
	}
	
	/**
	 * Adds an annotated edge to the graph
	 * @param label The edge label to be used
	 * @param bounds The annotation bounds of the edge
	 * @param nodes The list of nodes incident to this edge
	 * @return The added annotated edge
	 */
	public Edge addAnnotatedEdge(Label label, AnnotationBounds bounds, List<Node> nodes){
		
		Edge newEdge = addEdge(label,nodes);
		multiplicity.put(newEdge, bounds);
		return newEdge;
	}	
	
	/**
	 * Adds an annotated edge to the graph
	 * @param label The edge label to be used
	 * @param min The minimal annotation bound
	 * @param max The maximal annotation bound
	 * @param nodes The nodes incident to the edge
	 * @return The added annotated edge
	 */
	public Edge addAnnotatedEdge(Label label, int min, int max, Node... nodes){
		
		Edge newEdge = addEdge(label,nodes);
		multiplicity.put(newEdge, new AnnotationBounds(min,max));
		return newEdge;
	}
	
	/**
	 * Adds an annotated edge to the graph
	 * @param label The edge label to be used
	 * @param min The minimal annotation bound
	 * @param max The maximal annotation bound
	 * @param nodes The list of nodes incident to the edge
	 * @return The added annotated edge
	 */
	public Edge addAnnotatedEdge(Label label, int min, int max, List<Node> nodes){
		
		Edge newEdge = addEdge(label,nodes);
		multiplicity.put(newEdge, new AnnotationBounds(min,max));
		return newEdge;
	}
	
	//----------------------------------------------------------------------------------------
	//Remove [0,0] annotated edge section
	//----------------------------------------------------------------------------------------
	
	/**
	 * Removes an edge from the graph
	 * @param edge The edge to be removed
	 */
	public void removeAnnotatedEdge(Edge edge){
		
		for(HashMap<GraphElement,AnnotationBounds> mult: multiplicities){
			mult.remove(edge);
		}
		super.removeEdge(edge);
	}
	
	
	public Edge reAddAnnotatedEdge(Label label, List<Node> nodes){
		
		Edge newEdge = addEdge(label,nodes);
		for(HashMap<GraphElement,AnnotationBounds> mult: multiplicities){
			mult.put(newEdge, new AnnotationBounds(0,0));
		}
		return newEdge;
	}
	
	//----------------------------------------------------------------------------------------
	//Public methods
	//----------------------------------------------------------------------------------------
	
	/**
	 * Gets the active annotation bound of a graph element
	 * @param element The graph element whose active annotation bound is asked for
	 * @return The active annotation bound 
	 */
	public AnnotationBounds getAnnotationBounds(GraphElement element){
		
		return multiplicity.get(element);
	}

	/**
	 * Gets the active min bound of a graph element
	 * @param element The graph element whose minimal bound is asked for
	 * @return The active minimal bound value
	 */
	public int getMinBound(GraphElement element){
		
		return multiplicity.get(element).getMin();
	}

	/**
	 * Gets the list of min bounds of a graph element
	 * @param element The graph element whose minimal bounds is asked for
	 * @return A list of minimal bound values
	 */
	public List<Integer> getMinBounds(GraphElement element){
		
		List<Integer> bounds = new LinkedList<Integer>();
		for(HashMap<GraphElement,AnnotationBounds> mult: multiplicities) {
			bounds.add(mult.get(element).getMin());
		}
		return bounds;
	}
	
	/**
	 * Gets the active max bound of a graph element
	 * @param element The graph element whose maximal bound is asked for
	 * @return The active maximal bound value
	 */
	public int getMaxBound(GraphElement element){
		
		return multiplicity.get(element).getMax();
	}
	
	/**
	 * Gets the list of max bounds of a graph element
	 * @param element The graph element whose maximal bounds is asked for
	 * @return A list of maximal bound values
	 */
	public List<Integer> getMaxBounds(GraphElement element){
		
		List<Integer> bounds = new LinkedList<Integer>();
		for(HashMap<GraphElement,AnnotationBounds> mult: multiplicities) {
			bounds.add(mult.get(element).getMax());
		}
		return bounds;
	}
	
	/**
	 * Gets the String representation of the elements boundaries
	 * @param element The graph element whose bounds are asked for
	 * @return The String representing the elements current bounds
	 */
	public String getBoundLabel(GraphElement element){
		if(multiplicity.get(element).getMin() == limit){
			return "[m,m]";
		}
		else if(multiplicity.get(element).getMax() == limit){
			return "[" + multiplicity.get(element).getMin() + ",m]";
		}
		else{
			return "[" + multiplicity.get(element).getMin() + "," + multiplicity.get(element).getMax() + "]";
		}
	}
	
	/**
	 * Gets the limit of this graph
	 * @return The limit of this graph
	 */
	public int getLimit() 
	{
		return limit;
	}

	/**
	 * Sets the limit of this graph
	 * @param limit The limit to be used
	 */
	public void setLimit(int limit) 
	{
		this.limit = limit;
	}
	
	/**
	 * Gets the active multiplicity
	 * @return The currently active multiplicity
	 */
	public HashMap<GraphElement, AnnotationBounds> getCurrentMultiplicity() {
		return multiplicity;
	}

	/**
	 * Sets the active multiplicity
	 * @param multiplicity The multiplicity to be used
	 */
	private void setMultiplicity(HashMap<GraphElement, AnnotationBounds> multiplicity) {
		this.multiplicity = multiplicity;
	}
	
	/**
	 * Gets a node in this graph based on an ID
	 * @param id The ID of the node to be returned
	 * @return The node with the specific ID
	 */
	public Node getNodeByID(int id)
	{
		for(Node n: nodes)
		{
			if(n.getId() == id){return n;}
		}
		return null;
	}
	
	/**
	 * Gets a list of edges in  this graph
	 * 
	 * @return The list of edges contained
	 */
	public List<Edge> getEdgeList() {
		List<Edge> result = new LinkedList<Edge>();
		for(Edge e : edges) {
			result.add(e);
		}
		return result;
	}
	
	/**
	 * Gets a list of multiplicities for the multiply annotated type graph
	 * 
	 * @return              
	 *      the set of multiplicities as list
	 */
	public List<HashMap<GraphElement, AnnotationBounds>> getMultiplicities() {
		return multiplicities;
	}
	
	/**
	 * Adds a multiplicity to the set of annotations for this multiply annotated type graph
	 * 
	 * @param multiplicity          
	 *      the multiplicity to be added to the set of multiplicities
	 * @return Boolean if the addition was successful
	 */
	public boolean addMultiplicity(HashMap<GraphElement, AnnotationBounds> multiplicity) {
		
		if(!multiplicity.keySet().containsAll(getNodes())) {
			System.out.println("Given multiplicity does not define a bound for all nodes");
			return false;
		}
		if(!multiplicity.keySet().containsAll(getEdgeList())) {
			System.out.println("Given multiplicity does not define a bound for all edges");
			return false;
		}
		
		boolean newMult = multiplicities.isEmpty();
		for(HashMap<GraphElement, AnnotationBounds> oldMultiplicity : multiplicities) {
			newMult = false;
			for(Node n: getNodes()) {
				if((multiplicity.get(n).getMin() < oldMultiplicity.get(n).getMin()) || (oldMultiplicity.get(n).getMax() < multiplicity.get(n).getMax())) {
					newMult = true;
					break;
				}					
			}	
			for(Edge e: getEdgeList()) {
				if((multiplicity.get(e).getMin() < oldMultiplicity.get(e).getMin()) || (oldMultiplicity.get(e).getMax() < multiplicity.get(e).getMax())) {
					newMult = true;
					break;
				}					
			}
			if(!newMult) break;			
		}
		
		if(newMult) return multiplicities.add(multiplicity);
		return false;
	}
	
	/**
	 * Removes a multiplicity from the set of annotations for this multiply annotated type graph
	 * 
	 * @param index          
	 *      the multiplicity index to be removed
	 * @return Boolean if the removal was successful
	 */
	public boolean removeMultiplicity(int index) {
		
		if(multiplicities.size() == 1){
			return false;
		}
		multiplicities.remove(index);
		selectMultiplicity(0);
		return true;

	}
	
	/**
	 * Returns the annotated type graph with the n-th multiplicity stored in the set of multiplicities
	 * or the current annotated type graph if the position is undefined. 
	 * 
	 * @param n          
	 *      position of the multiplicity
	 * @return 
	 * 		the annotated type graph with the n-th multiplicity
	 */		
	public AnnotatedTypeGraph selectMultiplicity(int n) {
		if(n < multiplicities.size()) {
			setMultiplicity(multiplicities.get(n));
			activeMultiplicityIndex = n;
		}
		return this;
	}
	
	/**
	 * Returns the number of multiplicities stored in the annotation set of this multiply annotated type graph 
	 *
	 * @return 
	 * 		the number of multiplicities
	 */
	public int getNumberOfMultiplicities() {
		return multiplicities.size();
	}
	
	/**
	 * Returns the position of the current multiplicity of this multiply annotated type graph graph
	 * @return the activeMultiplicityIndex
	 */
	public int getActiveMultiplicityIndex() {
		return activeMultiplicityIndex;
	}
	
	/**
	 * Returns the node copy of a copied graph
	 * @param orig The original node that was copied
	 * @return The node copy
	 */
	public Node getCopiedNode(Node orig) {
		return nodeMap.get(orig);
	}
}
