/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.data;

import java.util.HashSet;
import java.util.Set;

import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.Morphism;
import de.uni_due.inf.ti.graph.Node;

/**
 * Class of rewritable materializations.
 * 
 * @author Dennis Nolte
 */
public class Materialization {

		private Morphism leftEmbedding;
		private Morphism rightEmbedding;
		
		/**
		 * Constructor for a new Materialization
		 * @param leftEmbedding The embedding of the left-hand side into the materialization graph
		 * @param rightEmbedding The embedding of the materialization graph into the multiply annotated type graph
		 */
		public Materialization(Morphism leftEmbedding, Morphism rightEmbedding) {

			this.leftEmbedding = leftEmbedding;
			this.rightEmbedding = rightEmbedding;
		}

		/**
		 * Gets the base morphisms from the left-hand side graph into the multiply annotated type graph
		 * @return The base morphism 
		 */
		public Morphism getBaseMorphism() {
			return leftEmbedding.compose(rightEmbedding);
		}

		/**
		 * Gets the morphism from the left-hand side graph into the materialization graph
		 * @return The embedding morphism
		 */
		public Morphism getLeftEmbedding() {
			return leftEmbedding;
		}

		/**
		 * Gets the morphism from the materialization graph into the multiply annotated type graph
		 * @return The embedding morphism
		 */
		public Morphism getRightEmbedding() {
			return rightEmbedding;
		}

		/**
		 * Gets the left-hand side graph of this materialization
		 * @return The left-hand side graph
		 */
		public AnnotatedTypeGraph getLeftHandSideGraph() {
			return (AnnotatedTypeGraph) leftEmbedding.getDomain();
		}

		/**
		 * Gets the materialization graph of this materialization
		 * @return The materialization graph
		 */
		public AnnotatedTypeGraph getMaterialization() {
			return (AnnotatedTypeGraph) leftEmbedding.getCodomain();
		}

		/**
		 * Gets the multiply annotated type graph of this materialization
		 * @return The multiply annotated type graph
		 */
		public AnnotatedTypeGraph getAbstractGraph() {
			return (AnnotatedTypeGraph) rightEmbedding.getCodomain();
		}
		
		/**
		 * Gets the node images of the left-hand side embedding
		 * @return The set of nodes
		 */
		public Set<Node> getLeftEmbeddingNodeImage() {
			Set<Node> image = new HashSet<Node>();
			for(Node n: leftEmbedding.getNodeMap().values()) {
				image.add(n);
			}
			return image;
		}

		/**
		 * Gets the edge images of the left-hand side embedding
		 * @return The set of edges
		 */
		public Set<Edge> getLeftEmbeddingEdgeImage() {
			Set<Edge> image = new HashSet<Edge>();
			for(Edge e: leftEmbedding.getEdgeMap().values()) {
				image.add(e);
			}
			return image;
		}
		
}
