/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.data;

import java.util.HashMap;
import java.util.Set;

import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.Morphism;
import de.uni_due.inf.ti.graph.Node;

/**
 * Class of postconditions.
 * 
 * @author Dennis Nolte
 */
public class Postcondition {

		private AnnotatedTypeGraph postGraph;
		private Set<Node> rightHandNodes;
		private Set<Edge> rightHandEdges;
		private Morphism legalMorphism;
		private HashMap<Integer,Integer> legalAnnotationIndexes;
		
		/**
		 * Constructor of the postcondition
		 * @param postGraph The annotated type graph representing the postcondition
		 * @param rightHandNodes The image nodes of the right-hand side embedding
		 * @param rightHandEdges The image edges of the right-hand side embedding
		 */
		public Postcondition(AnnotatedTypeGraph postGraph, Set<Node> rightHandNodes, Set<Edge> rightHandEdges) {

			this.postGraph = postGraph;
			this.rightHandNodes = rightHandNodes;
			this.rightHandEdges = rightHandEdges;
			this.legalMorphism = null;
			legalAnnotationIndexes = null;
		}
		
		/**
		 * Get the annotated type graph of this postcondition
		 * @return the post graph
		 */
		public AnnotatedTypeGraph getPostGraph() {
			return postGraph;
		}

		/**
		 * Get the right-hand side nodes of this postcondition
		 * @return the right-hand Nodes
		 */
		public Set<Node> getRightHandNodes() {
			return rightHandNodes;
		}
		
		/**
		 * Get the right-hand side edges of this postcondition
		 * @return the right-hand Edges
		 */
		public Set<Edge> getRightHandEdges() {
			return rightHandEdges;
		}
		
		/**
		 * Get the legal morphism from the postcondition to the annotated type graph
		 * @return the legal morphism
		 */
		public Morphism getLegalMorphism() {
			return legalMorphism;
		}

		/**
		 * Set the legal morphism from the postcondition to the annotated type graph
		 * @param legalMorphism the legal morphism to be set
		 */
		public void setLegalMorphism(Morphism legalMorphism) {
			this.legalMorphism = legalMorphism;
		}
		
		/**
		 * Get the legal morphism's annotation index mapping
		 * @return the legal morphism's annotation index mapping
		 */
		public HashMap<Integer,Integer> getAnnotationIndexes() {
			return legalAnnotationIndexes;
		}

		/**
		 * Set the legal morphism's annotation index mapping
		 * @param legalAnnotationIndexes the legal morphism's annotation index mapping
		 */
		public void setAnnotationIndexes(HashMap<Integer,Integer> legalAnnotationIndexes) {
			this.legalAnnotationIndexes = legalAnnotationIndexes;
		}
		
		
}
