/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.EnumSet;
import java.util.HashMap;

import javax.swing.BorderFactory;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.ListSelectionModel;
import javax.swing.ScrollPaneConstants;
import javax.swing.SpinnerListModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.dragom.data.AnnotationBounds;
import de.uni_due.inf.ti.dragom.general.ResourceKeys;
import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.GraphElement;
import de.uni_due.inf.ti.graph.Node;
import de.uni_due.inf.ti.swing.StandardAction;
import de.uni_due.inf.ti.swing.StandardDialog;
import de.uni_due.inf.ti.swing.UserInputException;
import de.uni_due.inf.ti.visigraph.ArrowHead;
import de.uni_due.inf.ti.visigraph.Style;
import de.uni_due.inf.ti.visigraph.VxComponent;
import de.uni_due.inf.ti.visigraph.VxEdge;
import de.uni_due.inf.ti.visigraph.VxExplicitNode;
import de.uni_due.inf.ti.visigraph.VxGraph;
import de.uni_due.inf.ti.visigraph.swing.GraphPanel;

/**
 * Dialog which displays the add multiplicity dialog.
 * 
 * @author Dennis Nolte
 */
class AddMultiplicityDialog extends StandardDialog<HashMap<GraphElement, AnnotationBounds>>
{
    private static final long serialVersionUID = -5346791005465206516L;
    private AnnotatedTypeGraph dataGraph;
    private VxGraph vxGraph;
    private JButton check;
    private JSpinner minSpinner;
    private JSpinner maxSpinner;
    
    private JScrollPane scroll;
    private GraphPanel graphView;
    private DefaultListModel<String> listModel;
    private DefaultListModel<String> multModel;
    private HashMap<VxComponent,Integer> vxComponentToInt;
    private HashMap<Integer,GraphElement> intToData;
    
    private JList<String> graphList;
    private JList<String> multList;
    private int limit;
    
    HashMap<GraphElement,AnnotationBounds> multiplicity;
    
    public AddMultiplicityDialog(Frame owner,AnnotatedTypeGraph dataGraph)
    {
        super(owner, ResourceKeys.TIT_CREATE_TYPE_GRAPH, EnumSet.of(Button.ACCEPT, Button.CANCEL));
        setAcceptKey(ResourceKeys.CMD_APPLY);
        setCancelKey(ResourceKeys.CMD_CANCEL);
        
        this.dataGraph= dataGraph;
        this.limit = dataGraph.getLimit();
        vxComponentToInt = new HashMap<VxComponent,Integer>();
        intToData = new HashMap<Integer,GraphElement>();
        multiplicity = new HashMap<GraphElement,AnnotationBounds>();
    }
    

    /**
     * {@inheritDoc}
     */
    @Override
    protected JComponent getDialogComponent()
    {
    	JPanel result = new JPanel();
    	result.setPreferredSize(new Dimension(1000,700));
    	result.setLayout(new BorderLayout(0,0));
    	result.setOpaque(true);

    	JPanel margin = new JPanel();
    	margin.setPreferredSize(new Dimension(1000,1));
    	   	
    	result.add(margin, BorderLayout.NORTH);
        result.add(createGraphPanel(), BorderLayout.CENTER);
        result.add(createOptionsPanel(), BorderLayout.WEST);

        result.setBorder(BorderFactory.createMatteBorder(1, 1, 1, 1, Color.BLACK));
        
        getContentPane().setBackground(new Color(0,0,0,0));
        getRootPane().setBackground(new Color(0,0,0,0));
        setUndecorated(true);
        setBackground(new Color(0,0,0,0));
        
        return result;
    }

    
    @Override
    protected HashMap<GraphElement, AnnotationBounds> parseOptions() throws UserInputException 
    {
    	for(Node n: dataGraph.getNodes()) {
    		if(!multiplicity.keySet().contains(n)) {
    			multiplicity.put(n, new AnnotationBounds(0, limit));
    		}
    	}
    	
    	for(Edge e: dataGraph.getEdges()) {
    		if(!multiplicity.keySet().contains(e)) {
    			multiplicity.put(e, new AnnotationBounds(0, limit));
    		}
    	}
    	   	
        return multiplicity;
    }
	
	private JPanel createOptionsPanel() {
		String labelSuffix = ": ";

		FormLayout layout = new FormLayout("50dlu, 3dlu, 30dlu:grow", "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout);
		builder.border(Borders.DIALOG);
		builder.background(Color.WHITE);
		builder.appendSeparator("Graph Elements");
		builder.nextLine();

		listModel = new DefaultListModel<String>();
		multModel = new DefaultListModel<String>();
		int index = 0;
		
		for(VxExplicitNode vxe: vxGraph.getNodes()) {
			
			Node dataNode = (Node) vxe.getReference();
			listModel.addElement("Vertex " + dataNode.getId());
			multModel.addElement("[0,m]");
			vxComponentToInt.put(vxe, index);
			intToData.put(index, dataNode);
			index++;
		}
		
		for(VxEdge vxe: vxGraph.getEdges()) {
			
			Edge dataEdge = (Edge) vxe.getReference();
			listModel.addElement("Edge " + dataEdge.getLabel().getName() + " (" + dataEdge.getSource().getId() + "-" + dataEdge.getTarget().getId() + ")");
			multModel.addElement("[0,m]");
			vxComponentToInt.put(vxe, index);
			intToData.put(index, dataEdge);
			index++;
		}

		graphList = new JList<String>(listModel);
		graphList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		graphList.setLayoutOrientation(JList.VERTICAL);
		graphList.setVisibleRowCount(-1);
		graphList.setEnabled(false);	
		

		graphList.addListSelectionListener(new ListSelectionListener() {
		    
			@Override
			public void valueChanged(ListSelectionEvent e) {
				if (e.getValueIsAdjusting() == false) {

			        if (graphList.getSelectedIndex() == -1) {
			            check.setEnabled(false);
			        } else {
			        	check.setEnabled(true);
			        }
			    }	
			}
		});
		
		
		
		JScrollPane listScroller = new JScrollPane(graphList);
		listScroller.setPreferredSize(new Dimension(60, 530));
		
		multList = new JList<String>(multModel);
		multList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		multList.setLayoutOrientation(JList.VERTICAL);
		multList.setVisibleRowCount(-1);
		multList.setEnabled(false);	
		
		JScrollPane multScroller = new JScrollPane(multList);
		multScroller.setPreferredSize(new Dimension(60, 530));
		
		builder.append(listScroller,multScroller);
		
		

		builder.nextLine();
		builder.appendSeparator("Multiplicity");
		builder.nextLine();
		
		minSpinner = new JSpinner();
		minSpinner.setModel(new SpinnerListModel(new String[] {"0"}));
		minSpinner.addChangeListener(new ChangeListener() {
			
			@Override
			public void stateChanged(ChangeEvent e) {
				if(minSpinner.getValue().equals("m")){
					maxSpinner.setValue("m");
				}
				else if(maxSpinner.getValue().equals("m")){
					
				} else if(Integer.parseInt((String) minSpinner.getValue()) > Integer.parseInt((String) maxSpinner.getValue())){
					maxSpinner.setValue(minSpinner.getValue());
				}
			}
		});

		builder.append("Min Bound" + labelSuffix, minSpinner);
		builder.nextLine();
		
		maxSpinner = new JSpinner();
		maxSpinner.setModel(new SpinnerListModel(new String[] {"0"}));
		maxSpinner.addChangeListener(new ChangeListener() {
			
			@Override
			public void stateChanged(ChangeEvent e) {
				if(minSpinner.getValue().equals("m")){
					minSpinner.setValue(maxSpinner.getValue());
				}
				else if(maxSpinner.getValue().equals("m")){
					
				} else if(Integer.parseInt((String) minSpinner.getValue()) > Integer.parseInt((String) maxSpinner.getValue())){
					minSpinner.setValue(maxSpinner.getValue());
				}				
				
			}
		});

		builder.append("Max Bound" + labelSuffix, maxSpinner);
		
		int maxValue = dataGraph.getLimit();
		String[] annotations = new String[maxValue + 1];
		
		for(int i=0; i<maxValue; i++){
			annotations[i] = "" + i;
		}
		annotations[maxValue] = "m";
		
		minSpinner.setModel(new SpinnerListModel(annotations));
		maxSpinner.setModel(new SpinnerListModel(annotations));
		maxSpinner.setValue(annotations[maxValue]);
		
		builder.nextLine();	
		
		check = new JButton(new AddActiveMultiplicityAction());
		check.setEnabled(false);
		builder.append(check,3);
		builder.nextLine();

		return builder.getPanel();
	}


	private JScrollPane createGraphPanel() {
                     
        Style edgeStyle = new Style();
        edgeStyle.setLineColor(new Color(.3f, .3f, .3f));
        edgeStyle.setLineColor(true, new Color(.6f, .6f, .6f));
        edgeStyle.setFrontArrow(ArrowHead.newTriangleArrowHead(8, 8));
        edgeStyle.setStroke(Style.getDefaultStroke(Style.LineWidth.MEDIUM, 
                Style.DashType.SOLID));
        
        Style nodeStyle = new Style();
        nodeStyle.setFilled(true);
        nodeStyle.setFillColor(false, Color.WHITE);
        nodeStyle.setFillColor(true, Color.LIGHT_GRAY);  
		
        graphView = new GraphPanel();
        graphView.addItemListener(new ItemListener() {
			
			@Override
			public void itemStateChanged(ItemEvent e) {
				switch(e.getStateChange()) {
				case ItemEvent.DESELECTED: graphList.clearSelection(); multList.clearSelection(); break;
				default: 
					for(VxComponent vxe: graphView.getSelectedObjects()){
						if(vxComponentToInt.get(vxe) != null) {
							graphList.setSelectedIndex(vxComponentToInt.get(vxe));
							multList.setSelectedIndex(vxComponentToInt.get(vxe));
						}
					}
				}
								
			}
		});
        
        AnnotatedTypeGraphVisualizer visualizer = VisiTools.getInstance().getVisualizer();
        visualizer.addDefaultStyles(graphView.getDrawer().getStyleMap());
        visualizer.setMultiplicityStyle(false);
        visualizer.setNodeIDs(true);
        vxGraph = visualizer.visualizeGraph(dataGraph);
        graphView.setGraph(vxGraph);
        visualizer.setMultiplicityStyle(true);
        visualizer.setNodeIDs(false);
        
        graphView.setMultiSelect(false);
        
        scroll = new JScrollPane(graphView);
        scroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        scroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
        
        return scroll;
	}
    
    // ---------------------------------------------------------------------- //
    //                              Actions                                   //
    // ---------------------------------------------------------------------- //
    
	/**
     * Action which shows the add multiplicity dialog
     */
    
    private class AddActiveMultiplicityAction extends StandardAction
    {
        /** 
         * Creates a new AddActiveMultiplicityAction instance.
         */
    	AddActiveMultiplicityAction()
        {
            super(ResourceKeys.CMD_ADD_MULTIPLICITY);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
			
			multModel.set(multList.getSelectedIndex(), "[" + getMinValue() + "," + getMaxValue() + "]");
			
			String minValue = getMinValue();
			String maxValue = getMaxValue(); 
			
			if(minValue.equals("m")){
				multiplicity.put(intToData.get(multList.getSelectedIndex()), new AnnotationBounds(limit, limit));    			
    		}
    		else if(maxValue.equals("m")){
    			multiplicity.put(intToData.get(multList.getSelectedIndex()), new AnnotationBounds(Integer.parseInt(minValue), limit));
    		}
    		else{
    			multiplicity.put(intToData.get(multList.getSelectedIndex()), new AnnotationBounds(Integer.parseInt(minValue), Integer.parseInt(maxValue)));
    		}		
		}
    
    }
    
	// ---------------------------------------------------------------------- //
    //                              Informations                              //
    // ---------------------------------------------------------------------- //
	
    /**
     * Gets the currently set minimal bound value
     * @return The spinners minimal bound value
     */
	public String getMinValue()
	{
		return minSpinner.getValue() + "";
	}

    /**
     * Gets the currently set maximal bound value
     * @return The spinners maximal bound value
     */
	public String getMaxValue()
	{
		return maxSpinner.getValue() + "";
	}

}
