/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.gui;

import java.awt.geom.Point2D;
import java.util.HashMap;

import de.uni_due.inf.ti.annotation.Nullable;
import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.graph.Label;
import de.uni_due.inf.ti.graph.Node;
import de.uni_due.inf.ti.swing.StandardDialog;
import de.uni_due.inf.ti.visigraph.VxEdge;
import de.uni_due.inf.ti.visigraph.VxGraph;
import de.uni_due.inf.ti.visigraph.VxLabelNode;
import de.uni_due.inf.ti.visigraph.VxNode;
import de.uni_due.inf.ti.visigraph.swing.GraphEditSynchronizer;

/**
 * AnnotatedGraphEditSynchronizer which uses {@link VxGraph} as the underlying model.
 * 
 * @author Dennis Nolte
 */
public class AnnotatedTypeGraphEditSynchronizer implements GraphEditSynchronizer
{
    // Graph characteristics
    private boolean edgesToLabels;
    private boolean enabled;
    private int maxLabels;
    private StandardDialog<?> informationDealer;
    private AnnotatedTypeGraph dataGraph;
    private HashMap<VxNode,Node> nodeMap;
    
    
    // Styles
    private String newNodeStyleName;
    private String newLabelStyleName;
    private String newEdgeStyleName;
        
    /**
     * Creates a new BasicGraphEditSynchronizer. 
     * The new BasicGraphEditSynchronizer does not allow edges to have edge
     * labels as source or target, and allows a single label per edge.
     */
    public AnnotatedTypeGraphEditSynchronizer()
    {
        this(false, 1, null);
    }
    
    /**
     * Creates a new BasicGraphEditSynchronizer with custom behavior.
     * 
     * @param edgesToLabels
     *      whether the new BasicGraphEditSynchronizer allows edges to have
     *      edge labels as source or target
     * @param maxLabels
     *      maximum number of label nodes on an edge; -1 means unlimited
     * @throws IllegalArgumentException
     *      if {@code maxLabels < -1}
     */
    public AnnotatedTypeGraphEditSynchronizer(boolean edgesToLabels, int maxLabels)
    {
        this(edgesToLabels, maxLabels, null);
    }
  
    public AnnotatedTypeGraphEditSynchronizer(boolean edgesToLabels, int maxLabels,@Nullable StandardDialog<?> informationDealer)
    {
        if (maxLabels < -1) throw new IllegalArgumentException("maxLabels must be >= -1");
        this.edgesToLabels = edgesToLabels;
        this.maxLabels = maxLabels;
        this.informationDealer = informationDealer;
        dataGraph = new AnnotatedTypeGraph();
        nodeMap = new HashMap<>();
        enabled = true;
    }
    
    /**
     * Returns the name of the style of nodes that are added to the graph.
     * @return The String which represents the style
     */
    public String getNewNodeStyleName()
    {
        return newNodeStyleName;
    }
    
    /**
     * Sets the name of the style of nodes that are added to the graph. 
     * 
     * @param style
     *      style of nodes added to the graph
     */
    public void setNewNodeStyleName(String style)
    {
        newNodeStyleName = style;
    }
    
    /**
     * Returns the name of the style of labels that are added to the graph.
     * @return The String which represents the style
     */
    public String getNewLabelStyleName()
    {
        return newLabelStyleName;
    }
    
    /**
     * Sets the style of labels that are added to the graph by the user.
     * @param style
     * 		style of labels added to the graph
     */
    public void setNewLabelStyleName(String style)
    {
        newLabelStyleName = style;
    }   
    
    /**
     * Returns the name of the style of edges that are added to the graph.
     * @return The String which represents the style
     */
    public String getNewEdgeStyleName()
    {
        return newEdgeStyleName;
    }
    
    /**
     * Sets the name of the style of edges that are added to the graph by this 
     * GraphEditor.
     * @param style
     *    style of edges added to the graph
     */
    public void setNewEdgeStyleName(String style)
    {
        newEdgeStyleName = style;
    }   
  
    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isPossibleSourceNode(VxNode source)
    {
        if (edgesToLabels) {
            return true;
        }
        else {
            return !(source instanceof VxLabelNode);
        }
    }
    
    /**
     * {@inheritDoc}
     * <p>
     * This implementation simply returns {@code true}.
     */
    @Override
    public boolean canAddNode(VxGraph graph)
    {
        return true;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public boolean canAddEdge(VxNode source, VxNode target)
    {
        if (!edgesToLabels) {
            if (source instanceof VxLabelNode) return false;
            if (target instanceof VxLabelNode) return false;
        }
        return true;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public boolean canAddLabel(VxEdge edge)
    {
    	if(enabled){
	        int numOfLabels = edge.getLabels().size(); // throws NPE when edge == null
	        return (maxLabels < 0 || numOfLabels + 1 <= maxLabels);
    	}
    	return false;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public boolean canEditLabel(VxNode node)
    {
        return true;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void addNode(VxGraph graph, Point2D pt) 
    {
    	if(enabled){
    		if(informationDealer != null){
    			VxNode newNode = graph.addNode(pt,((CreateTypeGraphDialog) informationDealer).getNodeLabel());
        		newNode.setStyle(newNodeStyleName);
        		
        		String minValue = ((CreateTypeGraphDialog) informationDealer).getMinValue();
        		String maxValue = ((CreateTypeGraphDialog) informationDealer).getMaxValue();
        		int limit = ((CreateTypeGraphDialog) informationDealer).getLimit();
        		
        		if(minValue.equals("m")){
        			nodeMap.put(newNode, dataGraph.addAnnotatedNode(limit, limit));
        		}
        		else if(maxValue.equals("m")){
        			nodeMap.put(newNode, dataGraph.addAnnotatedNode(Integer.parseInt(minValue), limit));
        		}
        		else{
        			nodeMap.put(newNode, dataGraph.addAnnotatedNode(Integer.parseInt(minValue), Integer.parseInt(maxValue)));
        		}	
    		}
    		
    	}
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void addEdge(VxNode source, VxNode target) 
    {
    	if(enabled){
    		if(informationDealer != null){
    			VxGraph graph = source.getGraph();
    	        VxEdge edge = graph.addEdge(source, target, newEdgeStyleName); 
    	        
    	        String edgeLabel = ((CreateTypeGraphDialog) informationDealer).getEdgeLabel();
        		String minValue = ((CreateTypeGraphDialog) informationDealer).getMinValue();
        		String maxValue = ((CreateTypeGraphDialog) informationDealer).getMaxValue();
        		int limit = ((CreateTypeGraphDialog) informationDealer).getLimit();
        		
        		addLabel(edge, edgeLabel +" [" + minValue + ";" + maxValue + "]");
        		if(minValue.equals("m")){
        			dataGraph.addAnnotatedEdge(new Label(edgeLabel), limit, limit, nodeMap.get(source), nodeMap.get(target));
        		}
        		else if(maxValue.equals("m")){
        			dataGraph.addAnnotatedEdge(new Label(edgeLabel), Integer.parseInt(minValue), limit, nodeMap.get(source), nodeMap.get(target));
        		}
        		else{
        			dataGraph.addAnnotatedEdge(new Label(edgeLabel), Integer.parseInt(minValue), Integer.parseInt(maxValue), nodeMap.get(source), nodeMap.get(target));
        		}
    		}
    	}
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void addLabel(VxEdge edge, String label)
    {
    	if(enabled){
	        VxLabelNode ln = edge.addLabel(label);
	        if (newLabelStyleName != null) {
	            ln.setStyle(newLabelStyleName);
	        }
    	}
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void editLabel(VxNode node, String newLabel)
    {
        node.setLabel(newLabel);
    }
    
    public AnnotatedTypeGraph getResultGraph(){
    	return dataGraph;
    }
    
    /**
     * Sets whether this synchronizer may add anything to the graph
     * @param bool 
     * 		whether it is allowed or not
     */
    public void setEnabled(boolean bool)
    {
    	enabled = bool;
    }
}

