/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.dragom.general.ResourceKeys;
import de.uni_due.inf.ti.graph.Morphism;
import de.uni_due.inf.ti.swing.GuiContext;
import de.uni_due.inf.ti.swing.StandardDialog;
import de.uni_due.inf.ti.swing.UserInputException;
import de.uni_due.inf.ti.visigraph.VxGraph;
import de.uni_due.inf.ti.visigraph.swing.GraphPanel;

/**
 * Dialog which displays the choose base morphism dialog.
 * 
 * @author Dennis Nolte
 */
class ChooseBaseMorphismDialog extends StandardDialog<Morphism>
{
	private static final long serialVersionUID = -4255861163073697037L;
	
	private Morphism currentBaseMorphism;
	private List<Morphism> allBaseMorphisms;
	private HashMap<Morphism,List<Integer>> morphAnnotPairs;
    private List<VxGraph> visis;
    
    @SuppressWarnings("unused")
	private int totalNumberOfAnnotations;
   
    private JScrollPane scroll;
	private GraphPanel morphismView;

	private JComboBox<String> morphismCombo;

    public ChooseBaseMorphismDialog(Frame owner,HashMap<Morphism,List<Integer>> morphAnnotPairs)
    {
        super(owner, ResourceKeys.TIT_CHOOSE_BASE_MORPHISM, EnumSet.of(Button.ACCEPT, Button.CANCEL));
        setAcceptKey(ResourceKeys.CMD_APPLY);
        setCancelKey(ResourceKeys.CMD_CANCEL);

        visis = new ArrayList<>();
        this.morphAnnotPairs = morphAnnotPairs;
        
        allBaseMorphisms = new LinkedList<Morphism>();
        for(Morphism m: morphAnnotPairs.keySet()){
        	allBaseMorphisms.add(m);
        }
        currentBaseMorphism = allBaseMorphisms.get(0);
        totalNumberOfAnnotations = ((AnnotatedTypeGraph)currentBaseMorphism.getCodomain()).getNumberOfMultiplicities();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected JComponent getDialogComponent()
    {
    	JPanel result = new JPanel();
    	result.setPreferredSize(new Dimension(600,600));
    	result.setLayout(new BorderLayout(0,1));
    	result.setOpaque(false);
    	
    	JPanel margin = new JPanel();
    	margin.setPreferredSize(new Dimension(1,600));
    	
    	
    	result.add(margin, BorderLayout.WEST);
        result.add(createMorphismPanel(), BorderLayout.CENTER);
        result.add(createMorphismCombo(), BorderLayout.NORTH);
        result.setBorder(BorderFactory.createMatteBorder(1, 1, 1, 1, Color.BLACK));
        
        getContentPane().setBackground(new Color(0,0,0,0));
        getRootPane().setBackground(new Color(0,0,0,0));
        setUndecorated(true);
        setBackground(new Color(0,0,0,0));
        return result;
    }

    
    @Override
    protected Morphism parseOptions() throws UserInputException //TODO
    {
        return currentBaseMorphism;
    }
	
	private JScrollPane createMorphismPanel() {
      
		morphismView = new GraphPanel();
        VisiTools.getInstance().getVisualizer().addDefaultStyles(morphismView.getDrawer().getStyleMap());
		
        scroll = new JScrollPane(morphismView);
        scroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        scroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
        
        return scroll;
	}
	
    /**
     * Creates the combo box displaying rule names.
     */
    private JPanel createMorphismCombo()
    {
        DefaultComboBoxModel<String> model = new DefaultComboBoxModel<>(); 
        morphismCombo = new JComboBox<String>(model);
        setMorphisms();
        morphismCombo.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e)
            {
                int index = morphismCombo.getSelectedIndex();
                if (index >= 0) changeMorphism(index);
            }

        });
        
        JLabel morphismLabel = GuiContext.createLabel(ResourceKeys.LAB_SELECT_MORPHISM);        

        JPanel comboPanel = new JPanel();
        comboPanel.setLayout(new BoxLayout(comboPanel, BoxLayout.X_AXIS));
        comboPanel.add(morphismLabel);
        comboPanel.add(Box.createHorizontalStrut(10));
        comboPanel.add(morphismCombo);
        
        return comboPanel;
    }

    /**
     * Changes the morphism displayed by this dialog.
     */
    public void setMorphisms()
    {        
        morphismView.setEnabled(true);
        morphismCombo.setEnabled(true);

        for (int n=0; n < allBaseMorphisms.size();n++) {
        	visis.add(null);
            morphismCombo.addItem("Base Morphism " + (n+1) + " (semi-legal for " + morphAnnotPairs.get(allBaseMorphisms.get(n)).size() + 
            																	//" of " + totalNumberOfAnnotations + 
            																	" annotations)");
        }

        morphismCombo.setSelectedIndex(0);
        changeMorphism(0);
    }
    
    // ---------------------------------------------------------------------- //
    //                              Informations                              //
    // ---------------------------------------------------------------------- //

	// ---------------------------------------------------------------------- //
    //                              Private Methods                           //
    // ---------------------------------------------------------------------- //
		
    /**
     * Changes the morphism which is displayed in the graph view.
     * 
     * @param index
     *      index of the rule to be displayed
     */
    private void changeMorphism(int index) 
    {
    	currentBaseMorphism = allBaseMorphisms.get(index);
    	VxGraph vx = visis.get(index);
    	if (vx == null) {
    		Morphism visiMorph = new Morphism(currentBaseMorphism);
    		((AnnotatedTypeGraph) visiMorph.getCodomain()).selectMultiplicity(morphAnnotPairs.get(currentBaseMorphism).get(0));
    		vx = VisiTools.getInstance().visualizeMatch(visiMorph, false, true);
    		visis.set(index, vx);
    	}
    	morphismView.setGraph(vx);
    }
}
