/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridLayout;
import java.util.EnumSet;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ScrollPaneConstants;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.dragom.general.ResourceKeys;
import de.uni_due.inf.ti.graph.Rule;
import de.uni_due.inf.ti.swing.GuiContext;
import de.uni_due.inf.ti.swing.StandardDialog;
import de.uni_due.inf.ti.swing.UserInputException;
import de.uni_due.inf.ti.visigraph.ArrowHead;
import de.uni_due.inf.ti.visigraph.Style;
import de.uni_due.inf.ti.visigraph.VxGraph;

/**
 * Dialog which displays the rule construction dialog.
 * 
 * @author Dennis Nolte
 */
class CreateRuleDialog extends StandardDialog<Rule>
{
    private static final long serialVersionUID = -5346791005465206516L;
    private AnnotatedTypeGraph leftGraph; 
    private AnnotatedTypeGraph interfaceGraph;
    private AnnotatedTypeGraph rightGraph;
    private RuleLeftGraphEditor leftEditor;
    private RuleInterfaceGraphEditor interfaceEditor;
    private RuleRightGraphEditor rightEditor;
    
    private JScrollPane leftScroll;
    private JScrollPane interfaceScroll;
    private JScrollPane rightScroll;
    
    private JTextField edgeField;
    
    private RuleEditSynchronizer syncer;
    
    public CreateRuleDialog(Frame owner)
    {
        super(owner, ResourceKeys.TIT_CREATE_RULE, EnumSet.of(Button.ACCEPT, Button.CANCEL));
        setAcceptKey(ResourceKeys.CMD_APPLY);
        setCancelKey(ResourceKeys.CMD_CANCEL);
        
        leftGraph = new AnnotatedTypeGraph();
        interfaceGraph = new AnnotatedTypeGraph();
        rightGraph = new AnnotatedTypeGraph();
    }
    

    /**
     * {@inheritDoc}
     */
    @Override
    protected JComponent getDialogComponent()
    {
    	JPanel result = new JPanel();
    	result.setPreferredSize(new Dimension(1150,700));
    	result.setLayout(new BorderLayout(0,0));
    	result.setOpaque(true);

    	JPanel bottomMargin = new JPanel();
    	bottomMargin.setPreferredSize(new Dimension(1000,1));
    	
    	JPanel leftMargin = new JPanel();
    	leftMargin.setPreferredSize(new Dimension(1,700));
    	
    	result.add(bottomMargin, BorderLayout.SOUTH);
    	result.add(leftMargin, BorderLayout.WEST);
        result.add(createRuleEditorPanel(), BorderLayout.CENTER);
        result.add(createOptionsPanel(), BorderLayout.NORTH);
        
        result.setBorder(BorderFactory.createMatteBorder(1, 1, 1, 1, Color.BLACK));
        getContentPane().setBackground(new Color(0,0,0,0));
        getRootPane().setBackground(new Color(0,0,0,0));
        setUndecorated(true);
        setBackground(new Color(0,0,0,0));
        return result;
    }

    
    @Override
    protected Rule parseOptions() throws UserInputException //TODO
    {
    	Rule result = syncer.getResultRule();
    	return result;
    }
	
    
	@SuppressWarnings("static-access")
	private JPanel createRuleEditorPanel() {
      
        JPanel mainPanel = new JPanel();
        mainPanel.setOpaque(false);
        mainPanel.setLayout(new GridLayout(1, 3, 0, 0));
        
        //Set up syncer
        syncer = new RuleEditSynchronizer(false, 1, this);
        syncer.setNewLeftNodeStyleName(VisiTools.getInstance().getVisualizer().createdNodeStyleName);
        syncer.setNewLeftEdgeStyleName(VisiTools.getInstance().getVisualizer().createdEdgeStyleName);
        syncer.setNewRightNodeStyleName(VisiTools.getInstance().getVisualizer().rightNodeStyleName);
        syncer.setNewRightEdgeStyleName(VisiTools.getInstance().getVisualizer().rightEdgeStyleName);
        syncer.setNewLabelStyleName(VisiTools.getInstance().getVisualizer().labelStyleName);
        syncer.setNewInterfaceNodeStyleName("node-style");
        syncer.setNewInterfaceEdgeStyleName("edge-style");
        syncer.setEnabled(true);
                
        Style edgeStyle = new Style();
        edgeStyle.setLineColor(new Color(.3f, .3f, .3f));
        edgeStyle.setLineColor(true, new Color(.6f, .6f, .6f));
        edgeStyle.setFrontArrow(ArrowHead.newTriangleArrowHead(8, 8));
        edgeStyle.setStroke(Style.getDefaultStroke(Style.LineWidth.MEDIUM, 
                Style.DashType.SOLID));

                
        Style nodeStyle = new Style();
        nodeStyle.setFilled(true);
        nodeStyle.setFillColor(false, Color.WHITE);
        nodeStyle.setFillColor(true, Color.LIGHT_GRAY);

        
        
        // Create left Editor and scrollpanel
        leftEditor = new RuleLeftGraphEditor();
        VisiTools.getInstance().getVisualizer().addDefaultStyles(leftEditor.getDrawer().getStyleMap());
        leftEditor.setStyle("node-style", nodeStyle);
        leftEditor.setStyle("edge-style", edgeStyle);
        leftEditor.setGraph(VisiTools.getInstance().visualizeGraph(leftGraph));
        leftEditor.setEditSynchronizer(syncer);
        leftEditor.setAntiAliased(true); 
		
        leftScroll = new JScrollPane(leftEditor);
        leftScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        leftScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
        
        JPanel leftGraphPanel = new JPanel();
    	leftGraphPanel.setLayout(new BorderLayout());
    	
    	JLabel leftGraphLabel = GuiContext.createLabel(ResourceKeys.LAB_LEFT_GRAPH);
    	leftGraphLabel.setText(" " + leftGraphLabel.getText());
    	
    	leftGraphPanel.add(leftGraphLabel, BorderLayout.PAGE_START);
    	leftGraphPanel.add(leftScroll, BorderLayout.CENTER);
    	
    	JPanel margin = new JPanel();
    	margin.setPreferredSize(new Dimension(1,700));
    	leftGraphPanel.add(margin, BorderLayout.WEST);
    	
    	//Create interface editor and scrollpanel
        interfaceEditor = new RuleInterfaceGraphEditor();
        VisiTools.getInstance().getVisualizer().addDefaultStyles(interfaceEditor.getDrawer().getStyleMap());
        interfaceEditor.setStyle("node-style", nodeStyle);
        interfaceEditor.setStyle("edge-style", edgeStyle);
        interfaceEditor.setGraph(VisiTools.getInstance().visualizeGraph(interfaceGraph));
        interfaceEditor.setEditSynchronizer(syncer);
        interfaceEditor.setAntiAliased(true);   
		
        interfaceScroll = new JScrollPane(interfaceEditor);
        interfaceScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        interfaceScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
        
        JPanel interfaceGraphPanel = new JPanel();
        interfaceGraphPanel.setLayout(new BorderLayout());
    	
    	JLabel interfaceGraphLabel = GuiContext.createLabel(ResourceKeys.LAB_INTERFACE_GRAPH);
    	interfaceGraphLabel.setText(" " + interfaceGraphLabel.getText());
    	
    	interfaceGraphPanel.add(interfaceGraphLabel, BorderLayout.PAGE_START);
    	interfaceGraphPanel.add(interfaceScroll, BorderLayout.CENTER);
    	
    	interfaceGraphPanel.add(margin, BorderLayout.WEST);
    	
    	
    	//Create right editor and scrollpanel
        rightEditor = new RuleRightGraphEditor();
        VisiTools.getInstance().getVisualizer().addDefaultStyles(rightEditor.getDrawer().getStyleMap());
        rightEditor.setStyle("node-style", nodeStyle);
        rightEditor.setStyle("edge-style", edgeStyle);
        rightEditor.setGraph(VisiTools.getInstance().visualizeGraph(rightGraph));
        rightEditor.setEditSynchronizer(syncer);
        rightEditor.setAntiAliased(true);   
		
        rightScroll = new JScrollPane(rightEditor);
        rightScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        rightScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
        
        JPanel rightGraphPanel = new JPanel();
    	rightGraphPanel.setLayout(new BorderLayout());
    	
    	JLabel rightGraphLabel = GuiContext.createLabel(ResourceKeys.LAB_RIGHT_GRAPH);
    	rightGraphLabel.setText(" " + rightGraphLabel.getText());
    	
    	rightGraphPanel.add(rightGraphLabel, BorderLayout.PAGE_START);
    	rightGraphPanel.add(rightScroll, BorderLayout.CENTER);
    	
    	rightGraphPanel.add(margin, BorderLayout.WEST);
    	
    	
    	//Add everything to the main panel
    	mainPanel.add(leftGraphPanel, BorderLayout.CENTER);
    	mainPanel.add(interfaceGraphPanel, BorderLayout.CENTER);
    	mainPanel.add(rightGraphPanel, BorderLayout.CENTER);
    	
        return mainPanel;
	}
	
	private JPanel createOptionsPanel() {
		String labelSuffix = ": ";

		FormLayout layout = new FormLayout("left:max(40dlu;pref), 3dlu, 60dlu, 1dlu:grow", "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout);
		builder.border(Borders.DIALOG);
		builder.background(Color.WHITE);
		
		edgeField = new JTextField();
		edgeField.setText("A");
		builder.append("Active Edge Label" + labelSuffix, edgeField);
		builder.nextLine();	

		builder.appendSeparator("Double Pushout Rule");
		builder.nextLine();
		
		return builder.getPanel();
	}
    
    // ---------------------------------------------------------------------- //
    //                              Actions                                   //
    // ---------------------------------------------------------------------- //

    
	// ---------------------------------------------------------------------- //
    //                              Informations                              //
    // ---------------------------------------------------------------------- //

	
	public VxGraph getLeftGraph() {
		return leftEditor.getGraph();
	}
	
	public VxGraph getInterfaceGraph() {
		return interfaceEditor.getGraph();
	}
	
	public VxGraph getRightGraph() {
		return rightEditor.getGraph();
	}
	
	public String getEdgeLabel()
	{
		return edgeField.getText();
	}

	// ---------------------------------------------------------------------- //
    //                              Private Methods                           //
    // ---------------------------------------------------------------------- //


	
}
