/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.text.NumberFormat;
import java.util.EnumSet;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFormattedTextField;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.ScrollPaneConstants;
import javax.swing.SpinnerListModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.text.NumberFormatter;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.dragom.general.ResourceKeys;
import de.uni_due.inf.ti.swing.StandardAction;
import de.uni_due.inf.ti.swing.StandardDialog;
import de.uni_due.inf.ti.swing.UserInputException;
import de.uni_due.inf.ti.visigraph.ArrowHead;
import de.uni_due.inf.ti.visigraph.Style;

/**
 * Dialog which displays the type graph construction dialog.
 * 
 * @author Dennis Nolte
 */
class CreateTypeGraphDialog extends StandardDialog<AnnotatedTypeGraph>
{
    private static final long serialVersionUID = -5346791005465206516L;
    private AnnotatedTypeGraph graph; 
    private JTextField nameField;
    private JTextField mField;
    private JTextField edgeField;
    private JButton check;
    private JSpinner minSpinner;
    private JSpinner maxSpinner;
    
    private JScrollPane scroll;
    
    private AnnotatedTypeGraphEditSynchronizer syncer;
    
    public CreateTypeGraphDialog(Frame owner)
    {
        super(owner, ResourceKeys.TIT_CREATE_TYPE_GRAPH, EnumSet.of(Button.ACCEPT, Button.CANCEL));
        setAcceptKey(ResourceKeys.CMD_APPLY);
        setCancelKey(ResourceKeys.CMD_CANCEL);
        
        graph = new AnnotatedTypeGraph();
    }
    

    /**
     * {@inheritDoc}
     */
    @Override
    protected JComponent getDialogComponent()
    {
    	JPanel result = new JPanel();
    	result.setPreferredSize(new Dimension(1000,700));
    	result.setLayout(new BorderLayout(0,0));
    	result.setOpaque(true);

    	JPanel margin = new JPanel();
    	margin.setPreferredSize(new Dimension(1000,1));
    	
    	
    	result.add(margin, BorderLayout.NORTH);
        result.add(createGraphEditorPanel(), BorderLayout.CENTER);
        result.add(createOptionsPanel(), BorderLayout.WEST);

        result.setBorder(BorderFactory.createMatteBorder(1, 1, 1, 1, Color.BLACK));
        
        getContentPane().setBackground(new Color(0,0,0,0));
        getRootPane().setBackground(new Color(0,0,0,0));
        setUndecorated(true);
        setBackground(new Color(0,0,0,0));
        
        return result;
    }

    
    @Override
    protected AnnotatedTypeGraph parseOptions() throws UserInputException //TODO
    {
    	AnnotatedTypeGraph result = syncer.getResultGraph();
        result.setName(getInputName());
        result.setLimit(getLimit());
        result.addMultiplicity(result.getCurrentMultiplicity());
        return result;
    }
	
	private JPanel createOptionsPanel() {
		String labelSuffix = ": ";

		FormLayout layout = new FormLayout("left:max(40dlu;pref), 3dlu, 60dlu:grow", "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout);
		builder.border(Borders.DIALOG);
		builder.background(Color.WHITE);
		builder.appendSeparator("General  Properties");
		builder.nextLine();

		nameField = new JTextField();
		nameField.setText("AbstractGraph");
		builder.append("Graph Name" + labelSuffix, nameField);
		builder.nextLine();

		mField = new JFormattedTextField(createNumberFormatter());
		builder.append("Value of m" + labelSuffix, mField);
		builder.nextLine();
		
		check = new JButton(new createGeneralGraphInfoAction());
		builder.append(check);
		builder.nextLine();
		
		builder.appendSeparator();
		builder.appendSeparator("Element Properties");
		builder.nextLine();
		
		edgeField = new JTextField();
		edgeField.setText("A");
		edgeField.setEnabled(false);
		builder.append("Edge Label" + labelSuffix, edgeField);
		builder.nextLine();
		
		minSpinner = new JSpinner();
		minSpinner.setModel(new SpinnerListModel(new String[] {"0"}));
		minSpinner.addChangeListener(new ChangeListener() {
			
			@Override
			public void stateChanged(ChangeEvent e) {
				if(minSpinner.getValue().equals("m")){
					maxSpinner.setValue("m");
				}
				else if(maxSpinner.getValue().equals("m")){
					
				} else if(Integer.parseInt((String) minSpinner.getValue()) > Integer.parseInt((String) maxSpinner.getValue())){
					maxSpinner.setValue(minSpinner.getValue());
				}
			}
		});
		minSpinner.setEnabled(false);
		builder.append("Min Bound" + labelSuffix, minSpinner);
		builder.nextLine();
		
		maxSpinner = new JSpinner();
		maxSpinner.setModel(new SpinnerListModel(new String[] {"0"}));
		maxSpinner.addChangeListener(new ChangeListener() {
			
			@Override
			public void stateChanged(ChangeEvent e) {
				if(minSpinner.getValue().equals("m")){
					minSpinner.setValue(maxSpinner.getValue());
				}
				else if(maxSpinner.getValue().equals("m")){
					
				} else if(Integer.parseInt((String) minSpinner.getValue()) > Integer.parseInt((String) maxSpinner.getValue())){
					minSpinner.setValue(maxSpinner.getValue());
				}				
				
			}
		});
		maxSpinner.setEnabled(false);
		builder.append("Max Bound" + labelSuffix, maxSpinner);
		builder.nextLine();		

		return builder.getPanel();
	}


	@SuppressWarnings("static-access")
	private JScrollPane createGraphEditorPanel() {
      
        syncer = new AnnotatedTypeGraphEditSynchronizer(false, 1, this);
        syncer.setNewEdgeStyleName("edge-style");
        syncer.setNewNodeStyleName("node-style");
        syncer.setNewLabelStyleName(VisiTools.getInstance().getVisualizer().labelStyleName);
        syncer.setEnabled(false);
                
        Style edgeStyle = new Style();
        edgeStyle.setLineColor(new Color(.3f, .3f, .3f));
        edgeStyle.setLineColor(true, new Color(.6f, .6f, .6f));
        edgeStyle.setFrontArrow(ArrowHead.newTriangleArrowHead(8, 8));
        edgeStyle.setStroke(Style.getDefaultStroke(Style.LineWidth.MEDIUM, 
                Style.DashType.SOLID));
        
        Style nodeStyle = new Style();
        nodeStyle.setFilled(true);
        nodeStyle.setFillColor(false, Color.WHITE);
        nodeStyle.setFillColor(true, Color.LIGHT_GRAY);
        
        AnnotatedTypeGraphEditor editor = new AnnotatedTypeGraphEditor();
        VisiTools.getInstance().getVisualizer().addDefaultStyles(editor.getDrawer().getStyleMap());
        editor.setStyle("edge-style", edgeStyle);
        editor.setStyle("node-style", nodeStyle);
        editor.setGraph(VisiTools.getInstance().visualizeGraph(graph));
        editor.setEditSynchronizer(syncer);
        editor.setAntiAliased(true);   
		
        scroll = new JScrollPane(editor);
        scroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        scroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
        
        return scroll;
	}
    
    // ---------------------------------------------------------------------- //
    //                              Actions                                   //
    // ---------------------------------------------------------------------- //
    
	/**
     * Action which shows the graph creation dialog
     */
    
    private class createGeneralGraphInfoAction extends StandardAction
    {
        /** 
         * Creates a new CreateGraphAction instance.
         */
    	createGeneralGraphInfoAction()
        {
            super(ResourceKeys.CMD_USE_THESE_PROPERTIES);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
			check.setEnabled(false);
			nameField.setEnabled(false);
			mField.setEnabled(false);
			syncer.setEnabled(true);
			edgeField.setEnabled(true);
			minSpinner.setEnabled(true);
			maxSpinner.setEnabled(true);
			
			int maxValue = 2;
			String[] annotations;
			
			if(mField.getText().equals("")){
				mField.setText("2");
				annotations = new String[3];
			}
			else{
				maxValue = Integer.parseInt(mField.getText());
				annotations = new String[maxValue + 1];
			}
			
			for(int i=0; i<maxValue; i++){
				annotations[i] = "" + i;
			}
			annotations[maxValue] = "m";
			
			minSpinner.setModel(new SpinnerListModel(annotations));
			maxSpinner.setModel(new SpinnerListModel(annotations));
			maxSpinner.setValue(annotations[maxValue]);
		}
    
    }
    
	// ---------------------------------------------------------------------- //
    //                              Informations                              //
    // ---------------------------------------------------------------------- //

	public String getInputName()
	{
		return nameField.getText().replaceAll("[^a-zA-Z]", "");
	}

	public String getNodeLabel()
	{
		return "[" + minSpinner.getValue() + ";" + maxSpinner.getValue() + "]";
	}
	
	public String getEdgeLabel()
	{
		return edgeField.getText();
	}
	
	public String getMinValue()
	{
		return minSpinner.getValue() + "";
	}
	
	public String getMaxValue()
	{
		return maxSpinner.getValue() + "";
	}
	
	public int getLimit()
	{
		if(mField.getText().equals("")){
			return 2;
		}
		return Integer.parseInt(mField.getText());
	}
	

	// ---------------------------------------------------------------------- //
    //                              Private Methods                           //
    // ---------------------------------------------------------------------- //
	
	private NumberFormatter createNumberFormatter() {
		NumberFormat format = NumberFormat.getInstance();
	    NumberFormatter formatter = new NumberFormatter(format);
	    formatter.setValueClass(Integer.class);
	    formatter.setMinimum(1);
	    formatter.setMaximum(999);
	    formatter.setAllowsInvalid(false);
	    formatter.setCommitsOnValidEdit(true);
	    return formatter;
	}
	
}
