/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingUtilities;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.dragom.data.Postcondition;
import de.uni_due.inf.ti.dragom.general.ResourceKeys;
import de.uni_due.inf.ti.dragom.gui.AnnotatedTypeGraphVisualizer.GraphColor;
import de.uni_due.inf.ti.graph.Morphism;
import de.uni_due.inf.ti.graph.Rule;
import de.uni_due.inf.ti.swing.GuiContext;
import de.uni_due.inf.ti.swing.StandardAction;
import de.uni_due.inf.ti.swing.StandardDialog;
import de.uni_due.inf.ti.swing.UserInputException;
import de.uni_due.inf.ti.visigraph.VxGraph;
import de.uni_due.inf.ti.visigraph.swing.GraphPanel;

/**
 * Dialog which displays the switch rule dialog.
 * 
 * @author Dennis Nolte
 */
class InvariantCheckResultDialog extends StandardDialog<Integer>
{
	private static final long serialVersionUID = -4255861163073697037L;
	
	private List<Rule> rules;
	private List<Morphism> matches;
	private Rule currentRule;
	private Morphism currentMatch;
	private Morphism currentPostMatch;
	private Postcondition currentPostcondition;
	HashMap<Rule,HashMap<Morphism,Postcondition>> invariantList;
    private List<VxGraph> visisRule;
    private List<VxGraph> visisMatch;
    private List<VxGraph> visisPost;
    
    private JMenuItem simplifyBox;
    private JMenuItem anItem;
    
    private JScrollPane ruleScroll;
	private GraphPanel ruleView;

    private JScrollPane matchScroll;
	private GraphPanel matchView;
	
    private JScrollPane postScroll;
	private GraphPanel postView;
	
	private JComboBox<String> ruleCombo;
	private JComboBox<String> matchCombo;
	
    class PostMorphismPopUp extends JPopupMenu {

		private static final long serialVersionUID = -3487707252263996312L;

		public PostMorphismPopUp(){
        	anItem = new JMenuItem(new SwitchPostconditionMultiplicityAction());
        	anItem.setToolTipText("Displayed multiplicity: " + (currentPostcondition.getPostGraph().getActiveMultiplicityIndex()+1) + "/" + currentPostcondition.getPostGraph().getNumberOfMultiplicities() + 
        						  " (legal for abstract graph annotation " + (currentPostcondition.getAnnotationIndexes().get(currentPostcondition.getPostGraph().getActiveMultiplicityIndex())+1) + ")");
            if(currentPostcondition.getPostGraph().getNumberOfMultiplicities()<=1) anItem.setEnabled(false);
            add(anItem);
            add(simplifyBox);
        }
    }
	
    public InvariantCheckResultDialog(Frame owner,List<Rule> rules, HashMap<Rule,HashMap<Morphism,Postcondition>> invariantList)
    {
        super(owner, ResourceKeys.TIT_INVARIANT_RESULT, EnumSet.of(Button.CANCEL));
        setCancelKey(ResourceKeys.CMD_OKAY);

        visisRule = new ArrayList<>();
        visisMatch = new ArrayList<>();
        visisPost = new ArrayList<>();
        
        matches = new ArrayList<Morphism>();
        
        this.rules = rules;
        this.invariantList = invariantList;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected JComponent getDialogComponent()
    {
    	JPanel result = new JPanel();
    	result.setPreferredSize(new Dimension((int) (getOwner().getSize().getWidth()*0.85),(int) (getOwner().getSize().getHeight()*0.85))); //new Dimension(1000,680)
    	result.setLayout(new BorderLayout(0,1));
    	result.setOpaque(true);
    	
        simplifyBox = new JCheckBoxMenuItem(new ToggleSimplifyAction());
        simplifyBox.setSelected(false);

    	JPanel mainPanel = new JPanel();
    	mainPanel.setOpaque(false);
    	mainPanel.setLayout(new GridLayout(2, 1, 0, 0));
    	
    	JPanel postPanel = new JPanel();
        postPanel.setOpaque(false);
        postPanel = createPostPanel();
        
        JPanel topPanels = new JPanel();
        topPanels.setOpaque(false);
        topPanels.setLayout(new GridLayout(1, 2, 0, 0));
        
    	JPanel rulePanel = new JPanel();
    	rulePanel.setOpaque(false);
    	rulePanel.setLayout(new BorderLayout());
        rulePanel.add(createRulePanel(), BorderLayout.CENTER);
        rulePanel.add(createRuleCombo(), BorderLayout.NORTH);
        
        JPanel matchPanel = new JPanel();
    	matchPanel.setOpaque(false);
    	matchPanel.setLayout(new BorderLayout());
    	matchPanel.add(createMatchPanel(), BorderLayout.CENTER);
    	matchPanel.add(createMatchCombo(), BorderLayout.NORTH);
          	
    	topPanels.add(rulePanel);
    	topPanels.add(matchPanel);
    	
    	mainPanel.add(topPanels);
    	mainPanel.add(postPanel);
    	
    	JPanel margin = new JPanel();
    	margin.setPreferredSize(new Dimension(1,680));

    	result.add(margin, BorderLayout.WEST);
        result.add(mainPanel, BorderLayout.CENTER);
        result.setBorder(BorderFactory.createMatteBorder(1, 1, 1, 1, Color.BLACK));
        getContentPane().setBackground(new Color(0,0,0,0));
        getRootPane().setBackground(new Color(0,0,0,0));
        setUndecorated(true);
        setBackground(new Color(0,0,0,0));
        return result;
    }

    
    @Override
    protected Integer parseOptions() throws UserInputException
    {
        return ruleCombo.getSelectedIndex();
    }
	
	private JScrollPane createRulePanel() {
      
		ruleView = new GraphPanel();
        VisiTools.getInstance().getVisualizer().addDefaultStyles(ruleView.getDrawer().getStyleMap());
		
        ruleScroll = new JScrollPane(ruleView);
        ruleScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        ruleScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
        
        return ruleScroll;
	}
	
	private JScrollPane createMatchPanel() {
	      
		matchView = new GraphPanel();
        VisiTools.getInstance().getVisualizer().addDefaultStyles(matchView.getDrawer().getStyleMap());
		
        matchScroll = new JScrollPane(matchView);
        matchScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        matchScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
        
        return matchScroll;
	}
	
    /**
     * Creates the combo box displaying rule names.
     */
    private JPanel createRuleCombo()
    {
        DefaultComboBoxModel<String> model = new DefaultComboBoxModel<>(); 
        ruleCombo = new JComboBox<String>(model);
        setRule();
        ruleCombo.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e)
            {
                int index = ruleCombo.getSelectedIndex();
                if (index >= 0) changeRule(index);
                resetMatch();
            }

        });
        
        JLabel ruleLabel = GuiContext.createLabel(ResourceKeys.LAB_SELECT_RULE);        

        JPanel comboPanel = new JPanel();
        comboPanel.setLayout(new BoxLayout(comboPanel, BoxLayout.X_AXIS));
        comboPanel.add(Box.createHorizontalStrut(5));
        comboPanel.add(ruleLabel);
        comboPanel.add(Box.createHorizontalStrut(10));
        comboPanel.add(ruleCombo);
        
        return comboPanel;
    }
    
    /**
     * Creates the combo box displaying match names.
     */
    private JPanel createMatchCombo()
    { 
    	DefaultComboBoxModel<String> model = new DefaultComboBoxModel<>();
        matchCombo = new JComboBox<String>(model);
        resetMatch();
        matchCombo.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e)
            {
                int index = matchCombo.getSelectedIndex();
                if (index >= 0) changeMatch(index);
            }

        });
        
        JLabel matchLabel = GuiContext.createLabel(ResourceKeys.LAB_SELECT_MATCH);        

        JPanel comboMatchPanel = new JPanel();
        comboMatchPanel.setLayout(new BoxLayout(comboMatchPanel, BoxLayout.X_AXIS));
        comboMatchPanel.add(Box.createHorizontalStrut(5));
        comboMatchPanel.add(matchLabel);
        comboMatchPanel.add(Box.createHorizontalStrut(10));
        comboMatchPanel.add(matchCombo);
        
        return comboMatchPanel;
    }
    
    /**
     * Creates the postcondition view component.
     */
    private JPanel createPostPanel()
    {
    	postView = new GraphPanel();
    	VisiTools.getInstance().getVisualizer().addDefaultStyles(postView.getDrawer().getStyleMap());
    	postView.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
            	if(currentPostcondition != null) {
            		if (SwingUtilities.isRightMouseButton(e)){
            			doPop(e);
            		}        		
            	}
            }
            private void doPop(MouseEvent e){
                PostMorphismPopUp menu = new PostMorphismPopUp();
                menu.show(e.getComponent(), e.getX(), e.getY());
            }
        });
            
    	postScroll = new JScrollPane(postView);
    	postScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    	postScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
    	
    	JPanel postMorphismPanel = new JPanel();
    	postMorphismPanel.setLayout(new BorderLayout());
    	
    	JLabel postMorphismLabel = GuiContext.createLabel(ResourceKeys.LAB_POSTCONDITION_MORPHISM);
    	postMorphismLabel.setText(" " + postMorphismLabel.getText());
    	
    	postMorphismPanel.add(postMorphismLabel, BorderLayout.PAGE_START);
    	postMorphismPanel.add(postScroll, BorderLayout.CENTER);
    	
    	JPanel margin = new JPanel();
    	margin.setPreferredSize(new Dimension(1,680));
    	postMorphismPanel.add(margin, BorderLayout.WEST);
    	
    	return postMorphismPanel;
    }
    
    // ---------------------------------------------------------------------- //
    //                                  Action	                              //
    // ---------------------------------------------------------------------- //

    /**
     * Action which shows the add rule dialog
     */   
    private class ToggleSimplifyAction extends StandardAction
    {
        /** 
         * Creates a new AddRuleAction instance.
         */
    	ToggleSimplifyAction()
        {
            super(ResourceKeys.CMD_TOGGLE_SIMPLIFY);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
				
			postView.setGraph(VisiTools.getInstance().visualizeLegalMorphism(currentPostMatch, 
																			 currentPostcondition.getRightHandNodes(), 
																			 currentPostcondition.getRightHandEdges(), 
																			 GraphColor.YELLOW, 
																			 simplifyBox.isSelected()));
			postScroll.revalidate();
			
		}
    
    }
    
    /**
     * Action which shows the switch postcondition multiplicity dialog
     */   
    private class SwitchPostconditionMultiplicityAction extends StandardAction
    {
        /** 
         * Creates a new SwitchPostconditionMultiplicityAction instance.
         */
    	SwitchPostconditionMultiplicityAction()
        {
            super(ResourceKeys.CMD_SWITCH_MULTIPLICITY);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
            SwitchAnnotationDialog dlg = new SwitchAnnotationDialog(InvariantCheckResultDialog.this,currentPostcondition.getPostGraph(),
																            	   currentPostcondition.getRightHandNodes(),
																            	   currentPostcondition.getRightHandEdges(),
            																	   GraphColor.YELLOW,
            																	   simplifyBox.isSelected());
            dlg.initialize();
            AnnotatedTypeGraph result = dlg.runDialog();
            if(result != null) {
            	((AnnotatedTypeGraph)currentPostMatch.getDomain()).selectMultiplicity(result.getActiveMultiplicityIndex());
            	((AnnotatedTypeGraph)currentPostMatch.getCodomain()).selectMultiplicity(currentPostcondition.getAnnotationIndexes().get(result.getActiveMultiplicityIndex()));
            	postView.setGraph(VisiTools.getInstance().visualizeLegalMorphism(currentPostMatch, currentPostcondition.getRightHandNodes(), currentPostcondition.getRightHandEdges(), GraphColor.YELLOW, simplifyBox.isSelected()));
            	postScroll.revalidate();
            }
		}
    
    }
    
	// ---------------------------------------------------------------------- //
    //                              Private Methods                           //
    // ---------------------------------------------------------------------- //
		
    /**
     * Changes the rule which is displayed in the rule view.
     * 
     * @param index
     *      index of the rule to be displayed
     */
    private void changeRule(int index) 
    {
    	currentRule = rules.get(index);
    	VxGraph vx = visisRule.get(index);
    	if (vx == null) {
    		vx = VisiTools.getInstance().visualizeRuleDPO(currentRule);
    		visisRule.set(index, vx);
    	}
    	ruleView.setGraph(vx);
    }
    
    /**
     * Changes the match which is displayed in the match view.
     * 
     * @param index
     *      index of the match to be displayed
     */
    private void changeMatch(int index) 
    {
    	currentMatch = matches.get(index);
    	VxGraph vx = visisMatch.get(index);
    	if (vx == null) {
    		vx = VisiTools.getInstance().visualizeMatch(currentMatch, false, true);
    		visisMatch.set(index, vx);
    	}
    	matchView.setGraph(vx);
    	
    	currentPostcondition = invariantList.get(currentRule).get(currentMatch);
    	currentPostMatch = currentPostcondition.getLegalMorphism();
    	((AnnotatedTypeGraph)currentPostMatch.getDomain()).selectMultiplicity(0);
    	((AnnotatedTypeGraph)currentPostMatch.getCodomain()).selectMultiplicity(currentPostcondition.getAnnotationIndexes().get(0));
    	VxGraph vx2 = visisPost.get(index);
    	if (vx2 == null) {
    		vx2 = VisiTools.getInstance().visualizeLegalMorphism(currentPostMatch, currentPostcondition.getRightHandNodes(), currentPostcondition.getRightHandEdges(), GraphColor.YELLOW, simplifyBox.isSelected());
    		visisPost.set(index, vx2);
    	}
    	postView.setGraph(vx2);
    }
    
    /**
     * Changes the rule displayed.
     */
    private void setRule()
    {        
        ruleView.setEnabled(true);
        ruleCombo.setEnabled(true);

        for (int n=0; n < rules.size();n++) {
        	visisRule.add(null);
            ruleCombo.addItem(rules.get(n).getName());
        }

        ruleCombo.setSelectedIndex(0);
        changeRule(0);
    }
    
    /**
     * Changes the match displayed.
     */
    private void resetMatch()
    {  
        matchCombo.removeAllItems();
        matches.clear();
        visisMatch = new ArrayList<>();
        visisPost = new ArrayList<>();
         
        for(Morphism m: invariantList.get(currentRule).keySet()) {
        	matches.add(m);
        }
        
        for (int n=0; n < matches.size();n++) {
        	visisMatch.add(null);
        	visisPost.add(null);
            matchCombo.addItem("Match " + (n+1));       
        }
        
        matchCombo.setSelectedIndex(0);
        changeMatch(0);
    }
}
