/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.gui;

import java.awt.AlphaComposite;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Desktop;
import java.awt.Dialog.ModalityType;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.ResourceBundle;
import javax.swing.Action;
import javax.swing.ImageIcon;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.UIManager;
import javax.swing.WindowConstants;

import de.uni_due.inf.ti.annotation.NonNull;
import de.uni_due.inf.ti.annotation.Nullable;
import de.uni_due.inf.ti.dragom.algorithms.LanguageAlgorithm;
import de.uni_due.inf.ti.dragom.algorithms.MaterializationConstructionAlgorithm;
import de.uni_due.inf.ti.dragom.algorithms.PostconditionConstructionAlgorithm;
import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.dragom.data.AnnotationBounds;
import de.uni_due.inf.ti.dragom.data.Materialization;
import de.uni_due.inf.ti.dragom.data.Postcondition;
import de.uni_due.inf.ti.dragom.general.AnnotatedTypeGraphException;
import de.uni_due.inf.ti.dragom.general.ResourceKeys;
import de.uni_due.inf.ti.dragom.gui.AnnotatedTypeGraphVisualizer.GraphColor;
import de.uni_due.inf.ti.dragom.io.GTSChooser;
import de.uni_due.inf.ti.dragom.io.GXLReader;
import de.uni_due.inf.ti.dragom.io.GXLWriter;
import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.GraphElement;
import de.uni_due.inf.ti.graph.Morphism;
import de.uni_due.inf.ti.graph.Node;
import de.uni_due.inf.ti.graph.Rule;
import de.uni_due.inf.ti.graph.TransformationSystem;
import de.uni_due.inf.ti.graph.io.SgfWriter;
import de.uni_due.inf.ti.swing.GuiContext;
import de.uni_due.inf.ti.swing.PlatformAdapter;
import de.uni_due.inf.ti.swing.StandardAction;
import de.uni_due.inf.ti.visigraph.swing.GraphPanel;

/**
 * Main frame of the DrAGoM GUI application.
 * 
 * @author Dennis Nolte
 */
public class MainFrame extends JFrame
{   
    private static final long serialVersionUID = 8566911575336591595L;
    
    private static final int DEFAULT_FRAME_WIDTH        = 1200;
    private static final int DEFAULT_FRAME_HEIGHT       = 800;
    
    private AnnotatedTypeGraph graph;
    private Materialization matti;
    private Postcondition postcondition;
    private List<Rule> rules;
    private Rule currentRule;
    private int activeRuleIndex;
    private HashMap<Rule,HashMap<Morphism,Postcondition>> invariantList;
    
    private boolean nowHyper = false;
    private boolean nowMultiplicities = true;
    private boolean validInvariant;
    
    private JFileChooser fileChooser;

    // GUI Elements
    private JPanel mainPanel;
    private JPanel workPanel;
    private JPanel gtsPanel;
    private JPanel rulePanel;
    private JPanel typeGraphPanel;
    private JPanel materializationGraphPanel;
    private JPanel postGraphPanel;
    private JMenu algoMenu;
    private JMenuItem materializationItem;
    private JMenuItem postconditionItem;
    private JMenuItem invariantItem;
    private JMenuItem saveRule;
    private JMenuItem saveGraph;
    private JMenuItem savePostcondition;
    private JMenuItem simplifyBox;
    private GraphPanel graphView;
    private GraphPanel ruleView;
    private GraphPanel materializationView;
    private GraphPanel postView;
    private JScrollPane graphScroll;
    private JScrollPane ruleScroll;
    private JScrollPane matScroll;
    private JScrollPane postScroll;
   
    class GraphContextPopUp extends JPopupMenu {

		private static final long serialVersionUID = 3222969040184143051L;
		JMenuItem anItem;
		JMenuItem anItem2;
		JMenuItem anItem3;
        public GraphContextPopUp(){
        	anItem = new JMenuItem(new AddMultiplicityAction()); 
        	anItem.setToolTipText("Number of multiplicities: " + graph.getNumberOfMultiplicities());
        	add(anItem);
        	anItem2 = new JMenuItem(new RemoveMultiplicityAction());
            if(graph.getNumberOfMultiplicities()<=1) anItem2.setEnabled(false);
            anItem2.setToolTipText("Displayed multiplicity: " + (graph.getActiveMultiplicityIndex()+1) + "/" + graph.getNumberOfMultiplicities());
            add(anItem2);
            anItem3 = new JMenuItem(new SwitchGraphMultiplicityAction());
            if(graph.getNumberOfMultiplicities()<=1) anItem3.setEnabled(false);
            anItem3.setToolTipText("Displayed multiplicity: " + (graph.getActiveMultiplicityIndex()+1) + "/" + graph.getNumberOfMultiplicities());
            add(anItem3);
        }
    }
    
    class MatContextPopUp extends JPopupMenu {

		private static final long serialVersionUID = 4707941529073715884L;
		JMenuItem anItem;
        public MatContextPopUp(){
        	anItem = new JMenuItem(new SwitchMaterializationMultiplicityAction());
        	anItem.setToolTipText("Displayed multiplicity: " + (matti.getMaterialization().getActiveMultiplicityIndex()+1) + "/" + matti.getMaterialization().getNumberOfMultiplicities());
            if(matti.getMaterialization().getNumberOfMultiplicities()<=1) anItem.setEnabled(false);
            add(anItem);      
            add(simplifyBox);
        }
    }
    
    class PostContextPopUp extends JPopupMenu {

		private static final long serialVersionUID = -8423505801356590261L;
		JMenuItem anItem;
        public PostContextPopUp(){
        	anItem = new JMenuItem(new SwitchPostconditionMultiplicityAction());
        	anItem.setToolTipText("Displayed multiplicity: " + (postcondition.getPostGraph().getActiveMultiplicityIndex()+1) + "/" + postcondition.getPostGraph().getNumberOfMultiplicities());
            if(postcondition.getPostGraph().getNumberOfMultiplicities()<=1) anItem.setEnabled(false);
            add(anItem);
            add(new JMenuItem(new CheckInclusionAction()));
            add(simplifyBox);
        }
    }
    
    class RuleContextPopUp extends JPopupMenu {

		private static final long serialVersionUID = 4324934756196344715L;
		JMenuItem anItem;
		JMenuItem anItem2;
		JMenuItem anItem3;
		public RuleContextPopUp(){
        	anItem = new JMenuItem(new AddRuleAction());
        	anItem.setToolTipText("Number of rules: " + rules.size());
            add(anItem);
            anItem2 = new JMenuItem(new RemoveRuleAction());
            if(rules.size()<=1) anItem2.setEnabled(false);
            anItem2.setToolTipText("Displayed rule: " + (activeRuleIndex+1) + "/" + rules.size());
            add(anItem2);
            anItem3 = new JMenuItem(new SwitchRuleAction());
            if(rules.size()<=1) anItem3.setEnabled(false);
            anItem3.setToolTipText("Displayed rule: " + (activeRuleIndex+1) + "/" + rules.size());
            add(anItem3);
        }
    }

    /**
     * Creates a new MainFrame instance.
     * 
     * @param system
     *      graph transformation system shown in this main frame
     * @param graph
     *      annotated type graph shown in this main frame
     * @param hyperGraphs
     *      boolean if all graphs are displayed as hypergraphs
     */
    public MainFrame(
            @Nullable TransformationSystem system,@Nullable AnnotatedTypeGraph graph, 
            boolean hyperGraphs) 
    {
    	
        this.graph = graph;
        nowHyper = hyperGraphs;
        
        if (system != null) {
            for (Rule rule : system.getRules()) {
                nowHyper |= !rule.getLeft().isBinary();
                nowHyper |= !rule.getRight().isBinary();
            }
        }
        
        VisiTools.getInstance().setHyperStyle(nowHyper);
        VisiTools.getInstance().setMultiplicityStyle(nowMultiplicities);
        
        rules = new ArrayList<>();
        
        if(System.getProperty("os.name").toLowerCase().contains("windows")) {
        	setTitle("r A G o M");
        }
        else {
        	setTitle("DrAGoM");
        }
    }

    /**
     * Creates the controls on this frame and shows this frame.
     */
    @SuppressWarnings("static-access")
	public void createAndShowGui()
    {
    	UIManager.put("OptionPane.background", Color.WHITE);
    	UIManager.getLookAndFeelDefaults().put("Panel.background", Color.WHITE);
    	UIManager.getLookAndFeelDefaults().put("PopupMenu.background", Color.WHITE);
    	UIManager.getLookAndFeelDefaults().put("CheckBoxMenuItem.background", Color.WHITE);
    	UIManager.getLookAndFeelDefaults().put("MenuItem.background", Color.WHITE);
    	UIManager.getLookAndFeelDefaults().put("ComboBox.background", Color.WHITE); 	
    	UIManager.getLookAndFeelDefaults().put("ScrollPane.background", Color.WHITE);
    	UIManager.getLookAndFeelDefaults().put("Label.disabledForeground", Color.BLACK);
    	UIManager.getLookAndFeelDefaults().put("Label.disabledText", Color.BLACK);
    	
    	mainPanel = new JPanel();
    	mainPanel.setOpaque(false);
    	mainPanel.setLayout(new GridLayout(2, 1, 0, 0));
    	
    	gtsPanel = new JPanel();
    	gtsPanel.setOpaque(false);
    	gtsPanel.setLayout(new BorderLayout());
    	
        workPanel = new JPanel();
        workPanel.setOpaque(false);
        workPanel.setLayout(new GridLayout(1, 3, 0, 0));

        mainPanel.add(gtsPanel);
        mainPanel.add(workPanel);
		
        URL bimageURL = MainFrame.class.getResource("/de/uni_due/inf/ti/dragom/resources/logo.png");
        ImageIcon bicon = new ImageIcon(bimageURL, "DrAGoM, directed graph abstraction over multiplicities");
			// Create output text area and install default printer
        JPanel	background = new JPanel() {
	
				private static final long serialVersionUID = 7066386582585853861L;
				
				private final Image backgroundImage;
	
				{
					backgroundImage = new BufferedImage(bicon.getIconWidth(), 
							bicon.getIconHeight(), BufferedImage.TYPE_INT_RGB);
					Graphics2D g2d = (Graphics2D) backgroundImage.getGraphics();
					g2d.setBackground(Color.WHITE);
					g2d.clearRect(0, 0, bicon.getIconWidth(), 
							bicon.getIconHeight());
					AlphaComposite alpha = AlphaComposite.getInstance(
							AlphaComposite.SRC_OVER, 0.3f);
					g2d.setComposite(alpha);
					g2d.drawImage(bicon.getImage(), 0, 0, null);
				}
	
				@Override
				public void paintComponent(Graphics g) {
					
					super.paintComponent(g);
					
					Rectangle rect = getVisibleRect();
					
					int x = rect.x + (rect.width-backgroundImage.getWidth(this));
					int y = rect.y + (rect.height-backgroundImage.getHeight(this));
					g.drawImage(backgroundImage, x, y, this);
					setOpaque(false);
	
				}
			};
        
		background.setLayout(new BorderLayout());
		background.setBackground(Color.WHITE);
		background.add(mainPanel, BorderLayout.CENTER);
		
        getContentPane().add(background, BorderLayout.CENTER);
        getContentPane().setBackground(Color.WHITE);
           	
        createRuleView();
        createGraphView();
        createMaterializationView();
        createPostconditionView();
        
        JMenuBar bar = new JMenuBar();
        
        simplifyBox = new JCheckBoxMenuItem(new ToggleSimplifyAction());
        simplifyBox.setSelected(false);
               
        bar.add(createSystemMenu());
        bar.add(createAlgorithmMenu());
        bar.add(createHelpMenu());
        setJMenuBar(bar);
               
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        setBounds(new Rectangle(
                    (screenSize.width - DEFAULT_FRAME_WIDTH) / 2,
                    (screenSize.height - DEFAULT_FRAME_HEIGHT) / 2,
                    DEFAULT_FRAME_WIDTH,
                    DEFAULT_FRAME_HEIGHT));
        
        this.setExtendedState(this.MAXIMIZED_BOTH);
        
        addComponentListener(new ComponentAdapter() {
            public void updateData() 
            {
                //PersistentData.setWindowBounds(PersistentData.MAIN_WINDOW, getBounds());
            }
            @Override public void componentMoved(ComponentEvent e) 
            {
                updateData();
            }
            @Override public void componentResized(ComponentEvent e) 
            {
                updateData();
            }
        });
        
        addWindowListener(new WindowAdapter() {
            @Override public void windowClosing(WindowEvent e) {
            	if (exit()) {
                    System.exit(0);
                }
            }
        });
        
        setVisible(true);
        URL imageURL = MainFrame.class.getResource("/de/uni_due/inf/ti/dragom/resources/icon.png");
        ImageIcon icon = new ImageIcon(imageURL, "DrAGoM, directed graph abstraction over multiplicities");
        setIconImage(icon.getImage());
        
        setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        
        PlatformAdapter pa = PlatformAdapter.getInstance();
        pa.setQuitHandler(() -> exit());
        
        try { 
        	System.loadLibrary("z3java");
        }
        catch (UnsatisfiedLinkError ex) {
        	try {
        		System.loadLibrary("libz3java");
        	}
        	catch (UnsatisfiedLinkError ex2) {
        		showInformation("Could not find libz3java in java.library.path.\n\n" +
                        "(z3java) " + ex.getMessage() + "\n" +
                        "(libz3java) " + ex2.getMessage() + "\n" +
                        "\nPlease check the website for information.\n(All algorithms are disabled for now)", JOptionPane.ERROR_MESSAGE);
        		algoMenu.setEnabled(false);
        	}
        }
        
    }

    /**
     * Creates the System menu.
     */
    private @NonNull JMenu createSystemMenu()
    {
        JMenu systemMenu = GuiContext.createMenu(ResourceKeys.MENU_SYSTEM);
       
        JMenu newMenu = GuiContext.createMenu(ResourceKeys.CMD_NEW);
        newMenu.add(new JMenuItem(new CreateGTSAction()));
        newMenu.add(new JMenuItem(new CreateTypeGraphAction()));
        systemMenu.add(newMenu);

        systemMenu.addSeparator();
        JMenu loadMenu = GuiContext.createMenu(ResourceKeys.CMD_LOAD);
        loadMenu.add(new JMenuItem(new LoadGTSAction()));
        loadMenu.add(new JMenuItem(new LoadGraphAction()));
        systemMenu.add(loadMenu);
        
        savePostcondition = new JMenuItem(new SavePostconditionAction());
        savePostcondition.setEnabled(false);
        saveRule = new JMenuItem(new SaveGTSAction());
        saveRule.setEnabled(false);
        saveGraph = new JMenuItem(new SaveGraphAction());
        saveGraph.setEnabled(false);
        JMenu saveMenu = GuiContext.createMenu(ResourceKeys.CMD_SAVE);
        saveMenu.add(saveRule);
        saveMenu.add(saveGraph);
        saveMenu.add(savePostcondition);
        
        systemMenu.add(saveMenu);
        
//        systemMenu.add(new JMenuItem(new PrintLanguageAction()));
        if (PlatformAdapter.getInstance().menuContainsQuit()) {        
        	systemMenu.addSeparator();
        	systemMenu.add(new JMenuItem(new ExitAction()));
        }
        
        return systemMenu;
    }
   
    /**
     * Creates the System menu.
     */
    private @NonNull JMenu createAlgorithmMenu()
    {
        algoMenu = GuiContext.createMenu(ResourceKeys.MENU_ALGORITHM);       

        JMenu computeMenu = GuiContext.createMenu(ResourceKeys.CMD_COMPUTE);
        materializationItem = new JMenuItem(new MaterializationAction());
        computeMenu.add(materializationItem);
        materializationItem.setEnabled(false);
        
        postconditionItem = new JMenuItem(new PostconditionAction());
        computeMenu.add(postconditionItem);
        postconditionItem.setEnabled(false);
        algoMenu.add(computeMenu);

        algoMenu.addSeparator();
        
        invariantItem = new JMenuItem(new InvariantCheckAction());
        invariantItem.setEnabled(false);
        algoMenu.add(invariantItem);
        
//        systemMenu.add(new JMenuItem(new PrintLanguageAction()));
        
        return algoMenu;
    }
    
    /**
     * Creates the Help menu.
     */
    private @NonNull JMenu createHelpMenu()
    {
        JMenu helpMenu = GuiContext.createMenu(ResourceKeys.MENU_HELP);
        
        try {
            URI websiteURI = new URI(ResourceBundle.getBundle("de.uni_due.inf.ti.dragom.resources.parameters").getString("website"));
            String key = ResourceKeys.CMD_WEBSITE;
            helpMenu.add(new JMenuItem(new WebsiteAction(key, websiteURI)));
        }
        catch (URISyntaxException e) {
            // if an error occurs, just do nothing
        }
        
        Action aboutAction = new AboutAction();
        PlatformAdapter pa = PlatformAdapter.getInstance();
        if (pa.menuContainsAboutBox()) {
            helpMenu.addSeparator();
        	helpMenu.add(new JMenuItem(aboutAction));
        }
        pa.setAboutHandler(aboutAction);
        return helpMenu;
    }
    
    /**
     * Prepares to exit the application, and returns whether or not the application
     * can exit.
     * <p> 
     * Currently, this method does nothing and always returns 
     * {@code true}. In the future, the method may check for unsaved changed or
     * ask for user confirmation.
     */
    private boolean exit()
    {
        return true;
    }
    
    /**
     * Creates the graph view component.
     */
    private void createGraphView()
    {
    	graphView = new GraphPanel();
    	VisiTools.getInstance().getVisualizer().addDefaultStyles(graphView.getDrawer().getStyleMap());
    	graphView.addMouseListener(new MouseAdapter() {
        @Override
        public void mouseClicked(MouseEvent e) {
        	if(graph != null) {
        		if (SwingUtilities.isRightMouseButton(e)){
        			 doPop(e);
        		}
        		if (SwingUtilities.isLeftMouseButton(e)) {
        			if(e.getClickCount()==2){
        				MaximizeDialog dlg = new MaximizeDialog(MainFrame.this,VisiTools.getInstance().visualizeGraph(graph));
        				dlg.initialize();
        				dlg.runDialog();
        			}
        		}        		
        	}
        }
        
        private void doPop(MouseEvent e){
            GraphContextPopUp menu = new GraphContextPopUp();
            menu.show(e.getComponent(), e.getX(), e.getY());
        }
    	});
    	
    	graphScroll = new JScrollPane(graphView);
    	graphScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    	graphScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
    	
    	typeGraphPanel = new JPanel();
    	typeGraphPanel.setLayout(new BorderLayout());
    	
    	JLabel typeGraphLabel = GuiContext.createLabel(ResourceKeys.LAB_ANNOTATED_TYPE_GRAPH);
    	typeGraphLabel.setText(" " + typeGraphLabel.getText());
    	
    	typeGraphPanel.add(typeGraphLabel, BorderLayout.PAGE_START);
    	typeGraphPanel.add(graphScroll, BorderLayout.CENTER);
    	
    	JPanel margin = new JPanel();
    	margin.setPreferredSize(new Dimension(1,DEFAULT_FRAME_HEIGHT/2));
    	typeGraphPanel.add(margin, BorderLayout.WEST);
    	typeGraphPanel.setVisible(false);
    	
    	workPanel.add(typeGraphPanel, BorderLayout.CENTER);
    }
    
    /**
     * Creates the rule view component.
     */
    private void createRuleView()
    {
        ruleView = new GraphPanel();
        VisiTools.getInstance().getVisualizer().addDefaultStyles(ruleView.getDrawer().getStyleMap());
    	ruleView.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
            	if(currentRule != null) {
            		if (SwingUtilities.isRightMouseButton(e)){
            			 doPop(e);
            		}        		
            	}
        		if (SwingUtilities.isLeftMouseButton(e)) {
        			if(e.getClickCount()==2){
        				MaximizeDialog dlg = new MaximizeDialog(MainFrame.this,VisiTools.getInstance().visualizeRuleDPO(currentRule));
        				dlg.initialize();
        				dlg.runDialog();
        			}
        		}  
            }
            
            private void doPop(MouseEvent e){
                RuleContextPopUp menu = new RuleContextPopUp();
                menu.show(e.getComponent(), e.getX(), e.getY());
            }
        });
        
        ruleScroll = new JScrollPane(ruleView);
        ruleScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        ruleScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
       
        rulePanel = new JPanel();
		rulePanel.setLayout(new BorderLayout());
		
    	JLabel ruleLabel = GuiContext.createLabel(ResourceKeys.LAB_CURRENT_RULE);
    	ruleLabel.setText(" " + ruleLabel.getText());
    	
    	rulePanel.add(ruleLabel, BorderLayout.PAGE_START);
		rulePanel.add(ruleScroll, BorderLayout.CENTER);
		
    	JPanel margin = new JPanel();
    	margin.setPreferredSize(new Dimension(1,DEFAULT_FRAME_HEIGHT/2));
    	rulePanel.add(margin, BorderLayout.WEST);
    	rulePanel.setVisible(false);
    	
        gtsPanel.add(rulePanel, BorderLayout.CENTER);
    }
    
    /**
     * Creates the materialization view component.
     */
    private void createMaterializationView()
    {
    	materializationView = new GraphPanel();
    	VisiTools.getInstance().getVisualizer().addDefaultStyles(materializationView.getDrawer().getStyleMap());
    	materializationView.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
            	if(matti != null) {
            		if (SwingUtilities.isRightMouseButton(e)){
            			doPop(e);
            		}
            		if (SwingUtilities.isLeftMouseButton(e)) {
            			if(e.getClickCount()==2){
            				MaximizeDialog dlg = new MaximizeDialog(MainFrame.this,VisiTools.getInstance().visualizeColoredGraph(matti.getMaterialization(),
																														  matti.getLeftEmbeddingNodeImage(),
																														  matti.getLeftEmbeddingEdgeImage(),
            																											  GraphColor.BLUE,
            																											  simplifyBox.isSelected()));
            				dlg.initialize();
            				dlg.runDialog();
            			}
            		}        		
            	}
            }
            private void doPop(MouseEvent e){
                MatContextPopUp menu = new MatContextPopUp();
                menu.show(e.getComponent(), e.getX(), e.getY());
            }
        });
            
    	matScroll = new JScrollPane(materializationView);
    	matScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    	matScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
    	
    	materializationGraphPanel = new JPanel();
    	materializationGraphPanel.setLayout(new BorderLayout());
    	
    	JLabel materializationGraphLabel = GuiContext.createLabel(ResourceKeys.LAB_MATERIALIZATION_GRAPH);
    	materializationGraphLabel.setText(" " + materializationGraphLabel.getText());
    	
    	materializationGraphPanel.add(materializationGraphLabel, BorderLayout.PAGE_START);
    	materializationGraphPanel.add(matScroll, BorderLayout.CENTER);
    	
    	JPanel margin = new JPanel();
    	margin.setPreferredSize(new Dimension(1,DEFAULT_FRAME_HEIGHT));
    	materializationGraphPanel.add(margin, BorderLayout.WEST);
    	
    	materializationGraphPanel.setVisible(false);
    	
    	workPanel.add(materializationGraphPanel, BorderLayout.CENTER);
    }  
    
    /**
     * Creates the postcondition view component.
     */
    private void createPostconditionView()
    {
    	postView = new GraphPanel();
    	VisiTools.getInstance().getVisualizer().addDefaultStyles(postView.getDrawer().getStyleMap());
    	postView.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
            	if(postcondition != null) {
            		if (SwingUtilities.isRightMouseButton(e)){
            			doPop(e);
            		}
            		if (SwingUtilities.isLeftMouseButton(e)) { 
            			if(e.getClickCount()==2){
            				MaximizeDialog dlg = new MaximizeDialog(MainFrame.this,VisiTools.getInstance().visualizeColoredGraph(postcondition.getPostGraph(),
            																											  postcondition.getRightHandNodes(),
            																											  postcondition.getRightHandEdges(),
            																											  GraphColor.YELLOW,
            																											  simplifyBox.isSelected()));
            				dlg.initialize();
            				dlg.runDialog();
            			}
            		}        		
            	}
            }
            private void doPop(MouseEvent e){
                PostContextPopUp menu = new PostContextPopUp();
                menu.show(e.getComponent(), e.getX(), e.getY());
            }
        });
            
    	postScroll = new JScrollPane(postView);
    	postScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    	postScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
    	
    	postGraphPanel = new JPanel();
    	postGraphPanel.setLayout(new BorderLayout());
    	
    	JLabel postGraphLabel = GuiContext.createLabel(ResourceKeys.LAB_POSTCONDITION_GRAPH);
    	postGraphLabel.setText(" " + postGraphLabel.getText());
    	
    	postGraphPanel.add(postGraphLabel, BorderLayout.PAGE_START);
    	postGraphPanel.add(postScroll, BorderLayout.CENTER);
    	
    	JPanel margin = new JPanel();
    	margin.setPreferredSize(new Dimension(1,DEFAULT_FRAME_HEIGHT));
    	postGraphPanel.add(margin, BorderLayout.WEST);
    	
    	postGraphPanel.setVisible(false);
    	
    	workPanel.add(postGraphPanel, BorderLayout.CENTER);
    }
    
    
    // ---------------------------------------------------------------------- //
    //                              Actions                                   //
    // ---------------------------------------------------------------------- //
    
    /**
     * Action which shows the graph creation dialog
     */   
    private class CreateTypeGraphAction extends StandardAction
    {
        /** 
         * Creates a new CreateGraphAction instance.
         */
    	CreateTypeGraphAction()
        {
            super(ResourceKeys.CMD_TYPE_GRAPH);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
            CreateTypeGraphDialog dlg = new CreateTypeGraphDialog(MainFrame.this);
            dlg.initialize();
            AnnotatedTypeGraph result = dlg.runDialog();
            if(result != null) {
            	graph = result.clone();
            	graphView.setGraph(VisiTools.getInstance().visualizeGraph(graph));
            	typeGraphPanel.setVisible(true);
            	saveGraph.setEnabled(true);
            	graphScroll.revalidate();
            	resetMatAndPostPanels();
            }
		} 
    }

    /**
     * Action which shows the rule creation dialog
     */   
    private class CreateGTSAction extends StandardAction
    {
        /** 
         * Creates a new CreateGTSAction instance.
         */
    	CreateGTSAction()
        {
            super(ResourceKeys.CMD_GTS_RULE);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
            CreateRuleDialog dlg = new CreateRuleDialog(MainFrame.this);
            dlg.initialize();
            Rule result = dlg.runDialog();
            if(result != null) {
            	currentRule = new Rule(result.getCorrespondence());
            	rules.clear();
            	rules.add(currentRule);
            	activeRuleIndex = 0;
            	ruleView.setGraph(VisiTools.getInstance().visualizeRuleDPO(currentRule)); 
            	rulePanel.setVisible(true);
            	saveRule.setEnabled(true);
            	ruleScroll.revalidate();
            	resetMatAndPostPanels();
            }
		}
    
    }
 
    /**
     * Action which shows the add multiplicity dialog
     */   
    private class AddMultiplicityAction extends StandardAction
    {
        /** 
         * Creates a new AddMultiplicityAction instance.
         */
    	AddMultiplicityAction()
        {
            super(ResourceKeys.CMD_ADD_MULTIPLICITY);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
            AddMultiplicityDialog dlg = new AddMultiplicityDialog(MainFrame.this,graph);
            dlg.initialize();
            HashMap<GraphElement, AnnotationBounds> result = dlg.runDialog();
            if(result != null) {
            	if(graph.addMultiplicity(result)) {
            		graph.selectMultiplicity(graph.getNumberOfMultiplicities()-1);
            		graphView.setGraph(VisiTools.getInstance().visualizeGraph(graph)); 
            		graphScroll.revalidate();
            		resetMatAndPostPanels();            		
            	}
            	else {
            		showInformation("The created multiplicity is already covered.\nNo new multiplicity added.", 
            						JOptionPane.WARNING_MESSAGE);
            	}
            	
            }
		} 
    }
    
    /**
     * Action which removes the active multiplicity
     */   
    private class RemoveMultiplicityAction extends StandardAction
    {
        /** 
         * Creates a new RemoveMultiplicityAction instance.
         */
    	RemoveMultiplicityAction()
        {
            super(ResourceKeys.CMD_REMOVE_MULTIPLICITY);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
            
			if(showConfirmation("Would you like to delete the currently shown multiplicity?", JOptionPane.QUESTION_MESSAGE)){
				graph.removeMultiplicity(graph.getActiveMultiplicityIndex());
				graphView.setGraph(VisiTools.getInstance().visualizeGraph(graph)); 
				graphScroll.revalidate();
				resetMatAndPostPanels();				
			}
		} 
    }
    
    /**
     * Action which shows the switch graph multiplicity dialog
     */   
    private class SwitchGraphMultiplicityAction extends StandardAction
    {
        /** 
         * Creates a new SwitchGraphMultiplicityAction instance.
         */
    	SwitchGraphMultiplicityAction()
        {
            super(ResourceKeys.CMD_SWITCH_MULTIPLICITY);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
            SwitchAnnotationDialog dlg = new SwitchAnnotationDialog(MainFrame.this,graph,null,null,null,false);
            dlg.initialize();
            AnnotatedTypeGraph result = dlg.runDialog();
            if(result != null) {
            	graph.selectMultiplicity(result.getActiveMultiplicityIndex());
            	graphView.setGraph(VisiTools.getInstance().visualizeGraph(graph)); 
            	graphScroll.revalidate();
            }
		}  
    }
    
    /**
     * Action which shows the switch materialization multiplicity dialog
     */   
    private class SwitchMaterializationMultiplicityAction extends StandardAction
    {
        /** 
         * Creates a new SwitchMaterializationMultiplicityAction instance.
         */
    	SwitchMaterializationMultiplicityAction()
        {
            super(ResourceKeys.CMD_SWITCH_MULTIPLICITY);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
            SwitchAnnotationDialog dlg = new SwitchAnnotationDialog(MainFrame.this,matti.getMaterialization(),
            																	   matti.getLeftEmbeddingNodeImage(),
            																	   matti.getLeftEmbeddingEdgeImage(),
            																	   GraphColor.BLUE,
            																	   simplifyBox.isSelected());
            dlg.initialize();
            AnnotatedTypeGraph result = dlg.runDialog();
            if(result != null) {
            	matti.getMaterialization().selectMultiplicity(result.getActiveMultiplicityIndex());
    			updateMaterializationView();
    			
    			if(postGraphPanel.isVisible()) {
    				postcondition.getPostGraph().selectMultiplicity(result.getActiveMultiplicityIndex());
    				updatePostconditionView();
    			}
            }
		}
    
    }
    
    /**
     * Action which shows the switch postcondition multiplicity dialog
     */   
    private class SwitchPostconditionMultiplicityAction extends StandardAction
    {
        /** 
         * Creates a new SwitchPostconditionMultiplicityAction instance.
         */
    	SwitchPostconditionMultiplicityAction()
        {
            super(ResourceKeys.CMD_SWITCH_MULTIPLICITY);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
            SwitchAnnotationDialog dlg = new SwitchAnnotationDialog(MainFrame.this,postcondition.getPostGraph(),
            																	   postcondition.getRightHandNodes(),
            																	   postcondition.getRightHandEdges(),
            																	   GraphColor.YELLOW,
            																	   simplifyBox.isSelected());
            dlg.initialize();
            AnnotatedTypeGraph result = dlg.runDialog();
            if(result != null) {
            	postcondition.getPostGraph().selectMultiplicity(result.getActiveMultiplicityIndex());
    			updatePostconditionView();
    			
    			matti.getMaterialization().selectMultiplicity(result.getActiveMultiplicityIndex());
    			updateMaterializationView();
            }
		}
    
    }
    
    
    /**
     * Action which checks for the language inclusion
     */   
    private class CheckInclusionAction extends StandardAction
    {
        /** 
         * Creates a new CheckInclusionAction instance.
         */
    	CheckInclusionAction()
        {
            super(ResourceKeys.CMD_CHECK_INCLUSION);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
            
			Morphism legalMorphism = checkInclusion(postcondition.getPostGraph(),graph);
			if(legalMorphism!=null){
				showInformation("The language of the abstract graph \nis an invariant for this match of rule " + (activeRuleIndex+1) + ".", JOptionPane.INFORMATION_MESSAGE);
				
				InclusionMorphismDialog dlg = new InclusionMorphismDialog(MainFrame.this,VisiTools.getInstance().visualizeLegalMorphism(legalMorphism, postcondition.getRightHandNodes(), postcondition.getRightHandEdges(), GraphColor.YELLOW, simplifyBox.isSelected()));
				dlg.initialize();
				dlg.runDialog();
				postScroll.revalidate();
			}
			else {
				showInformation("No legal morphism from the postcondition \nto the abstract graph was found for this match of rule " + (activeRuleIndex+1) + ".\nCan't prove invariant for the given system.", JOptionPane.ERROR_MESSAGE);
			}
				
		}
    
    }

    /**
     * Action which shows the add rule dialog
     */   
    private class ToggleSimplifyAction extends StandardAction
    {
        /** 
         * Creates a new AddRuleAction instance.
         */
    	ToggleSimplifyAction()
        {
            super(ResourceKeys.CMD_TOGGLE_SIMPLIFY);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
			
			if(materializationGraphPanel.isVisible()) {
				materializationView.setGraph(VisiTools.getInstance().visualizeColoredGraph(matti.getMaterialization(),
						matti.getLeftEmbeddingNodeImage(),
						matti.getLeftEmbeddingEdgeImage(),
						GraphColor.BLUE,
						simplifyBox.isSelected()));
				matScroll.revalidate();
			}
			if(postGraphPanel.isVisible()) {
				postView.setGraph(VisiTools.getInstance().visualizeColoredGraph(postcondition.getPostGraph(),
						 postcondition.getRightHandNodes(),
						 postcondition.getRightHandEdges(),
						 GraphColor.YELLOW,
						 simplifyBox.isSelected()));
				postScroll.revalidate();
			}
		}
    
    }

    /**
     * Action which prints the graph language to an HTML file
     */
	@SuppressWarnings("unused")
	private class PrintLanguageAction extends StandardAction
    {
        /**
         * Creates a new PrintLanguageAction instance.
         */
    	PrintLanguageAction() 
        {
            super(ResourceKeys.CMD_PRINT_LANGUAGE);
        }
        
        /**
         * Called when this action is executed.
         */
        @Override
        public void actionPerformed(@NonNull ActionEvent e)
        {
        	SwingWorker<Void, Void> mySwingWorker = new SwingWorker<Void, Void>(){
            @Override
            protected Void doInBackground() throws Exception {

             		LanguageAlgorithm algo = new LanguageAlgorithm(graph);
                	algo.computeLanguage();
                	return null;
               }
            };

             final JDialog dialog = new JDialog((Frame)null, "LanguagePrinter", ModalityType.APPLICATION_MODAL);

             mySwingWorker.addPropertyChangeListener(new PropertyChangeListener() {

                @Override
                public void propertyChange(PropertyChangeEvent evt) {
                   if (evt.getPropertyName().equals("state")) {
                      if (evt.getNewValue() == SwingWorker.StateValue.DONE) {
                         dialog.dispose();
                      }
                   }
                }
             });
             mySwingWorker.execute();

             JProgressBar progressBar = new JProgressBar();
             progressBar.setIndeterminate(true);
             progressBar.setOpaque(false);
             JPanel panel = new JPanel(new BorderLayout());
             JLabel text = new JLabel("Printing the language. Please wait...");
             text.setOpaque(false);
             panel.add(progressBar, BorderLayout.CENTER);
             panel.add(text, BorderLayout.PAGE_START);
             panel.setBackground(Color.WHITE);
             dialog.setUndecorated(true);
             dialog.add(panel);
             dialog.pack();
             dialog.setLocationRelativeTo(null);
             dialog.setVisible(true);
        }
    }
	
    /**
     * Action which shows the add rule dialog
     */   
    private class AddRuleAction extends StandardAction
    {
        /** 
         * Creates a new AddRuleAction instance.
         */
    	AddRuleAction()
        {
            super(ResourceKeys.CMD_ADD_RULE);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
			CreateRuleDialog dlg = new CreateRuleDialog(MainFrame.this);
            dlg.initialize();
            Rule result = dlg.runDialog();
            if(result != null) {
            	currentRule = new Rule(result.getCorrespondence());
            	rules.add(currentRule);
            	activeRuleIndex = rules.size()-1;
            	resetMatAndPostPanels();
            	ruleView.setGraph(VisiTools.getInstance().visualizeRuleDPO(currentRule)); 
            	ruleScroll.revalidate();
            }
		}
    
    }
    
    /**
     * Action which removes the active rule
     */   
    private class RemoveRuleAction extends StandardAction
    {
        /** 
         * Creates a new RemoveRuleAction instance.
         */
    	RemoveRuleAction()
        {
            super(ResourceKeys.CMD_REMOVE_RULE);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
            
			if(showConfirmation("Would you like to delete the currently shown rule?", JOptionPane.QUESTION_MESSAGE)){
				rules.remove(activeRuleIndex);
				activeRuleIndex = 0;
				currentRule = rules.get(activeRuleIndex);
				ruleView.setGraph(VisiTools.getInstance().visualizeRuleDPO(currentRule)); 
            	ruleScroll.revalidate();
	    		resetMatAndPostPanels();
			}
		} 
    }

    /**
     * Action which shows the switch rule dialog
     */   
    private class SwitchRuleAction extends StandardAction
    {
        /** 
         * Creates a new SwitchMaterializationMultiplicityAction instance.
         */
    	SwitchRuleAction()
        {
            super(ResourceKeys.CMD_SWITCH_RULE);
        }

        /**
         * Called when this action is executed.
         */
		@Override
		public void actionPerformed(ActionEvent arg0) {
            SwitchRuleDialog dlg = new SwitchRuleDialog(MainFrame.this,rules,activeRuleIndex);
            dlg.initialize();
            Integer result = dlg.runDialog();
            if(result != null) {
            	if(activeRuleIndex != result) resetMatAndPostPanels();
            	currentRule = rules.get(result);
            	activeRuleIndex = result;
            	ruleView.setGraph(VisiTools.getInstance().visualizeRuleDPO(currentRule)); 
            	ruleScroll.revalidate();
            }
		}
    
    }
    
    /**
     * Action which show the about box.
     */
    private class AboutAction extends StandardAction
    {
        /** 
         * Creates a new AboutAction instance.
         */
        AboutAction()
        {
            super(ResourceKeys.CMD_ABOUT);
        }

        /**
         * Called when this action is executed.
         */
        @Override
        public void actionPerformed(@NonNull ActionEvent e)
        {
            AboutDialog dlg = new AboutDialog(MainFrame.this);
            dlg.initialize();
            dlg.runDialog();
        }
    }
      
    /**
     * Action which opens a website in the browser.
     */
    private class WebsiteAction extends StandardAction
    {
        private URI uri;
        
        /**
         * Creates a new WebsiteAction instance.
         * 
         * @param key
         *      resource key for the caption of the action
         * @param uri
         *      URI of the website which is displayed if this action is 
         *      executed
         */
        WebsiteAction(@NonNull String key, @NonNull URI uri)
        {
            super(key);
            this.uri = uri;
        }

        /**
         * Called when this action is executed.
         */
        @Override
        public void actionPerformed(@NonNull ActionEvent e) 
        {
            try {
                if (uri != null && Desktop.isDesktopSupported()) {
                    Desktop dt = Desktop.getDesktop();
                    if (dt.isSupported(Desktop.Action.BROWSE)) {
                        dt.browse(uri);
                    }
                }
            }
            catch (IOException ex) {
                showError(ex);
            }
        }
    }
   
    /**
     * Action which exits the application.
     */
    private class ExitAction extends StandardAction
    {
        /**
         * Creates a new ExitAction instance.
         */
        ExitAction() 
        {
            super(ResourceKeys.CMD_EXIT);
        }
        
        /**
         * Called when this action is executed.
         */
        @Override
        public void actionPerformed(@NonNull ActionEvent e)
        {
            if (exit()) {
                System.exit(0);
            }
        }
    }
    
    /**
     * Action which performs an invariant check for the current GTS and annotated type graph
     */
	private class InvariantCheckAction extends StandardAction
    {
        /**
         * Creates a new InvariantCheckAction instance.
         */
    	InvariantCheckAction() 
        {
            super(ResourceKeys.CMD_INVARIANT_CHECK);
        }
        
        /**
         * Called when this action is executed.
         */
        @Override
        public void actionPerformed(@NonNull ActionEvent e)
        {
        	SwingWorker<Void, Void> mySwingWorker = new SwingWorker<Void, Void>(){
        		@Override
	            protected Void doInBackground() throws Exception {
        			
	            	invariantList = new HashMap<Rule,HashMap<Morphism,Postcondition>>();
	            	validInvariant = true;
	            	resetMatAndPostPanels();
	            	
	            	try {
		            	//For every rule in the GTS...
		        		for(Rule r: rules) {
		        			currentRule = r;
		        			List<Node> deadNodes = computeDeadNodes();
		        			HashMap<Morphism, List<Integer>> morphAnnotPairs = computeSemiLegalMorphisms();
		        			HashMap<Morphism,Postcondition> baseAndLegal = new HashMap<Morphism,Postcondition>();
		        			
	        				//...and for all semi-legal morphisms from the current left-hand side...
	        				for(Morphism morph: morphAnnotPairs.keySet()) {
	        					
	        					//...compute the rewritable materialization alongside the postcondition...
	        					MaterializationConstructionAlgorithm algo = new MaterializationConstructionAlgorithm(morph, morphAnnotPairs.get(morph), deadNodes);
	        					matti = algo.computeMaterialization();	        					
	        					
	        					PostconditionConstructionAlgorithm algo2 = new PostconditionConstructionAlgorithm(currentRule, matti.getLeftEmbedding());
	        					postcondition = algo2.computePostcondition();

	        					//...and check for a legal morphism from the postcondition to the abstract graph.
	        					Morphism legalMorphism = checkInclusion(postcondition.getPostGraph(),graph);
	        					if(legalMorphism != null) {
	        						postcondition.setLegalMorphism(legalMorphism);
	        						baseAndLegal.put(morph, postcondition);
	        					}
	        					else {
	        						validInvariant = false;
	        						break;
	        					}
	        				}
	        				invariantList.put(r, baseAndLegal);
	        				if(!validInvariant) break;		        				
		        			
		        		}
	            	} 
	            	catch (UnsatisfiedLinkError e) {
	            		showInformation("Could not find libz3java in java.library.path.", JOptionPane.ERROR_MESSAGE);
	            		validInvariant = false;
	            	}
	            	
	            	return null;
	          	}
            };

             final JDialog dialog = new JDialog((Frame)null, "InvariantCheck", ModalityType.APPLICATION_MODAL);

             mySwingWorker.addPropertyChangeListener(new PropertyChangeListener() {

                @Override
                public void propertyChange(PropertyChangeEvent evt) {
                   if (evt.getPropertyName().equals("state")) {
                      if (evt.getNewValue() == SwingWorker.StateValue.DONE) {
                         dialog.dispose();
                         
                         if(validInvariant) {
                        	currentRule = rules.get(activeRuleIndex);
                        	List<Rule> applicableRules = new LinkedList<>();
                        	String nonApplicableRule = "";
                        	
                        	for(int i=0; i<rules.size();i++){
                        		rules.get(i).setName("Rule " + (i+1));
                    			if(invariantList.get(rules.get(i)).keySet().isEmpty()) {
                    				nonApplicableRule = nonApplicableRule + (i+1) + ",";	
                    			}
                    			else {
                    				applicableRules.add(rules.get(i));
                    			}
                    		}
                        	
                        	String extraInformation = "(All rules are applicable)";
                        	if(!nonApplicableRule.equals("")) {
                        		int length = nonApplicableRule.length();
                        		extraInformation = "(Non-applicable rule ID's: " + nonApplicableRule.substring(0, length-1) + ")";
                        	}
                        		                        		
                    		showInformation("The language of the abstract graph \nis an invariant for the given system.\n" + extraInformation, JOptionPane.INFORMATION_MESSAGE);
                    		if(!applicableRules.isEmpty()){
                    			InvariantCheckResultDialog dlg = new InvariantCheckResultDialog(MainFrame.this,applicableRules,invariantList);
                    			dlg.initialize();
                    			dlg.runDialog();                    			
                    		}

                         }
                         else {                         	 
                        	 updateMaterializationView();
                        	 updatePostconditionView();
                        	 ruleView.setGraph(VisiTools.getInstance().visualizeRuleDPO(currentRule)); 
                         	 ruleScroll.revalidate();
                         	 activeRuleIndex = invariantList.keySet().size()-1;
                         	                       	 
                        	 showInformation("No legal morphism from the postcondition \nto the abstract graph was found for rule " + (activeRuleIndex+1) + ".\nCan't prove invariant for the given system.", JOptionPane.ERROR_MESSAGE);
                         }
                         mySwingWorker.cancel(true);
                      }
                   }
                }
             });
             mySwingWorker.execute();

             JProgressBar progressBar = new JProgressBar();
             progressBar.setIndeterminate(true);
             progressBar.setOpaque(false);
             JPanel panel = new JPanel(new BorderLayout());
             JLabel text = new JLabel("Checking for Invariant. Please wait...");
             text.setOpaque(false);
             panel.add(progressBar, BorderLayout.CENTER);
             panel.add(text, BorderLayout.PAGE_START);
             panel.setBackground(Color.WHITE);
             dialog.setUndecorated(true);
             dialog.add(panel);
             dialog.pack();
             dialog.setLocationRelativeTo(null);
             dialog.setVisible(true);
        }
    }
	
    
    /**
     * Action which saves a type graph.
     */
    private class SaveGraphAction extends StandardAction 
    {
        /**
         * Creates a new ExitAction instance.
         */
    	SaveGraphAction() 
        {
            super(ResourceKeys.CMD_TYPE_GRAPH);
        }
        
        /**
         * Called when this action is executed.
         */
        @Override
        public void actionPerformed(@NonNull ActionEvent e)
        {
        	new File("savedata/graphs").mkdirs();
        	fileChooser = new JFileChooser("savedata/graphs");
        	fileChooser.setFileFilter(
                    FileTypes.getFileFilterForType(FileTypes.FileType.GXL));
        	try{
        		int returnVal = fileChooser.showSaveDialog(MainFrame.this);
        		if (returnVal == JFileChooser.APPROVE_OPTION) {
        			File file = null;
        			String absolutePath = fileChooser.getSelectedFile().getAbsolutePath();
        			if(!absolutePath.endsWith(".gxl")){
        				file = new File(absolutePath + ".gxl");
        			}
        			else{
        				file = new File(absolutePath);
        			}
        			GXLWriter.getWriter().write(graph, file);
        			//SgfWriter writer = new SgfWriter(new PrintStream(file, StandardCharsets.UTF_8.name()));
        		}    		
        	}
            catch (IOException ex) {
                showError(ex);
            }
        }
    }
    
    
    /**
     * Action which loads a type graph.
     */
    private class LoadGraphAction extends StandardAction 
    {
        /**
         * Creates a new ExitAction instance.
         */
    	LoadGraphAction() 
        {
            super(ResourceKeys.CMD_TYPE_GRAPH);
        }
        
        /**
         * Called when this action is executed.
         */
        @Override
        public void actionPerformed(@NonNull ActionEvent e)
        {
        	new File("savedata/graphs").mkdirs();
        	fileChooser = new JFileChooser("savedata/graphs");
        	fileChooser.setFileFilter(
                    FileTypes.getFileFilterForType(FileTypes.FileType.GXL));
        	
        	int returnVal = fileChooser.showOpenDialog(MainFrame.this);
        		 
            if (returnVal == JFileChooser.APPROVE_OPTION) {
               File file = fileChooser.getSelectedFile();
               
               try {
            	AnnotatedTypeGraph result = GXLReader.readGraph(file.getAbsolutePath());
            	if(result != null){
            		graph = result.clone();
            		graphView.setGraph(VisiTools.getInstance().visualizeGraph(graph)); 
            		typeGraphPanel.setVisible(true);
            		saveGraph.setEnabled(true);
               	   	graphScroll.revalidate();
               	   	resetMatAndPostPanels();
                }
                else{
             	   System.out.println("The gxl file could not be parsed correctly");
             	  showInformation("The gxl file could not be parsed correctly.", JOptionPane.ERROR_MESSAGE);
             	   
                }
               } catch (FileNotFoundException e1) {
            	   e1.printStackTrace();
               } catch (IOException e1) {
            	   e1.printStackTrace();
               } catch (AnnotatedTypeGraphException e1) {
               	   showInformation("The gxl file could not be parsed correctly.\nThe annotation bounds are incorrect.", JOptionPane.ERROR_MESSAGE);
               }
               
               
            }
        }
    }
    
    
    /**
     * Action which saves a type graph.
     */
    private class SaveGTSAction extends StandardAction 
    {
        /**
         * Creates a new SaveGTSAction instance.
         */
    	SaveGTSAction() 
        {
            super(ResourceKeys.CMD_GTS);
        }
        
        /**
         * Called when this action is executed.
         */
        @Override
        public void actionPerformed(@NonNull ActionEvent e)
        {
        	new File("savedata/gts").mkdirs();
        	fileChooser = new JFileChooser("savedata/gts");
        	fileChooser.setFileFilter(
                    FileTypes.getFileFilterForType(FileTypes.FileType.SGF));
        	try{
        		int returnVal = fileChooser.showSaveDialog(MainFrame.this);
        		if (returnVal == JFileChooser.APPROVE_OPTION) {
        			File file = null;
        			String absolutePath = fileChooser.getSelectedFile().getAbsolutePath();
        			if(!absolutePath.endsWith(".sgf")){
        				file = new File(absolutePath + ".sgf");
        			}
        			else{
        				file = new File(absolutePath);
        			}
        			SgfWriter writer = new SgfWriter(new PrintStream(file, StandardCharsets.UTF_8.name()));
        			
        			for(int index=0; index < rules.size();index++){
        				rules.get(index).getLeft().setName("leftGraph" + index);
        				rules.get(index).getRight().setName("rightGraph" + index);
        				rules.get(index).setName("Rule" + index);
        			}
        			TransformationSystem system = TransformationSystem.create(null, rules);
        			system.setName(file.getName().substring(0, file.getName().lastIndexOf(".")));
        			writer.writeTransformationSystem(system);
        		}    		
        	}
            catch (IOException ex) {
                showError(ex);
            }
        }
    }
    
        
    /**
     * Action which loads another GTS.
     */
    private class LoadGTSAction extends StandardAction
    {
        /**
         * Creates a new LoadGTSAction instance.
         */
        LoadGTSAction()
        {
            super(ResourceKeys.CMD_GTS);
        }
        
        /**
         * Called when this action is executed.
         */
        @Override
        public void actionPerformed(@NonNull ActionEvent e)
        {
            GTSChooser chooser = new GTSChooser();
            
            TransformationSystem system = chooser.showDialog(MainFrame.this);
            if (system != null) {
            	try{
            		List<Rule> loadRules = new ArrayList<>();
	                for(Rule rule: system.getRules()){
	                	loadRules.add(convertToAnnotatedTypeGraphRule(rule));
	                }
	            	rules.clear();
	            	rules = loadRules;
	                currentRule = rules.get(0);
	                activeRuleIndex = 0;
	            	ruleView.setGraph(VisiTools.getInstance().visualizeRuleDPO(currentRule));
	            	rulePanel.setVisible(true);
	            	saveRule.setEnabled(true);
	            	ruleScroll.revalidate();
	            	resetMatAndPostPanels();
            	}
            	catch(UnsupportedOperationException err){
            		showInformation("DrAGoM does not support hypergraphs in its current version.", JOptionPane.ERROR_MESSAGE);
            	}
            }
        }
    }

    /**
     * Action which saves a postcondition graph.
     */
    private class SavePostconditionAction extends StandardAction 
    {
        /**
         * Creates a new SavePostconditionAction instance.
         */
    	SavePostconditionAction() 
        {
            super(ResourceKeys.CMD_POSTCONDITION);
        }
        
        /**
         * Called when this action is executed.
         */
        @Override
        public void actionPerformed(@NonNull ActionEvent e)
        {
        	new File("savedata/graphs").mkdirs();
        	fileChooser = new JFileChooser("savedata/graphs");
        	fileChooser.setFileFilter(
                    FileTypes.getFileFilterForType(FileTypes.FileType.GXL));
        	try{
        		int returnVal = fileChooser.showSaveDialog(MainFrame.this);
        		if (returnVal == JFileChooser.APPROVE_OPTION) {
        			File file = null;
        			String absolutePath = fileChooser.getSelectedFile().getAbsolutePath();
        			if(!absolutePath.endsWith(".gxl")){
        				file = new File(absolutePath + ".gxl");
        			}
        			else{
        				file = new File(absolutePath);
        			}
        			GXLWriter.getWriter().write(postcondition.getPostGraph(), file);
        			//SgfWriter writer = new SgfWriter(new PrintStream(file, StandardCharsets.UTF_8.name()));
        		}    		
        	}
            catch (IOException ex) {
                showError(ex);
            }
        }
    }
    
    
    /**
     * Action which computes the rewritable materialization.
     */
    private class MaterializationAction extends StandardAction
    {
        /**
         * Creates a new MaterializationAction instance.
         */
    	MaterializationAction() 
        {
    		super(ResourceKeys.CMD_MATERIALIZATION);
        }
        
        /**
         * Called when this action is executed.
         */
        @Override
        public void actionPerformed(@NonNull ActionEvent e)
        {
        	HashMap<Morphism, List<Integer>> morphAnnotPairs = computeSemiLegalMorphisms();
        	if(morphAnnotPairs.keySet().size() > 0) {
        		int formerActiveMultiplicity = graph.getActiveMultiplicityIndex();
        		ChooseBaseMorphismDialog dlg = new ChooseBaseMorphismDialog(MainFrame.this,morphAnnotPairs);
        		dlg.initialize();
        		Morphism result = dlg.runDialog();
        		graph.selectMultiplicity(formerActiveMultiplicity);
        		if(result != null) {
        			try{
        				List<Node> deadNodes = computeDeadNodes();
        				MaterializationConstructionAlgorithm algo = new MaterializationConstructionAlgorithm(result, morphAnnotPairs.get(result), deadNodes);
        				matti = algo.computeMaterialization();       			
        				updateMaterializationView();
        				postGraphPanel.setVisible(false);        				
        			}
	            	catch (UnsatisfiedLinkError err) {
	            		showInformation("Could not find libz3java in java.library.path.", JOptionPane.ERROR_MESSAGE);
	            	}
        		}
            }
            else {
            	showInformation("No semi-legal morphism was found.", JOptionPane.WARNING_MESSAGE);
            }
        	
        }

    }
    
    /**
     * Action which computes the strongest postcondition.
     */
    private class PostconditionAction extends StandardAction
    {
        /**
         * Creates a new PostconditionAction instance.
         */
    	PostconditionAction() 
        {
    		super(ResourceKeys.CMD_POSTCONDITION);
        }
        
        /**
         * Called when this action is executed.
         */
        @Override
        public void actionPerformed(@NonNull ActionEvent e)
        {
        		PostconditionConstructionAlgorithm algo = new PostconditionConstructionAlgorithm(currentRule, matti.getLeftEmbedding());
        		postcondition = algo.computePostcondition();
  		
        		postcondition.getPostGraph().selectMultiplicity(matti.getMaterialization().getActiveMultiplicityIndex());
        		updatePostconditionView();
        }

    }

    // ---------------------------------------------------------------------- //
    //                         Private Methods                                //
    // ---------------------------------------------------------------------- //
    /**
     * Shows an error message containing the message of an exception and the
     * messages of all its causes.
     * 
     * @param ex
     *      Throwable to display to the user
     */
    private void showError(@NonNull Throwable ex) 
    {
        StringBuilder buf = new StringBuilder();
        do {
            buf.append(ex.getMessage())
               .append("\n");
            ex = ex.getCause();
        } while (ex != null);
        JOptionPane.showMessageDialog(
                this, 
                buf.toString(),
                GuiContext.getSimpleGuiString(ResourceKeys.TIT_ERROR),
                JOptionPane.ERROR_MESSAGE);
    }
    
    /**
     *   Shows any type of information or notification that needs to be acknowledged by the user
     * @param message
     * 		The message to be shown
     * @param messageType
     * 		Message type to be used (use one of the default JOptionPane.TYPE) 
     * @param width
     * 		Sets the width of the window to be shown
     * @param height
     * 		Sets the height of the window to be shown
     */
    private void showInformation(String message, int messageType) {
    	JOptionPane jop = new JOptionPane(message, messageType,JOptionPane.DEFAULT_OPTION, null, new Object[]{}, null);          	
    	DragomInformationDialog info = new DragomInformationDialog(MainFrame.this, "", jop, true);
    	info.initialize();
    	info.runDialog();
    }
    
    /**
     *   Shows any type of confirmation that needs to be acknowledged by the user
     * @param message
     * 		The message to be shown
     * @param messageType
     * 		Message type to be used (use one of the default JOptionPane.TYPE) 
     * @param width
     * 		Sets the width of the window to be shown
     * @param height
     * 		Sets the height of the window to be shown
     */
    private boolean showConfirmation(String message, int messageType) {
    	JOptionPane jop = new JOptionPane(message, messageType,JOptionPane.DEFAULT_OPTION, null, new Object[]{}, null);          	
    	DragomConfirmationDialog info = new DragomConfirmationDialog(MainFrame.this, "", jop);
    	info.initialize();
    	
    	return info.runDialog() != null;
    }
    
    /**
     * Resets/hides materialization panel and the poscondition panel/action
     */
    private void resetMatAndPostPanels(){
    	postGraphPanel.setVisible(false);
    	postconditionItem.setEnabled(false);
    	savePostcondition.setEnabled(false);
    	materializationGraphPanel.setVisible(false);
    	if(typeGraphPanel.isVisible() && rulePanel.isVisible()) {
    		materializationItem.setEnabled(true);
    		invariantItem.setEnabled(true);
    	}
    }
    
    /**
     * Computes all semi-legal morphisms between the left-hand side of the active rule and the annotated type graph
     * @return 
     * 		A list of semi-legal morphisms
     */
	private HashMap<Morphism, List<Integer>> computeSemiLegalMorphisms() {
		HashMap<Morphism, List<Integer>> result = new HashMap<Morphism, List<Integer>>();
		List<Integer> validMultiplicityIndexes;
		AnnotatedTypeGraph codomain;
		int maxBound;
		
		List<Node> nodeImage;
		List<Edge> edgeImage;
		boolean add = false;
		int limit = graph.getLimit();
		List<HashMap<GraphElement,AnnotationBounds>> multiplicities = graph.getMultiplicities();
		
		for(Morphism morph: Morphism.getMatches(currentRule.getLeft(), graph, false)){
    		if(morph != null) {
    			validMultiplicityIndexes = new LinkedList<Integer>();
    			codomain = (AnnotatedTypeGraph) morph.getCodomain();

    			for(int index=0; index < multiplicities.size();index++) {
    				add = true;
    				
    				for(Node n: codomain.getNodes()) {
    					if(add) {
    						maxBound = multiplicities.get(index).get(n).getMax();
    						if(maxBound != limit) {
    							nodeImage = new LinkedList<Node>();
    							nodeImage.add(n); 
    							
    							if(morph.getPreimageOfNodes(nodeImage).size() > maxBound) {
    								add = false;
    							}
    						}        					
    					}
    				}
    				
    				if(add) {
    					for(Edge e: codomain.getEdgeList()) {
    						if(add) {
    							maxBound = multiplicities.get(index).get(e).getMax();
    							if(maxBound != limit) {
    								edgeImage = new LinkedList<Edge>();
    								edgeImage.add(e); 
    								
    								if(morph.getPreimageOfEdges(edgeImage).size() > maxBound) {
    									add = false;
    								}
    							}        					
    						}
    					}
    				}
    				
    				if(add) {
    					validMultiplicityIndexes.add(index);
    				}  
    			}
    			if(validMultiplicityIndexes.size() > 0) {
    				result.put(morph, validMultiplicityIndexes);
    			}
    		}
    	}
		return result;
	}
	
    /**
     * Checks for a language inclusion of the postcondition' language into the annotated type graph language
     * @return 
     * 		Whether or not the postcondition has a legal morphism into the annotated type graph. 
     * If the postcondition has a legal morphism, this morhpism is returned, null otherwise.
     */
	private Morphism checkInclusion(AnnotatedTypeGraph source, AnnotatedTypeGraph target) {
		Morphism legalMorphism = null;
		List<Integer> validTargetMultiplicityIndexes;
		HashMap<Integer,Integer> legalAnnotationIndexes;
		int maxBound;
		int minBound;
		
		List<Node> nodeImage;
		List<Edge> edgeImage;
		
		boolean add = false;
		boolean existsAnnot = false;
		boolean stillValidSourceMultiplicities = true;
		boolean fullCheck = false;
		int limit = target.getLimit();
		int minSum;
		int maxSum;
		List<HashMap<GraphElement,AnnotationBounds>> sourceMultiplicities = source.getMultiplicities();
		List<HashMap<GraphElement,AnnotationBounds>> targetMultiplicities = target.getMultiplicities();	
		
		for(Morphism morph: Morphism.getMatches(source, target, false)){
			fullCheck = true;
			if(morph != null) {
    			if(!existsAnnot) {
    				legalAnnotationIndexes = new HashMap<Integer,Integer>();
    				stillValidSourceMultiplicities = true;
    				for(int sIndex=0; sIndex < sourceMultiplicities.size();sIndex++) {
    					if(stillValidSourceMultiplicities) {
	    					validTargetMultiplicityIndexes = new LinkedList<Integer>();
	    					
	    	    			for(int tIndex=0; tIndex < targetMultiplicities.size();tIndex++) {
	    	    				if(validTargetMultiplicityIndexes.size() == 0) {
	        	    				add = true;
	        	    				
	        	    				for(Node n: target.getNodes()) {
	        	    					if(add) {
	        	    						minBound = targetMultiplicities.get(tIndex).get(n).getMin();
	        	    						maxBound = targetMultiplicities.get(tIndex).get(n).getMax();
	        	    						if((minBound != 0) || (maxBound != limit)) {
	        	    							nodeImage = new LinkedList<Node>();
	        	    							nodeImage.add(n); 
	        	    							minSum = 0;
	        	    							maxSum = 0;
	        	    							
	        	    							for(Node pINode: morph.getPreimageOfNodes(nodeImage)) {
	        	    								minSum = minSum + sourceMultiplicities.get(sIndex).get(pINode).getMin();
	        	    								maxSum = maxSum + sourceMultiplicities.get(sIndex).get(pINode).getMax();
	        	    							}
	        	    							if(maxSum > limit) maxSum = limit;
	        	    							
	        	    							
	        	    							if(maxSum > maxBound || minSum < minBound) {
	        	    								add = false;
	        	    							}
	        	    						}        					
	        	    					}
	        	    				}
	        	    				
	        	    				if(add) {
	        	    					for(Edge e: target.getEdgeList()) {
	        	    						if(add) {
		        	    						minBound = targetMultiplicities.get(tIndex).get(e).getMin();
		        	    						maxBound = targetMultiplicities.get(tIndex).get(e).getMax();
		        	    						if((minBound != 0) || (maxBound != limit)) {
		        	    							edgeImage = new LinkedList<Edge>();
		        	    							edgeImage.add(e); 
		        	    							minSum = 0;
		        	    							maxSum = 0;
		        	    							
		        	    							for(Edge pIEdge: morph.getPreimageOfEdges(edgeImage)) {
		        	    								minSum = minSum + sourceMultiplicities.get(sIndex).get(pIEdge).getMin();
		        	    								maxSum = maxSum + sourceMultiplicities.get(sIndex).get(pIEdge).getMax();
		        	    							}
		        	    							if(maxSum > limit) maxSum = limit;
		        	    							
		        	    							
		        	    							if(maxSum > maxBound || minSum < minBound) {
		        	    								add = false;
		        	    							}
		        	    						}        					
	        	    						}
	        	    					}
	        	    				}
	        	    				
	        	    				if(add) {
	        	    					validTargetMultiplicityIndexes.add(tIndex);
	        	    				} 
	    	    				}
	    	    			}//tIndexIteration
	    	    			
	    	    			if(validTargetMultiplicityIndexes.size() == 0) {
	    	    				stillValidSourceMultiplicities = false;
	    	    				matti.getMaterialization().selectMultiplicity(sIndex);
	    	    				postcondition.getPostGraph().selectMultiplicity(sIndex);
	    	    			}
	    	    			else {
	    	    				legalAnnotationIndexes.put(sIndex, validTargetMultiplicityIndexes.get(0));
	    	    			}
    					}
    					
        			}//sIndexIteration
    				if(stillValidSourceMultiplicities) {
    					existsAnnot = true;
    					postcondition.setAnnotationIndexes(legalAnnotationIndexes);
    					legalMorphism = morph;
    				}
    			}
    		}
    		else {
    			return legalMorphism;
    		}
    	}
		//In case we can not fully embed the postcondition due to edges with annotation [0,0]...
		if(!fullCheck){
			boolean removed = false;
			boolean removable = true;
			List<Edge> allEdges = new LinkedList<Edge>(source.getEdges());
			List<Edge> removedEdges = new LinkedList<Edge>();
			//...remove the [0,0] annotated edges to check for the partial morphism
			
			
			for(Edge e:allEdges){
				removable = true;
				for(int partialIndex=0; partialIndex < sourceMultiplicities.size(); partialIndex++){
					if((sourceMultiplicities.get(partialIndex).get(e).getMin() != 0) || (sourceMultiplicities.get(partialIndex).get(e).getMax() != 0)){
						removable = false;
					}
				}
				if(removable){
					removed = true;
					removedEdges.add(e);
				}
			}
			
			for(Edge rem: removedEdges){
				source.removeAnnotatedEdge(rem);
			}
				
			if(removed){
				for(Morphism morph: Morphism.getMatches(source, target, false)){
					if(morph != null) {
		    			if(!existsAnnot) {
		    				legalAnnotationIndexes = new HashMap<Integer,Integer>();
		    				stillValidSourceMultiplicities = true;
		    				for(int sIndex=0; sIndex < sourceMultiplicities.size();sIndex++) {
		    					if(stillValidSourceMultiplicities) {
			    					validTargetMultiplicityIndexes = new LinkedList<Integer>();
			    					
			    	    			for(int tIndex=0; tIndex < targetMultiplicities.size();tIndex++) {
			    	    				if(validTargetMultiplicityIndexes.size() == 0) {
			        	    				add = true;
			        	    				
			        	    				for(Node n: target.getNodes()) {
			        	    					if(add) {
			        	    						minBound = targetMultiplicities.get(tIndex).get(n).getMin();
			        	    						maxBound = targetMultiplicities.get(tIndex).get(n).getMax();
			        	    						if((minBound != 0) || (maxBound != limit)) {
			        	    							nodeImage = new LinkedList<Node>();
			        	    							nodeImage.add(n); 
			        	    							minSum = 0;
			        	    							maxSum = 0;
			        	    							
			        	    							for(Node pINode: morph.getPreimageOfNodes(nodeImage)) {
			        	    								minSum = minSum + sourceMultiplicities.get(sIndex).get(pINode).getMin();
			        	    								maxSum = maxSum + sourceMultiplicities.get(sIndex).get(pINode).getMax();
			        	    							}
			        	    							if(maxSum > limit) maxSum = limit;
			        	    							
			        	    							
			        	    							if(maxSum > maxBound || minSum < minBound) {
			        	    								add = false;
			        	    							}
			        	    						}        					
			        	    					}
			        	    				}
			        	    				
			        	    				if(add) {
			        	    					for(Edge e: target.getEdgeList()) {
			        	    						if(add) {
				        	    						minBound = targetMultiplicities.get(tIndex).get(e).getMin();
				        	    						maxBound = targetMultiplicities.get(tIndex).get(e).getMax();
				        	    						if((minBound != 0) || (maxBound != limit)) {
				        	    							edgeImage = new LinkedList<Edge>();
				        	    							edgeImage.add(e); 
				        	    							minSum = 0;
				        	    							maxSum = 0;
				        	    							
				        	    							for(Edge pIEdge: morph.getPreimageOfEdges(edgeImage)) {
				        	    								minSum = minSum + sourceMultiplicities.get(sIndex).get(pIEdge).getMin();
				        	    								maxSum = maxSum + sourceMultiplicities.get(sIndex).get(pIEdge).getMax();
				        	    							}
				        	    							if(maxSum > limit) maxSum = limit;
				        	    							
				        	    							
				        	    							if(maxSum > maxBound || minSum < minBound) {
				        	    								add = false;
				        	    							}
				        	    						}        					
			        	    						}
			        	    					}
			        	    				}
			        	    				
			        	    				if(add) {
			        	    					validTargetMultiplicityIndexes.add(tIndex);
			        	    				} 
			    	    				}
			    	    			}//tIndexIteration
			    	    			
			    	    			if(validTargetMultiplicityIndexes.size() == 0) {
			    	    				stillValidSourceMultiplicities = false;
			    	    				matti.getMaterialization().selectMultiplicity(sIndex);
			    	    				postcondition.getPostGraph().selectMultiplicity(sIndex);
			    	    			}
			    	    			else {
			    	    				legalAnnotationIndexes.put(sIndex, validTargetMultiplicityIndexes.get(0));
			    	    			}
		    					}
		    					
		        			}//sIndexIteration
		    				if(stillValidSourceMultiplicities) {
		    					existsAnnot = true;
		    					postcondition.setAnnotationIndexes(legalAnnotationIndexes);
		    					legalMorphism = morph;
		    				}
		    			}
		    		}
		    		else {
		    			return legalMorphism;
		    		}
		    	}
				
				for(Edge ret: removedEdges){
					source.reAddAnnotatedEdge(ret.getLabel(), ret.getNodes());
				}
				
				return legalMorphism;
			}				
			
		}
		
		return legalMorphism;
	}
    
	/**
	 * Computes all obsolete nodes in the current rule
	 * @return The list of obsolete nodes
	 */
	private List<Node> computeDeadNodes() {
		List<Node> livingNodes = new LinkedList<Node>();
		Morphism corr = currentRule.getCorrespondence();

		for(Node n: corr.getNodeMap().keySet()) {
			livingNodes.add(n);
		}
		List<Node> deadNodes = new LinkedList<Node>();
		for(Node m: currentRule.getLeft().getNodes()) {
			if(!livingNodes.contains(m)) deadNodes.add(m);
		}
		return deadNodes;
	}
	
	/**
	 * Converts a graph transformation rule into one where the graphs are standard annotated graphs
	 * @param rule The graph transformation rule
	 * @return The modified rule with standard annotations
	 */
	private Rule convertToAnnotatedTypeGraphRule(Rule rule){
		HashMap<GraphElement,GraphElement> leftConvert = new HashMap<GraphElement,GraphElement>(); 
		HashMap<GraphElement,GraphElement> rightConvert = new HashMap<GraphElement,GraphElement>();
		
		AnnotatedTypeGraph left = new AnnotatedTypeGraph();
		AnnotatedTypeGraph right = new AnnotatedTypeGraph();
		
		HashMap<Node,Node> nodeMap = new HashMap<Node,Node>();
		HashMap<Edge,Edge> edgeMap = new HashMap<Edge,Edge>();
		
		for(Node nl: rule.getLeft().getNodes()){
			leftConvert.put(nl, left.addStandardAnnotatedNode());
		}
		for(Edge el: rule.getLeft().getEdges()){
			leftConvert.put(el, left.addStandardAnnotatedEdge(el.getLabel(),(Node) leftConvert.get(el.getSource()),(Node) leftConvert.get(el.getTarget())));
		}
		
		for(Node nr: rule.getRight().getNodes()){
			rightConvert.put(nr, right.addStandardAnnotatedNode());
		}
		for(Edge er: rule.getRight().getEdges()){
			rightConvert.put(er, right.addStandardAnnotatedEdge(er.getLabel(),(Node) rightConvert.get(er.getSource()),(Node) rightConvert.get(er.getTarget())));
		}
		
		Map<Node,Node> oldNodeMap = rule.getCorrespondence().getNodeMap();
		for(Node nm: oldNodeMap.keySet()){
			nodeMap.put((Node) leftConvert.get(nm),(Node) rightConvert.get(oldNodeMap.get(nm)));
		}
		
		Map<Edge,Edge> oldEdgeMap = rule.getCorrespondence().getEdgeMap();
		for(Edge em: oldEdgeMap.keySet()){
			edgeMap.put((Edge) leftConvert.get(em),(Edge) rightConvert.get(oldEdgeMap.get(em)));
		}
		
		return new Rule(Morphism.create(left, right, nodeMap, edgeMap));
	}
	
	/**
	 * Updates the materialization view
	 */
	private void updateMaterializationView() {
   	 	materializationView.setGraph(VisiTools.getInstance().visualizeColoredGraph(matti.getMaterialization(),
																			matti.getLeftEmbeddingNodeImage(),
																			matti.getLeftEmbeddingEdgeImage(),
																			GraphColor.BLUE,
																			simplifyBox.isSelected()));
		materializationGraphPanel.setVisible(true);
		postconditionItem.setEnabled(true);
		matScroll.revalidate();
	}

	/**
	 * Updates the postcondition view
	 */
	private void updatePostconditionView() {
		postView.setGraph(VisiTools.getInstance().visualizeColoredGraph(postcondition.getPostGraph(),
																 postcondition.getRightHandNodes(),
																 postcondition.getRightHandEdges(),
																 GraphColor.YELLOW,
																 simplifyBox.isSelected()));
		postGraphPanel.setVisible(true);
		savePostcondition.setEnabled(true);
		postScroll.revalidate();
	}
}
