/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.gui;

import java.awt.geom.Point2D;
import java.util.HashMap;

import de.uni_due.inf.ti.annotation.Nullable;
import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.Label;
import de.uni_due.inf.ti.graph.Morphism;
import de.uni_due.inf.ti.graph.Node;
import de.uni_due.inf.ti.graph.Rule;
import de.uni_due.inf.ti.swing.StandardDialog;
import de.uni_due.inf.ti.visigraph.VxEdge;
import de.uni_due.inf.ti.visigraph.VxGraph;
import de.uni_due.inf.ti.visigraph.VxLabelNode;
import de.uni_due.inf.ti.visigraph.VxNode;
import de.uni_due.inf.ti.visigraph.swing.GraphEditSynchronizer;

/**
 * AnnotatedGraphEditSynchronizer which uses {@link VxGraph} as the underlying model.
 * 
 * @author Dennis Nolte
 */
public class RuleEditSynchronizer implements GraphEditSynchronizer
{
    // Graph characteristics
    private boolean edgesToLabels;
    private boolean enabled;
    private int maxLabels;
    private StandardDialog<?> informationDealer;
    private AnnotatedTypeGraph leftGraph;
    private AnnotatedTypeGraph interfaceGraph;
    private AnnotatedTypeGraph rightGraph;
    private HashMap<VxNode,Node> leftNodeMap;
    private HashMap<VxNode,Node> interfaceNodeMap;
    private HashMap<VxNode,Node> rightNodeMap;
    private HashMap<VxNode,VxNode> interfaceToLeftNodeMap;
    private HashMap<VxNode,VxNode> interfaceToRightNodeMap;    
    
    private HashMap<Node,Node> ruleNodeMap;
    private HashMap<Edge,Edge> ruleEdgeMap;
    
    // Styles
    private String newLeftNodeStyleName;
    private String newInterfaceNodeStyleName;
    private String newRightNodeStyleName;
    private String newLeftEdgeStyleName;
    private String newInterfaceEdgeStyleName;
    private String newRightEdgeStyleName;
    private String newLabelStyleName;
        
    /**
     * Creates a new BasicGraphEditSynchronizer. 
     * The new BasicGraphEditSynchronizer does not allow edges to have edge
     * labels as source or target, and allows a single label per edge.
     */
    public RuleEditSynchronizer()
    {
        this(false, 1, null);
    }
    
    /**
     * Creates a new BasicGraphEditSynchronizer with custom behavior.
     * 
     * @param edgesToLabels
     *      whether the new BasicGraphEditSynchronizer allows edges to have
     *      edge labels as source or target
     * @param maxLabels
     *      maximum number of label nodes on an edge; -1 means unlimited
     * @throws IllegalArgumentException
     *      if {@code maxLabels < -1}
     */
    public RuleEditSynchronizer(boolean edgesToLabels, int maxLabels)
    {
        this(edgesToLabels, maxLabels, null);
    }
  
    public RuleEditSynchronizer(boolean edgesToLabels, int maxLabels,@Nullable StandardDialog<?> informationDealer)
    {
        if (maxLabels < -1) throw new IllegalArgumentException("maxLabels must be >= -1");
        this.edgesToLabels = edgesToLabels;
        this.maxLabels = maxLabels;
        this.informationDealer = informationDealer;
        leftGraph = new AnnotatedTypeGraph();
        interfaceGraph = new AnnotatedTypeGraph();
        rightGraph = new AnnotatedTypeGraph();
        
        leftNodeMap = new HashMap<>();
        interfaceNodeMap = new HashMap<>();
        rightNodeMap = new HashMap<>();
        
        interfaceToLeftNodeMap = new HashMap<>();
        interfaceToRightNodeMap = new HashMap<>();
        
        ruleNodeMap = new HashMap<Node,Node>();
        ruleEdgeMap = new HashMap<Edge,Edge>();
        enabled = true;
    }
    
    /**
     * Sets the name of the style of nodes that are added to the graph. 
     * 
     * @param style
     *      style of nodes added to the graph
     */
    public void setNewLeftNodeStyleName(String style)
    {
        newLeftNodeStyleName = style;
    }
    
    /**
     * Sets the name of the style of nodes that are added to the interface. 
     * 
     * @param style
     *      style of nodes added to the graph
     */
    public void setNewInterfaceNodeStyleName(String style)
    {
        newInterfaceNodeStyleName = style;
    }
    
    /**
     * Sets the name of the style of nodes that are added to the graph. 
     * 
     * @param style
     *      style of nodes added to the graph
     */
    public void setNewRightNodeStyleName(String style)
    {
        newRightNodeStyleName = style;
    }
        
    
    /**
     * Sets the name of the style of edges that are added to the graph by this 
     * GraphEditor.
     * @param style
     * 		style of edges added to a left-hand side graph
     */
    public void setNewLeftEdgeStyleName(String style)
    {
        newLeftEdgeStyleName = style;
    } 
    
    /**
     * Sets the name of the style of edges that are added to the interface. 
     * 
     * @param style
     *      style of nodes added to the graph
     */
    public void setNewInterfaceEdgeStyleName(String style)
    {
        newInterfaceEdgeStyleName = style;
    }
    
    /**
     * Sets the name of the style of edges that are added to the graph by this 
     * GraphEditor.
     * @param style
     * 		style of edges added to a right-hand side graph
     */
    public void setNewRightEdgeStyleName(String style)
    {
        newRightEdgeStyleName = style;
    } 
    
    /**
     * Returns the name of the style of labels that are added to the graph.
     * @return the String representing the label style
     */
    public String getNewLabelStyleName()
    {
        return newLabelStyleName;
    }
    
    /**
     * Sets the style of labels that are added to the graph by the user.
     * @param style
     * 		style of labels added to the graph
     */
    public void setNewLabelStyleName(String style)
    {
        newLabelStyleName = style;
    } 
  
    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isPossibleSourceNode(VxNode source)
    {
        if (edgesToLabels) {
            return true;
        }
        else {
            return !(source instanceof VxLabelNode);
        }
    }
    
    /**
     * {@inheritDoc}
     * <p>
     * This implementation simply returns {@code true}.
     */
    @Override
    public boolean canAddNode(VxGraph graph)
    {
        return true;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public boolean canAddEdge(VxNode source, VxNode target)
    {
        if (!edgesToLabels) {
            if (source instanceof VxLabelNode) return false;
            if (target instanceof VxLabelNode) return false;
        }
        return true;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public boolean canAddLabel(VxEdge edge)
    {
    	if(enabled){
	        int numOfLabels = edge.getLabels().size(); // throws NPE when edge == null
	        return (maxLabels < 0 || numOfLabels + 1 <= maxLabels);
    	}
    	return false;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public boolean canEditLabel(VxNode node)
    {
        return true;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void addNode(VxGraph graph, Point2D pt) 
    {
    	//Do nothing
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void addEdge(VxNode source, VxNode target) 
    {
    	//Do nothing
    }
    
    // Add left methods
    public void addLeftNode(VxGraph graph, Point2D pt) 
    {
    	if(enabled){
    		if(informationDealer != null){
    			VxNode newNode = graph.addNode(pt);
        		newNode.setStyle(newLeftNodeStyleName);
        		        		
        		leftNodeMap.put(newNode, leftGraph.addStandardAnnotatedNode());  
    		}
    		
    	}
    }
    
    public void addLeftEdge(VxNode source, VxNode target) 
    {
    	if(enabled){
    		if(informationDealer != null){
    			VxGraph graph = source.getGraph();
    	        VxEdge edge = graph.addEdge(source, target, newLeftEdgeStyleName); 
    	        
    	        String edgeLabel = ((CreateRuleDialog) informationDealer).getEdgeLabel();
    	        
        		addLabel(edge, edgeLabel);

        		leftGraph.addStandardAnnotatedEdge(new Label(edgeLabel), leftNodeMap.get(source), leftNodeMap.get(target));
    		}
    	}
    }

    // Add right methods
    public void addInterfaceNode(VxGraph graph, Point2D pt) 
    {
    	if(enabled){
    		if(informationDealer != null){
    			VxNode newNode = graph.addNode(pt);
        		newNode.setStyle(newInterfaceNodeStyleName);
        		
        		VxNode leftNode = ((CreateRuleDialog) informationDealer).getLeftGraph().addNode(pt);
        		leftNode.setStyle(newInterfaceNodeStyleName);
        		Node leftDataNode = leftGraph.addStandardAnnotatedNode();
        		
        		VxNode rightNode = ((CreateRuleDialog) informationDealer).getRightGraph().addNode(pt);        		
        		rightNode.setStyle(newInterfaceNodeStyleName);
        		Node rightDataNode = rightGraph.addStandardAnnotatedNode();
        		
        		leftNodeMap.put(leftNode, leftDataNode); 
        		interfaceNodeMap.put(newNode, interfaceGraph.addStandardAnnotatedNode()); 
        		rightNodeMap.put(rightNode, rightDataNode);
        		
        		ruleNodeMap.put(leftDataNode, rightDataNode);
        		
        		interfaceToLeftNodeMap.put(newNode, leftNode);
        		interfaceToRightNodeMap.put(newNode, rightNode);
    		}
    		
    	}
    }
    
    public void addInterfaceEdge(VxNode source, VxNode target) 
    {
    	if(enabled){
    		if(informationDealer != null){
    			VxGraph graph = source.getGraph();
    	        VxEdge edge = graph.addEdge(source, target, newInterfaceEdgeStyleName); 
    	        
    	        VxGraph leftVxGraph = ((CreateRuleDialog) informationDealer).getLeftGraph();
    	        VxEdge leftEdge = leftVxGraph.addEdge(interfaceToLeftNodeMap.get(source), interfaceToLeftNodeMap.get(target), newInterfaceEdgeStyleName);

    	        VxGraph rightVxGraph = ((CreateRuleDialog) informationDealer).getRightGraph();
    	        VxEdge rightEdge = rightVxGraph.addEdge(interfaceToRightNodeMap.get(source), interfaceToRightNodeMap.get(target), newInterfaceEdgeStyleName);
    	        
    	        String edgeLabel = ((CreateRuleDialog) informationDealer).getEdgeLabel();
    	        
        		addLabel(edge, edgeLabel);
        		addLabel(leftEdge, edgeLabel);
        		addLabel(rightEdge, edgeLabel);
        		
        		interfaceGraph.addStandardAnnotatedEdge(new Label(edgeLabel), interfaceNodeMap.get(source), interfaceNodeMap.get(target));	
      		
        		ruleEdgeMap.put(leftGraph.addStandardAnnotatedEdge(new Label(edgeLabel), leftNodeMap.get(interfaceToLeftNodeMap.get(source)), leftNodeMap.get(interfaceToLeftNodeMap.get(target))), 
        						rightGraph.addStandardAnnotatedEdge(new Label(edgeLabel), rightNodeMap.get(interfaceToRightNodeMap.get(source)), rightNodeMap.get(interfaceToRightNodeMap.get(target))));
    		}
    	}
    }
    
    // Add right methods
    public void addRightNode(VxGraph graph, Point2D pt) 
    {
    	if(enabled){
    		if(informationDealer != null){
    			VxNode newNode = graph.addNode(pt);
        		newNode.setStyle(newRightNodeStyleName);
        		        		
        		rightNodeMap.put(newNode, rightGraph.addStandardAnnotatedNode());
    		}
    		
    	}
    }
    
    public void addRightEdge(VxNode source, VxNode target) 
    {
    	if(enabled){
    		if(informationDealer != null){
    			VxGraph graph = source.getGraph();
    	        VxEdge edge = graph.addEdge(source, target, newRightEdgeStyleName); 
    	        
    	        String edgeLabel = ((CreateRuleDialog) informationDealer).getEdgeLabel();      
        		addLabel(edge, edgeLabel);

        		rightGraph.addStandardAnnotatedEdge(new Label(edgeLabel), rightNodeMap.get(source), rightNodeMap.get(target));
    		}
    	}
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void addLabel(VxEdge edge, String label)
    {
    	if(enabled){
	        VxLabelNode ln = edge.addLabel(label);
	        if (newLabelStyleName != null) {
	            ln.setStyle(newLabelStyleName);
	        }
    	}
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void editLabel(VxNode node, String newLabel)
    {
        node.setLabel(newLabel);
    }
    
    public Rule getResultRule(){
    	return new Rule(Morphism.create(leftGraph, rightGraph, ruleNodeMap, ruleEdgeMap));
    }
    
    /**
     * Sets whether this synchronizer may add anything to the graph
     * @param bool 
     * 		whether it is allowed or not
     */
    public void setEnabled(boolean bool)
    {
    	enabled = bool;
    }
}

