/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Set;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.dragom.general.ResourceKeys;
import de.uni_due.inf.ti.dragom.gui.AnnotatedTypeGraphVisualizer.GraphColor;
import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.Node;
import de.uni_due.inf.ti.swing.GuiContext;
import de.uni_due.inf.ti.swing.StandardDialog;
import de.uni_due.inf.ti.swing.UserInputException;
import de.uni_due.inf.ti.visigraph.VxGraph;
import de.uni_due.inf.ti.visigraph.swing.GraphPanel;

/**
 * Dialog which displays the switch annotation dialog.
 * 
 * @author Dennis Nolte
 */
class SwitchAnnotationDialog extends StandardDialog<AnnotatedTypeGraph>
{
	private static final long serialVersionUID = -4255861163073697037L;
	
	private AnnotatedTypeGraph graph;
    private List<VxGraph> visis;
    private Set<Node> colouredNodes;
    private Set<Edge> colouredEdges;
    private GraphColor color;
    private boolean simplified;
   
    private JScrollPane scroll;
	private GraphPanel graphView;

	private JComboBox<String> annotationCombo;

    public SwitchAnnotationDialog(Frame owner,AnnotatedTypeGraph graph,Set<Node> colouredNodes,Set<Edge> colouredEdges, GraphColor color, boolean simplified)
    {
        super(owner, ResourceKeys.TIT_SWITCH_ANNOTATION, EnumSet.of(Button.ACCEPT, Button.CANCEL));
        setAcceptKey(ResourceKeys.CMD_APPLY);
        setCancelKey(ResourceKeys.CMD_CANCEL);

        visis = new ArrayList<>();
        this.graph = graph;
        this.colouredNodes = colouredNodes;
        this.colouredEdges = colouredEdges;
        this.color = color;
        this.simplified = simplified;
        
    }
    
    public SwitchAnnotationDialog(Dialog owner,AnnotatedTypeGraph graph,Set<Node> colouredNodes,Set<Edge> colouredEdges, GraphColor color, boolean simplified)
    {
        super(owner, ResourceKeys.TIT_SWITCH_ANNOTATION, EnumSet.of(Button.ACCEPT, Button.CANCEL));
        setAcceptKey(ResourceKeys.CMD_APPLY);
        setCancelKey(ResourceKeys.CMD_CANCEL);

        visis = new ArrayList<>();
        this.graph = graph;
        this.colouredNodes = colouredNodes;
        this.colouredEdges = colouredEdges;
        this.color = color;
        this.simplified = simplified;
        
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected JComponent getDialogComponent()
    {
    	JPanel result = new JPanel();
    	result.setPreferredSize(new Dimension(700,700));
    	result.setLayout(new BorderLayout(0,1));
    	result.setOpaque(true);
    	
    	JPanel margin = new JPanel();
    	margin.setPreferredSize(new Dimension(1,700));
    	
    	
    	result.add(margin, BorderLayout.WEST);
        result.add(createGraphPanel(), BorderLayout.CENTER);
        result.add(createAnnotationCombo(), BorderLayout.NORTH);
        result.setBorder(BorderFactory.createMatteBorder(1, 1, 1, 1, Color.BLACK));
        getContentPane().setBackground(new Color(0,0,0,0));
        getRootPane().setBackground(new Color(0,0,0,0));
        setUndecorated(true);
        setBackground(new Color(0,0,0,0));
        return result;
    }

    
    @Override
    protected AnnotatedTypeGraph parseOptions() throws UserInputException //TODO
    {
        return graph;
    }
	
	private JScrollPane createGraphPanel() {
      
		graphView = new GraphPanel();
        VisiTools.getInstance().getVisualizer().addDefaultStyles(graphView.getDrawer().getStyleMap());
		
        scroll = new JScrollPane(graphView);
        scroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        scroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
        
        return scroll;
	}
	
    /**
     * Creates the combo box displaying annotation names.
     */
    private JPanel createAnnotationCombo()
    {
        DefaultComboBoxModel<String> model = new DefaultComboBoxModel<>(); 
        annotationCombo = new JComboBox<String>(model);
        setAnnotations();
        annotationCombo.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e)
            {
                int index = annotationCombo.getSelectedIndex();
                if (index >= 0) changeAnnotation(index);
            }

        });
        
        JLabel morphismLabel = GuiContext.createLabel(ResourceKeys.LAB_SELECT_ANNOTATION);        

        JPanel comboPanel = new JPanel();
        comboPanel.setLayout(new BoxLayout(comboPanel, BoxLayout.X_AXIS));
        comboPanel.add(Box.createHorizontalStrut(5));
        comboPanel.add(morphismLabel);
        comboPanel.add(Box.createHorizontalStrut(10));
        comboPanel.add(annotationCombo);
        
        return comboPanel;
    }

    /**
     * Changes the annotation displayed by the graph.
     */
    public void setAnnotations()
    {        
        graphView.setEnabled(true);
        annotationCombo.setEnabled(true);

        for (int n=0; n < graph.getNumberOfMultiplicities();n++) {
        	visis.add(null);
            annotationCombo.addItem("Annotation " + (n+1));
        }

        annotationCombo.setSelectedIndex(graph.getActiveMultiplicityIndex());
        changeAnnotation(graph.getActiveMultiplicityIndex());
    }
    
    // ---------------------------------------------------------------------- //
    //                              Informations                              //
    // ---------------------------------------------------------------------- //

	// ---------------------------------------------------------------------- //
    //                              Private Methods                           //
    // ---------------------------------------------------------------------- //
		
    /**
     * Changes the annotation which is displayed in the graph view.
     * 
     * @param index
     *      index of the multiplicity to be displayed
     */
    private void changeAnnotation(int index) 
    {
    	graph.selectMultiplicity(index);
    	VxGraph vx = visis.get(index);
    	if (vx == null) {
    		if(colouredNodes != null && colouredEdges != null) {
    			vx = VisiTools.getInstance().visualizeColoredGraph(graph, colouredNodes, colouredEdges,color,simplified);
    		}
    		else {
    			vx = VisiTools.getInstance().visualizeGraph(graph);    			
    		}
    		
    		visis.set(index, vx);
    	}
    	graphView.setGraph(vx);
    }
}
