/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;

import de.uni_due.inf.ti.dragom.general.ResourceKeys;
import de.uni_due.inf.ti.graph.Rule;
import de.uni_due.inf.ti.swing.GuiContext;
import de.uni_due.inf.ti.swing.StandardDialog;
import de.uni_due.inf.ti.swing.UserInputException;
import de.uni_due.inf.ti.visigraph.VxGraph;
import de.uni_due.inf.ti.visigraph.swing.GraphPanel;

/**
 * Dialog which displays the switch rule dialog.
 * 
 * @author Dennis Nolte
 */
class SwitchRuleDialog extends StandardDialog<Integer>
{
	private static final long serialVersionUID = -4255861163073697037L;
	
	private List<Rule> rules;
	private Rule currentRule;
    private List<VxGraph> visis;
   
    private JScrollPane scroll;
	private GraphPanel ruleView;

	private JComboBox<String> ruleCombo;
	private int activeRuleIndex;

    public SwitchRuleDialog(Frame owner,List<Rule> rules, int activeRuleIndex)
    {
        super(owner, ResourceKeys.TIT_SWITCH_RULE, EnumSet.of(Button.ACCEPT, Button.CANCEL));
        setAcceptKey(ResourceKeys.CMD_APPLY);
        setCancelKey(ResourceKeys.CMD_CANCEL);

        visis = new ArrayList<>();
        this.rules = rules;
        this.activeRuleIndex = activeRuleIndex;
        
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected JComponent getDialogComponent()
    {
    	JPanel result = new JPanel();
    	result.setPreferredSize(new Dimension(1000,680));
    	result.setLayout(new BorderLayout(0,1));
    	result.setOpaque(true);
    	
    	JPanel margin = new JPanel();
    	margin.setPreferredSize(new Dimension(1,680));
    	
    	
    	result.add(margin, BorderLayout.WEST);
        result.add(createRulePanel(), BorderLayout.CENTER);
        result.add(createRuleCombo(), BorderLayout.NORTH);
        result.setBorder(BorderFactory.createMatteBorder(1, 1, 1, 1, Color.BLACK));
        getContentPane().setBackground(new Color(0,0,0,0));
        getRootPane().setBackground(new Color(0,0,0,0));
        setUndecorated(true);
        setBackground(new Color(0,0,0,0));
        return result;
    }

    
    @Override
    protected Integer parseOptions() throws UserInputException
    {
        return ruleCombo.getSelectedIndex();
    }
	
	private JScrollPane createRulePanel() {
      
		ruleView = new GraphPanel();
        VisiTools.getInstance().getVisualizer().addDefaultStyles(ruleView.getDrawer().getStyleMap());
		
        scroll = new JScrollPane(ruleView);
        scroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        scroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
        
        return scroll;
	}
	
    /**
     * Creates the combo box displaying rule names.
     */
    private JPanel createRuleCombo()
    {
        DefaultComboBoxModel<String> model = new DefaultComboBoxModel<>(); 
        ruleCombo = new JComboBox<String>(model);
        setRule();
        ruleCombo.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e)
            {
                int index = ruleCombo.getSelectedIndex();
                if (index >= 0) changeRule(index);
            }

        });
        
        JLabel ruleLabel = GuiContext.createLabel(ResourceKeys.LAB_SELECT_RULE);        

        JPanel comboPanel = new JPanel();
        comboPanel.setLayout(new BoxLayout(comboPanel, BoxLayout.X_AXIS));
        comboPanel.add(Box.createHorizontalStrut(5));
        comboPanel.add(ruleLabel);
        comboPanel.add(Box.createHorizontalStrut(10));
        comboPanel.add(ruleCombo);
        
        return comboPanel;
    }

    /**
     * Changes the rule displayed.
     */
    public void setRule()
    {        
        ruleView.setEnabled(true);
        ruleCombo.setEnabled(true);

        for (int n=0; n < rules.size();n++) {
        	visis.add(null);
            ruleCombo.addItem("Rule " + (n+1));
        }

        ruleCombo.setSelectedIndex(activeRuleIndex);
        changeRule(activeRuleIndex);
    }
    
    // ---------------------------------------------------------------------- //
    //                              Informations                              //
    // ---------------------------------------------------------------------- //

	// ---------------------------------------------------------------------- //
    //                              Private Methods                           //
    // ---------------------------------------------------------------------- //
		
    /**
     * Changes the rule which is displayed in the rule view.
     * 
     * @param index
     *      index of the rule to be displayed
     */
    private void changeRule(int index) 
    {
    	currentRule = rules.get(index);
    	VxGraph vx = visis.get(index);
    	if (vx == null) {
    		vx = VisiTools.getInstance().visualizeRuleDPO(currentRule);
    		visis.set(index, vx);
    	}
    	ruleView.setGraph(vx);
    }
}
