/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.dragom.gui;

import java.awt.Component;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.Spring;
import javax.swing.SpringLayout;
import javax.swing.border.Border;

/**
 * A collection of utility methods for the graphical user interface of DrAGoM
 * 
 * @author Dennis Nolte
 */
public final class Utilities {
		
	/**
	 * Creates a new grid-based layout {@code layout} placing all components of the panel {@code panel}
	 * into {@code cols} different columns.
	 * @param layout the layout which is used for the laying out the panel's components
	 * @param panel the panel whose components will be laid out
	 * @param cols the number of columns of the grid layout
	 * @param xPadding the padding in x-direction
	 * @param yPadding the padding in y-direction
	 */
	protected static void makeGridLayout(SpringLayout layout, JPanel panel, 
			int cols, int xPadding, int yPadding) {
		Spring[] maxWidthSprings = null;
		Component comp = null;
		Component adjustmentComp = null;
		Component eastComp = null;
		Component southComp = null;
		SpringLayout.Constraints cons = null;
	
		if(cols<1) {
			throw new IllegalArgumentException();
		}
		else if(cols==1) {
			makeListLayout(layout, panel, xPadding, yPadding);
		}
		else {
			if(panel.getComponentCount()==0) {
				return;
			}
			
			//Initialize springs for width of the several columns
			maxWidthSprings = new Spring[cols];
			for(int i=0;i<cols;i++) {
				maxWidthSprings[i] = Spring.constant(0);
			}
			
			//Get the maximum width for each column (i.e. the maximum width of each component in a column) 
			for(int i=0;i<panel.getComponentCount();i++) {
				comp = panel.getComponent(i);
				comp = panel.getComponent(i);
				
				cons = layout.getConstraints(comp);
				//Ignore the last component if it is the only one in that row
				if(i%cols!=0 || i!=panel.getComponentCount()-1) {
					maxWidthSprings[i%cols] = Spring.max(maxWidthSprings[i%cols], 
							cons.getWidth());
				}
			}
			
			//Initialize most-east and most-south component (to compute the width and the height of the panel)
			eastComp = panel.getComponent(0);
			southComp = panel.getComponent(0);
			//Lay out the components on the panel
			for(int i=0;i<panel.getComponentCount();i++) {
				comp = panel.getComponent(i);
				//The first component is attached to the left and the top edge of the panel
				if(i==0) {
					layout.putConstraint(SpringLayout.NORTH, comp, yPadding, 
							SpringLayout.NORTH, panel);
					layout.putConstraint(SpringLayout.WEST, comp, xPadding, 
							SpringLayout.WEST, panel);
				}
				//every component which belongs to the first column is adjusted by the most-south component
				//in the row above
				else if(i%cols==0) {
					for(int j=i-cols;j<i;j++) {
						if(layout.getConstraint(SpringLayout.SOUTH, adjustmentComp).getValue() <
								layout.getConstraint(SpringLayout.SOUTH, panel.getComponent(j)).getValue()) {
							adjustmentComp = panel.getComponent(j);
						}
					}
					
					layout.putConstraint(SpringLayout.NORTH, comp, yPadding, 
							SpringLayout.SOUTH, adjustmentComp);
					layout.putConstraint(SpringLayout.WEST, comp, xPadding, 
							SpringLayout.WEST, panel);
				}
				//every other component is adjusted by its left neighbor
				else {
					adjustmentComp = panel.getComponent(i-1);
					layout.putConstraint(SpringLayout.NORTH, comp, 0, 
							SpringLayout.NORTH, adjustmentComp);
					layout.putConstraint(SpringLayout.WEST, comp, xPadding, 
							SpringLayout.EAST, adjustmentComp);
				}
				
				//Set the width of each component of the column to the columns' width
				//except for the last component if it is the only one in that row
				if(i%cols!=0 || i!=panel.getComponentCount()-1) {
					layout.getConstraints(comp).setWidth(maxWidthSprings[i%cols]);
				}
				
				//Check whether the current component lays more to right than the previous
				if(layout.getConstraint(SpringLayout.EAST, eastComp).getValue() < 
						layout.getConstraint(SpringLayout.EAST, comp).getValue()) {
					eastComp = comp;
				}
				
				//Check whether the current component lays more to bottom than the previous
				if(layout.getConstraint(SpringLayout.SOUTH, southComp).getValue() < 
						layout.getConstraint(SpringLayout.SOUTH, comp).getValue()) {
					southComp = comp;
				}
			}
			
			//Set the panel's width and height such that 
			//every component on this panel is visible
			layout.putConstraint(SpringLayout.EAST, panel, xPadding, 
					SpringLayout.EAST, eastComp);
			layout.putConstraint(SpringLayout.SOUTH, panel, yPadding, 
					SpringLayout.SOUTH, southComp);
		}
	}
	
	/**
	 * Sets the border of the component {@code component} to a lowered bevel border with 
	 * title {@code title}.
	 * @param component the component whose border is set
	 * @param title the title of the border
	 */
	protected static void setCompoundBorder(JComponent component, String title) {
		Border loweredbevel = BorderFactory.createLoweredBevelBorder();
		Border titledBorder = BorderFactory.createTitledBorder(loweredbevel, " "+title+" ");
		
		component.setBorder(titledBorder);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Static Methods																*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new list-based layout {@code layout} placing all components of the panel {@code panel}.
	 * @param layout the layout which is used for the laying out the panel's components
	 * @param panel the panel whose components will be laid out
	 * @param xPadding the padding in x-direction
	 * @param yPadding the padding in y-direction
	 */
	private static void makeListLayout(SpringLayout layout, JPanel panel, int xPadding, int yPadding) {
		Spring maxWidthSpring = null;
		Component comp = null;
		Component adjustmentComp = null;
		SpringLayout.Constraints cons = null;		
		
		if(panel.getComponentCount()==0) {
			return;
		}
		
		//Initialize spring for the maximum width of the column
		maxWidthSpring = Spring.constant(0);
		
		//Get the maximum width for each column (i.e. the maximum width of each component in a column) 
		for(int i=0;i<panel.getComponentCount();i++) {
			comp = panel.getComponent(i);
			comp = panel.getComponent(i);
			
			cons = layout.getConstraints(comp);
			maxWidthSpring = Spring.max(maxWidthSpring, cons.getWidth());
		}
		
		//Lay out the components on the panel
		for(int i=0;i<panel.getComponentCount();i++) {
			comp = panel.getComponent(i);
			//The first component is attached to the left and the top edge of the panel
			if(i==0) {
				layout.putConstraint(SpringLayout.NORTH, comp, yPadding, 
						SpringLayout.NORTH, panel);
				layout.putConstraint(SpringLayout.WEST, comp, xPadding, 
						SpringLayout.WEST, panel);
			}
			//every other component is adjusted by its left neighbor
			else {
				adjustmentComp = panel.getComponent(i-1);
				layout.putConstraint(SpringLayout.NORTH, comp, yPadding, 
						SpringLayout.SOUTH, adjustmentComp);
				layout.putConstraint(SpringLayout.WEST, comp, 0, 
						SpringLayout.WEST, adjustmentComp);
			}
			
			//Set the width of each component of the column to the columns' maximum width
			layout.getConstraints(comp).setWidth(maxWidthSpring);
		}
		
		//Set the panel's width and height such that 
		//every component on this panel is visible
		layout.putConstraint(SpringLayout.EAST, panel, xPadding, 
				SpringLayout.EAST, panel.getComponent(panel.getComponentCount()-1));
		layout.putConstraint(SpringLayout.SOUTH, panel, yPadding, 
				SpringLayout.SOUTH, panel.getComponent(panel.getComponentCount()-1));
	}
	
}
