/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.gui;

import java.util.Set;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.dragom.gui.AnnotatedTypeGraphVisualizer.GraphColor;
import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.Morphism;
import de.uni_due.inf.ti.graph.Node;
import de.uni_due.inf.ti.graph.Rule;
import de.uni_due.inf.ti.graph.Rule.DoublePushoutMorphisms;
import de.uni_due.inf.ti.visigraph.GraphDrawer;
import de.uni_due.inf.ti.visigraph.Layouter;
import de.uni_due.inf.ti.visigraph.LevelLayouter;
import de.uni_due.inf.ti.visigraph.SpringLayouter;
import de.uni_due.inf.ti.visigraph.VxGraph;

/**
 * Wrapper class for the visualizer to store certain default options.
 * 
 * @author Dennis Nolte
 */
final class VisiTools
{
    private static VisiTools instance = null;
    
    private AnnotatedTypeGraphVisualizer visualizer;
    private GuiOptions.Layouter layouterStyle;
    private Layouter layouter;
    
    
    /**
     * Returns the application-wide instance of VisiTools.
     */
    public static VisiTools getInstance()
    {
        if (instance == null) {
            instance = new VisiTools();
        }
        return instance;
    }
    
    
    /**
     * Creates a new VisiTools instance.
     */
    private VisiTools()
    {
        visualizer = new AnnotatedTypeGraphVisualizer();
        visualizer.setHyperStyle(false);
        visualizer.setMultiplicityStyle(false);
    }
    

    /**
     * Creates a layouter for layouting hypergraphs.
     */
    private void createLayouter()
    {
        GuiOptions.Layouter style = GuiOptions.Layouter.SPRING;
        if (layouter == null || layouterStyle != style) {
            switch (style) {
                case LEVELED: {
                    LevelLayouter ll = new LevelLayouter();
                    GraphDrawer drawer = new GraphDrawer();
                    visualizer.addDefaultStyles(drawer.getStyleMap());
                    ll.setGraphDrawer(drawer);
                    layouter = ll;
                    break;
                }
                case SPRING: {
                    SpringLayouter sl = new SpringLayouter();
                    GraphDrawer drawer = new GraphDrawer();
                    visualizer.addDefaultStyles(drawer.getStyleMap());
                    sl.setGraphDrawer(drawer);
                    layouter = sl;
                    break;
                }
            }
        }
        layouterStyle = style;
    }

    
    /**
     * Creates a visualization of a graph.
     */
    public VxGraph visualizeGraph(AnnotatedTypeGraph graph) 
    {
        createLayouter();
        VxGraph vx = visualizer.visualizeGraph(graph);
        layouter.layoutGraph(vx);
        return vx;
    }
    

    /**
     * Creates a visualization of a graph.
     */
    public VxGraph visualizeColoredGraph(AnnotatedTypeGraph graph,Set<Node> nodeImage, Set<Edge> edgeImage, GraphColor color, boolean simplified) 
    {
        createLayouter();
        VxGraph vx = visualizer.visualizeColoredGraph(graph,nodeImage,edgeImage, color, simplified);
        layouter.layoutGraph(vx);
        return vx;
    }
    
    /**
     * Creates a visualization of a rule.
     */
    public VxGraph visualizeRuleSPO(Rule rule)
    {
        createLayouter();
        VxGraph vx = null;
        setMultiplicityStyle(false); 
        AnnotatedTypeGraphVisualizer.MorphismVisi visi = visualizer.visualizeMorphism(rule.getCorrespondence(),false,false);
        layouter.layoutGroups(visi.getLeftGroup(), visi.getRightGroup());
        vx = visi.getVxGraph();
        setMultiplicityStyle(true);
        return vx;
    }
    
    /**
     * Creates a visualization of a rule.
     */
    public VxGraph visualizeRuleDPO(Rule rule)
    {
        createLayouter();
        VxGraph vx = null;
        setMultiplicityStyle(false); 
        DoublePushoutMorphisms dpo = rule.getDoublePushoutMorphisms();
        AnnotatedTypeGraphVisualizer.RuleVisi visi = visualizer.visualizeDPO(dpo.getLeftMorphism(),dpo.getRightMorphism());
        layouter.layoutGroups(visi.getLeftGroup(), visi.getMiddleGroup(), visi.getRightGroup());
        vx = visi.getVxGraph();
        setMultiplicityStyle(true);
        return vx;
    }
    
    /**
     * Creates a visualization of a morphism.
     */
    public VxGraph visualizeMatch(Morphism morph,boolean leftMulti,boolean rightMulti)
    {
        createLayouter();
        VxGraph vx = null;
        AnnotatedTypeGraphVisualizer.MorphismVisi visi = visualizer.visualizeMorphism(morph,leftMulti,rightMulti);
        layouter.layoutGroups(visi.getLeftGroup(), visi.getRightGroup());
        vx = visi.getVxGraph();
        setMultiplicityStyle(true);
        return vx;
    }
    
    /**
     * Creates a visualization of a morphism.
     */
    public VxGraph visualizeLegalMorphism(Morphism morph,Set<Node> coloredNodes, Set<Edge> coloredEdges, GraphColor color, boolean simplified)
    {
        createLayouter();
        VxGraph vx = null;
        setMultiplicityStyle(true);
        AnnotatedTypeGraphVisualizer.MorphismVisi visi = visualizer.visualizePostconditionMatch(morph,coloredNodes,coloredEdges,color,simplified);
        layouter.layoutGroups(visi.getLeftGroup(), visi.getRightGroup());
        vx = visi.getVxGraph();
        return vx;
    }

    /**
     * Sets the property whether all graphs should have multiplicities.
     */
    public void setMultiplicityStyle(boolean alwaysMulti)
    {
        visualizer.setMultiplicityStyle(alwaysMulti);
    }

    
    /**
     * Sets the property whether all graphs should be represented in hypergraph
     * form.
     */
    public void setHyperStyle(boolean alwaysHyper)
    {
        visualizer.setHyperStyle(alwaysHyper);
    }
    
    
    /**
     * Returns the GraphVisualizer used by this VisiTools instance.
     */
    public AnnotatedTypeGraphVisualizer getVisualizer()
    {
        return visualizer;
    }
}
