/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.io;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.JFileChooser;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;

import de.uni_due.inf.ti.dragom.gui.FileTypes;
import de.uni_due.inf.ti.dragom.general.ResourceKeys;
import de.uni_due.inf.ti.graph.TransformationSystem;
import de.uni_due.inf.ti.graph.io.ParseException;
import de.uni_due.inf.ti.graph.io.SgfParser;
import de.uni_due.inf.ti.gui.CollectionListModel;
import de.uni_due.inf.ti.swing.GuiContext;

/**
 * Class which shows a file chooser with an additional accessory to choose a
 * single graph transformation system from a SGF file possibly containing more
 * than one system.
 * 
 * @author Dennis Nolte
 */
public class GTSChooser implements PropertyChangeListener
{
    /** The file chooser dialog used. */
    private JFileChooser fileChooser;
    
    /** The accessory of the file chooser which lets the user select one of
        the systems from the selected file. */
    private JPanel accessory;
    
    /** List model of the accessory. */
    private CollectionListModel<String> systems;
    
    /** List control of the accessory. */
    private JList<String> systemList;
    
    /** Map which maps system names to corresponding systems. */
    private Map<String,TransformationSystem> sgfSystems = null;
    
    /** If an IO error occurs, store it here so that an error message can be
    presented when the user clicks Ok. */
    private String error = null;
    
    /** If an IO error occurs, store the title of the error here. */
    private String errorTitle = null;
    
    /**
     * Creates the file chooser accessory.
     */
    private void createAccessory()
    {
        accessory = new JPanel();
        accessory.setLayout(new BorderLayout(5, 5));
        accessory.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
        accessory.add(GuiContext.createLabel(ResourceKeys.LAB_SYSTEMS), BorderLayout.NORTH);
        systems = new CollectionListModel<>();
        systemList = new JList<>(systems);
        JScrollPane scroll = new JScrollPane(systemList);
        scroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        scroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
        accessory.add(scroll, BorderLayout.CENTER);
    }
    
    /**
     * Creates the file chooser. The method {@link createAccessory()} must have
     * been called before calling this method.
     */
    private void createFileChooser()
    {
    	new File("savedata/gts").mkdirs();
        fileChooser = new JFileChooser("savedata/gts");
        fileChooser.setMultiSelectionEnabled(false);
        fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
        fileChooser.setFileFilter(
                FileTypes.getFileFilterForType(FileTypes.FileType.SGF));
        fileChooser.setAccessory(accessory);
        fileChooser.addPropertyChangeListener(this);
        accessory.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
                if (e.getClickCount() >= 2) {
                    fileChooser.approveSelection();
                }
            }
        });
    }
    
    /**
     * Creates a new SystemChooser instance.
     */
    public GTSChooser()
    {
        createAccessory();
        createFileChooser();
    }
    
    /**
     * Sets the current directory of the file chooser.
     * 
     * @param dir
     *      the new current directory
     */
    public void setCurrentDirectory(File dir)
    {
        fileChooser.setCurrentDirectory(dir);
    }
    
    /**
     * Returns the current directory of the file chooser.
     * @return The current directory
     */
    public File getCurrentDirectory()
    {
        return fileChooser.getCurrentDirectory();
    }
    
    /**
     * Shows the dialog. The method returns the transformation system selected
     * by the user, or {@code null} if no transformation system was chosen or
     * an error occurred.
     * 
     * @param parent
     *      parent component of the dialog
     * @return 
     *      the transformation system selected by the user, or {@code null} if
     *      no system was selected or an error occurred
     */
    public TransformationSystem showDialog(Component parent)
    {
        int answer = fileChooser.showOpenDialog(parent);
        if (answer == JFileChooser.APPROVE_OPTION) {
            if (error != null) {
                JOptionPane.showMessageDialog(parent, error, errorTitle, JOptionPane.ERROR_MESSAGE);
                return null;
            }
            else {
                String systemName = systemList.getSelectedValue();
                if (systemName == null) {
                    return null;
                }
                else {
                    return sgfSystems.get(systemName);
                }
            }
        }
        else {
            return null;
        }
    }

    /**
     * Called when a property of the JFileChooser changes.
     */
    @Override
    public void propertyChange(PropertyChangeEvent e)
    {
        if (e.getPropertyName().equals(JFileChooser.SELECTED_FILE_CHANGED_PROPERTY)) {
            systems.clear();
            error = null;
            errorTitle = null;
            File newFile = (File)e.getNewValue();
            if (newFile != null) {
                try {
                    if (sgfSystems == null) sgfSystems = new HashMap<>();
                                       else sgfSystems.clear();
                    List<String> newSystems = new ArrayList<>();
                    FileReader in = new FileReader(newFile);
                    Map<String,Object> sgfObjects = SgfParser.parseSgfFile(in);
                    for (Map.Entry<String,Object> entry : sgfObjects.entrySet()) {
                        Object value = entry.getValue();
                        if (value instanceof TransformationSystem) {
                            String key = entry.getKey();
                            sgfSystems.put(key, (TransformationSystem)value);
                            newSystems.add(key);
                        }
                    }
                    if (newSystems.size() >= 1) {
                        Collections.sort(newSystems);
                        systems.addAll(newSystems);
                        systemList.setEnabled(true);
                        systemList.setSelectedIndex(0);
                    }
                    else {
                        systemList.setEnabled(false);
                    }
                }
                catch (IOException ex) {
                    Throwable cause = ex.getCause();
                    if (cause instanceof ParseException) {
                        errorTitle = GuiContext.getSimpleGuiString(ResourceKeys.TIT_PARSE_ERROR);
                        error = cause.getMessage();
                    }
                    else {
                        errorTitle = GuiContext.getSimpleGuiString(ResourceKeys.TIT_IOERROR); 
                        error = ex.getMessage();
                    }
                    systemList.setEnabled(false);
                }
            }
            else {
                systemList.setEnabled(false);
            }
        }
    }
}
