/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.io;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.dragom.data.AnnotationBounds;
import de.uni_due.inf.ti.dragom.general.AnnotatedTypeGraphException;
import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.GraphElement;
import de.uni_due.inf.ti.graph.Label;
import de.uni_due.inf.ti.graph.Node;

/**
 * This class provides operations for reading graphs from files.
 * 
 * @author Dennis Nolte
 */
public class GXLReader {

	// GXL Tags
	private static final String GXL_TAG_GRAPH = "graph";
	private static final String GXL_TAG_NODE = "node";
	private static final String GXL_TAG_EDGE = "edge";

	// GXL Attributes
	private static final String GXL_ATTRIBUTE_ID = "id";
	private static final String GXL_ATTRIBUTE_NAME = "name";
	private static final String GXL_ATTRIBUTE_SOURCE = "from";
	private static final String GXL_ATTRIBUTE_TARGET = "to";
	
	//GXL Values
	private static final String GXL_VALUE_LABEL = "label";
	private static final String GXL_VALUE_MINBOUND = "min";
	private static final String GXL_VALUE_MAXBOUND = "max";	
	private static final String GXL_VALUE_LIMIT = "limit";
	
	private static final Logger logger = Logger.getLogger(GXLReader.class.getName());

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Static Methods																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Reads a graph from a GXL file and returns it.
	 * 
	 * @param filename the name of the file from which the graph is read
	 * @return the graph read from the file
	 * @throws FileNotFoundException thrown if the file to be read could not be found or if the 
	 * corresponding DTD file could not be found
	 * @throws IOException thrown if an I/O error occurred
	 * @throws AnnotatedTypeGraphException thrown if the parsing went wrong
	 */
	public static AnnotatedTypeGraph readGraph(String filename) throws FileNotFoundException, IOException, AnnotatedTypeGraphException {
		AnnotatedTypeGraph graph = null;
		HashMap<String,Node> nodeMap = new HashMap<String,Node>();
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		
		// Initialize XML-Parser
		// With validation
		factory.setValidating(true);
		
		// Without validation
		/*
		factory.setValidating(false);
		try {
			factory.setFeature("http://xml.org/sax/features/validation", false);
			factory.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd", false);
		} catch (ParserConfigurationException e1) {
			e1.printStackTrace();
		}
		*/
		
		factory.setIgnoringElementContentWhitespace(true);
		DocumentBuilder builder;
		try {
			builder = factory.newDocumentBuilder();
			GXLErrorHandler errorHandler = new GXLErrorHandler();
			builder.setErrorHandler(errorHandler);
			Document document = builder.parse(new File(filename));

			// check whether parsing went okay
			// otherwise return null
			if (errorHandler.getErrorOccured()) {
				return null;
			}
			// open file and get the graph node
			Element root = document.getDocumentElement();
			Element graphElement = (Element) root.getElementsByTagName(GXL_TAG_GRAPH).item(0);
			String graphName = graphElement.getAttribute(GXL_ATTRIBUTE_ID);
			
			NodeList nodeElements = graphElement.getElementsByTagName(GXL_TAG_NODE);
			// get the limit node
			org.w3c.dom.Node limitNode = nodeElements.item(0);
			int limit = 0;
			if(limitNode.getAttributes().getNamedItem(GXL_ATTRIBUTE_ID).getNodeValue().equals(GXL_VALUE_LIMIT)){
				org.w3c.dom.Node attributeNode = limitNode.getFirstChild();
				if(attributeNode.getAttributes().getNamedItem(GXL_ATTRIBUTE_NAME).getNodeValue().equals(GXL_VALUE_LIMIT)){
					limit = Integer.parseInt(attributeNode.getFirstChild().getTextContent());
				}
			}

			graph = new AnnotatedTypeGraph(graphName);
			graph.setLimit(limit);

			int min = 0;
			int max = 0;
			int numberOfMultiplicities = 0;
			HashMap<GraphElement,List<Integer>> minValues = new HashMap<GraphElement,List<Integer>>();
			HashMap<GraphElement,List<Integer>> maxValues = new HashMap<GraphElement,List<Integer>>();
			List<Integer> minList = null;
			List<Integer> maxList = null;
			
			// get regular nodes and remember them in the nodemap
			for (int i = 1; i < nodeElements.getLength(); i++) {
				org.w3c.dom.Node n = nodeElements.item(i);
				String id = n.getAttributes().getNamedItem(GXL_ATTRIBUTE_ID).getNodeValue();
				NodeList attributeNodes = n.getChildNodes();
				for (int a = 0; a<attributeNodes.getLength(); a++) {
					if(attributeNodes.item(a).getAttributes().getNamedItem(GXL_ATTRIBUTE_NAME).getNodeValue().equals(GXL_VALUE_MINBOUND)){
						NodeList multNodes = attributeNodes.item(a).getFirstChild().getChildNodes();
						minList = new LinkedList<Integer>();
						numberOfMultiplicities = multNodes.getLength();
						for (int multa = 0; multa<multNodes.getLength(); multa++) {
							min = Integer.parseInt(multNodes.item(multa).getTextContent());
							minList.add(min);
						}
					}
					if(attributeNodes.item(a).getAttributes().getNamedItem(GXL_ATTRIBUTE_NAME).getNodeValue().equals(GXL_VALUE_MAXBOUND)){
						NodeList multNodes = attributeNodes.item(a).getFirstChild().getChildNodes();
						maxList = new LinkedList<Integer>();
						for (int multa = 0; multa<multNodes.getLength(); multa++) {
							max = Integer.parseInt(multNodes.item(multa).getTextContent());
							maxList.add(max);
						}
					}
				}
				if(minList.get(0) > maxList.get(0)) throw new AnnotatedTypeGraphException();
				Node newNode = graph.addAnnotatedNode(minList.get(0), maxList.get(0));
				nodeMap.put(id, newNode);
				minValues.put(newNode, minList);
				maxValues.put(newNode, maxList);
			}

			// get edges
			NodeList edgeElements = graphElement.getElementsByTagName(GXL_TAG_EDGE);
			for (int j = 0; j < edgeElements.getLength(); j++) {
				org.w3c.dom.Node e = edgeElements.item(j);
				String source = e.getAttributes().getNamedItem(GXL_ATTRIBUTE_SOURCE).getNodeValue();
				String target = e.getAttributes().getNamedItem(GXL_ATTRIBUTE_TARGET).getNodeValue();
				String label = "";
				NodeList edgeAttributeNodes = e.getChildNodes();
				for (int b = 0; b<edgeAttributeNodes.getLength(); b++) {
					if(edgeAttributeNodes.item(b).getAttributes().getNamedItem(GXL_ATTRIBUTE_NAME).getNodeValue().equals(GXL_VALUE_LABEL)){
						label = edgeAttributeNodes.item(b).getFirstChild().getTextContent();
					}
					if(edgeAttributeNodes.item(b).getAttributes().getNamedItem(GXL_ATTRIBUTE_NAME).getNodeValue().equals(GXL_VALUE_MINBOUND)){
						NodeList multEdges = edgeAttributeNodes.item(b).getFirstChild().getChildNodes();
						minList = new LinkedList<Integer>();
						for (int multc = 0; multc<multEdges.getLength(); multc++) {
							min = Integer.parseInt(multEdges.item(multc).getTextContent());
							minList.add(min);
						}
					}
					if(edgeAttributeNodes.item(b).getAttributes().getNamedItem(GXL_ATTRIBUTE_NAME).getNodeValue().equals(GXL_VALUE_MAXBOUND)){
						NodeList multEdges = edgeAttributeNodes.item(b).getFirstChild().getChildNodes();
						maxList = new LinkedList<Integer>();
						for (int multd = 0; multd<multEdges.getLength(); multd++) {
							max = Integer.parseInt(multEdges.item(multd).getTextContent());
							maxList.add(max);
						}
					}
				}
				if(minList.get(0) > maxList.get(0)) throw new AnnotatedTypeGraphException();
				Edge newEdge = graph.addAnnotatedEdge(new Label(label), minList.get(0), maxList.get(0), nodeMap.get(source), nodeMap.get(target));
				minValues.put(newEdge, minList);
				maxValues.put(newEdge, maxList);
			}
			
			graph.addMultiplicity(graph.getCurrentMultiplicity());
			

			HashMap<GraphElement,AnnotationBounds> multiplicity;
				
			for(int multIndex = 1;multIndex < numberOfMultiplicities; multIndex++) {
				multiplicity = new HashMap<GraphElement,AnnotationBounds>();
				
				for(Node dataN: graph.getNodes()) {
					if(minValues.get(dataN).get(multIndex) > maxValues.get(dataN).get(multIndex)) throw new AnnotatedTypeGraphException();
					multiplicity.put(dataN, new AnnotationBounds(minValues.get(dataN).get(multIndex), maxValues.get(dataN).get(multIndex)));
				}
				for(Edge dataE: graph.getEdges()) {
					if(minValues.get(dataE).get(multIndex) > maxValues.get(dataE).get(multIndex)) throw new AnnotatedTypeGraphException();
					multiplicity.put(dataE, new AnnotationBounds(minValues.get(dataE).get(multIndex), maxValues.get(dataE).get(multIndex)));
				}
				
				graph.addMultiplicity(multiplicity);
			}
			
		} catch (ParserConfigurationException e) {
			logger.throwing(GXLReader.class.getName(), "readGraph", e);
			return null;
		} catch (NumberFormatException e) {
			logger.throwing(GXLReader.class.getName(), "readGraph", e);
			return null;
		} catch (SAXException e) {
			logger.throwing(GXLReader.class.getName(), "readGraph", e);
			return null;
		}
		return graph;
	}


	/*------------------------------------------------------------------------------------------*/
	/* 		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	private GXLReader() {
		// Do nothing
	}


}

