/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.io;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Set;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.Node;

/**
 * Writes a graph to a GXL file.
 * 
 * @author Dennis Nolte
 */
public class GXLWriter {
	
	private static final String XML_DECLARATION = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
	private static final String DOCUMENT_TYPE_DECLARATION = "<!DOCTYPE gxl SYSTEM \"http://www.gupro.de/GXL/gxl-1.0.dtd\">";
	
	// GXL Tags
	private static final String GXL_TAG_GRAPH = "graph";
	private static final String GXL_TAG_GXL = "gxl";
	private static final String GXL_TAG_NODE = "node";
	private static final String GXL_TAG_SEQUENCE = "seq";
	private static final String GXL_TAG_INTEGER = "int";
	private static final String GXL_TAG_STRING = "string";
	private static final String GXL_TAG_EDGE = "edge";
	private static final String GXL_TAG_ATTR = "attr";

	// GXL Attributes
	private static final String GXL_ATTRIBUTE_ID = "id";
	private static final String GXL_ATTRIBUTE_NAME = "name";
	private static final String GXL_ATTRIBUTE_SOURCE = "from";
	private static final String GXL_ATTRIBUTE_TARGET = "to";
	
	//GXL Values
	private static final String GXL_VALUE_LABEL = "label";
	private static final String GXL_VALUE_MINBOUND = "min";
	private static final String GXL_VALUE_MAXBOUND = "max";	
	private static final String GXL_VALUE_LIMIT = "limit";
	
	private static GXLWriter writer;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns an instance of the GXL Writer
	 * @return an GXL Writer instance
	 */
	public static GXLWriter getWriter() {
		if(writer==null) {
			writer = new GXLWriter();
		}
		
		return writer;
	}
	
	/**
	 * Writes the graph {@code graph} using the printer {@code out}.
	 * @param graph the graph to be written
	 * @param out the printer to be used
	 */
	private static void writeGraph(AnnotatedTypeGraph graph, PrintWriter out) {
		out.println("<"+GXL_TAG_GRAPH+" " + 
						GXL_ATTRIBUTE_ID + "=\"" + graph.getName() + "\">");

		out.println("<"+GXL_TAG_NODE+" " + GXL_ATTRIBUTE_ID + "=\"" + GXL_VALUE_LIMIT + "\">");
			out.println("<"+GXL_TAG_ATTR+" "+GXL_ATTRIBUTE_NAME+ "=\"" + GXL_VALUE_LIMIT + "\">");
				out.println("<"+GXL_TAG_INTEGER+">" + graph.getLimit() + "</"+GXL_TAG_INTEGER+">");
			out.println("</"+GXL_TAG_ATTR+">");
		out.println("</"+GXL_TAG_NODE+">");
		
		List<Node> nodes = graph.getNodes();
		Set<Edge> edges = graph.getEdges();

		// write nodes to the file
		for (Node n : nodes){
			out.println("<"+GXL_TAG_NODE+" " + GXL_ATTRIBUTE_ID + "=\"n" + n.getId() + "\">");
				out.println("<"+GXL_TAG_ATTR+" "+GXL_ATTRIBUTE_NAME+ "=\"" + GXL_VALUE_MINBOUND + "\">");
					out.println("<"+GXL_TAG_SEQUENCE+">");
						for(int minN: graph.getMinBounds(n)) {
							out.println("<"+GXL_TAG_INTEGER+">" + minN + "</"+GXL_TAG_INTEGER+">");							
						}
					out.println("</"+GXL_TAG_SEQUENCE+">");
				out.println("</"+GXL_TAG_ATTR+">");
				out.println("<"+GXL_TAG_ATTR+" "+GXL_ATTRIBUTE_NAME+ "=\"" + GXL_VALUE_MAXBOUND + "\">");
					out.println("<"+GXL_TAG_SEQUENCE+">");
						for(int maxN: graph.getMaxBounds(n)) {
							out.println("<"+GXL_TAG_INTEGER+">" + maxN + "</"+GXL_TAG_INTEGER+">");							
						}
					out.println("</"+GXL_TAG_SEQUENCE+">");
				out.println("</"+GXL_TAG_ATTR+">");
			out.println("</"+GXL_TAG_NODE+">");
		}

		// write edges to the file
		int eIndex = 0;
		for (Edge e : edges){
			out.println("<"+GXL_TAG_EDGE+" "+ GXL_ATTRIBUTE_ID + "=\"e" + eIndex + "\" " + GXL_ATTRIBUTE_SOURCE + "=\"n" + e.getSource().getId() + "\" " + GXL_ATTRIBUTE_TARGET + "=\"n" + e.getTarget().getId() + "\">");
				out.println("<"+GXL_TAG_ATTR+" "+GXL_ATTRIBUTE_NAME+ "=\"" + GXL_VALUE_LABEL + "\">");
					out.println("<"+GXL_TAG_STRING+">" + e.getLabel().getName() + "</"+GXL_TAG_STRING+">");
				out.println("</"+GXL_TAG_ATTR+">");
				out.println("<"+GXL_TAG_ATTR+" "+GXL_ATTRIBUTE_NAME+ "=\"" + GXL_VALUE_MINBOUND + "\">");
					out.println("<"+GXL_TAG_SEQUENCE+">");
						for(int minE: graph.getMinBounds(e)) {
							out.println("<"+GXL_TAG_INTEGER+">" + minE + "</"+GXL_TAG_INTEGER+">");							
						}
					out.println("</"+GXL_TAG_SEQUENCE+">");
				out.println("</"+GXL_TAG_ATTR+">");
				out.println("<"+GXL_TAG_ATTR+" "+GXL_ATTRIBUTE_NAME+ "=\"" + GXL_VALUE_MAXBOUND + "\">");
					out.println("<"+GXL_TAG_SEQUENCE+">");
						for(int maxE: graph.getMaxBounds(e)) {
							out.println("<"+GXL_TAG_INTEGER+">" + maxE + "</"+GXL_TAG_INTEGER+">");							
						}
					out.println("</"+GXL_TAG_SEQUENCE+">");
				out.println("</"+GXL_TAG_ATTR+">");
			out.println("</"+GXL_TAG_EDGE+">");
			eIndex++;
		}
		
		out.println("</"+GXL_TAG_GRAPH+">");
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	private GXLWriter() {
		//Do nothing
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Writes an annotated type graph to a file
	 * @param graph The annotated type graph to be saved
	 * @param file The file to which the graph is stored
	 * @throws FileNotFoundException thrown if a file was not found
	 * @throws IOException thrown if an I/O error occurred
	 * @throws UnsupportedEncodingException thrown if the encoding isn't supported
	 */
	public void write(AnnotatedTypeGraph graph, File file) 
			throws FileNotFoundException, IOException, UnsupportedEncodingException {		
			
		try (PrintWriter out = new PrintWriter(file, StandardCharsets.UTF_8.name())) {
			// print gxl header
			out.println(XML_DECLARATION);
			out.println(DOCUMENT_TYPE_DECLARATION);
			out.println("<"+GXL_TAG_GXL+">");
			writeGraph(graph, out);
			out.println("</"+GXL_TAG_GXL+">");
			out.flush();
		}
	}
}