/*
 * Copyright (c) 2018-2020, Dennis Nolte, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.   
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.dragom.io;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.Iterator;
import java.util.Properties;

import de.uni_due.inf.ti.dragom.data.AnnotatedTypeGraph;
import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.Node;

/**
 * Writes a graph language abstracted by an annotated type graph.
 * 
 * @author Dennis Nolte
 */
public class LanguagePrinter
{
	 /**
	  * Load the config.properties file.
	  */
	 private final static String cfgProp = "src/de/uni_due/inf/ti/dragom/resources/graphviz.properties";
	 private final static Properties configFile = new Properties() {
	     private final static long serialVersionUID = 1L; {
	         try {
	             load(new FileInputStream(cfgProp));
	         } catch (Exception e) {}
	     }
	 };
	
	 /**
	  * The dir. where temporary files will be created.
	  */
	private static String TEMP_DIR = "tmp";
	
	 /**
	  * Where is your dot program located? It will be called externally.
	  */
	private static String DOT = configFile.getProperty("dotExecutable");
	
	 /**
	  * The source of the graph written in dot language.
	  */
	 private StringBuilder dotSource = new StringBuilder();
	 
	 /**
	  * Boolean determining if all printed annotations should be [1,1]
	  */
	 private boolean alwaysStandardAnnotations;
	
	 /**
	  * Boolean determining if annotations should be printed
	  */
	 private boolean printAnnotations;
	 
	 /**
	  * Constructor: creates a new GraphViz object that will contain
	  * a graph.
	  */
	 public LanguagePrinter() {
		 alwaysStandardAnnotations = false;
		 printAnnotations = true;
	 }
	
	 /**
	  * Returns the graph's source description in dot language.
	  * @return Source of the graph in dot language.
	  */
	 public String getDotSource() {
	     return this.dotSource.toString();
	 }
	
	 /**
	  * Adds a string to the graph's source (without newline).
	  * @param line The line to be added.
	  */
	 public void add(String line) {
	     this.dotSource.append(line);
	 }
	
	 /**
	  * Adds a string to the graph's source (with newline).
	  * @param line The line to be added.
	  */
	 public void addln(String line) {
	     this.dotSource.append(line + "\n");
	 }
	
	 /**
	  * Adds a newline to the graph's source.
	  */
	 public void addln() {
	     this.dotSource.append('\n');
	 }
	
	 public void clearDotSource(){
	     this.dotSource = new StringBuilder();
	 }
	 
	 public void setAlwaysStandardAnnotations(boolean bool){
		 alwaysStandardAnnotations = bool;
	 }
	 
	 public void setPrintAnnotations(boolean bool){
		 printAnnotations = bool;
	 }
	
	 public void printGraph(AnnotatedTypeGraph graph, String languageFolderName, String fileName)
	 {
		 clearDotSource();
	     addln(startDotSource(graph.getName()));
	     add(generateDotString(graph));
	     addln(endDotSource());
	     String type = "svg"; //TODO Create more save options
	     File out = new File("output/languages/" + languageFolderName + "/graphs/" + fileName+"."+ type); 
	     writeGraphToFile( getGraph( getDotSource(), type ), out );
	 }
	 
	 public String generateDotString(AnnotatedTypeGraph graph){
	 	String dotFormat = "";
	 	
	 	Iterator<Node> nodes = graph.getNodes().iterator();   	    	
	 	while(nodes.hasNext()){
	 		Node nextNode = nodes.next();
	 		if(alwaysStandardAnnotations && printAnnotations){
	 			dotFormat += nextNode.getId() + " [label=\"[1,1]\"]; \n"; 			
	 		}
	 		else if(printAnnotations){
	 			dotFormat += nextNode.getId() + " [label=\"" + graph.getBoundLabel(nextNode) + "\"]; \n"; 			
	 		}
	 		else {
	 			dotFormat += nextNode.getId() + " [label=\"" + nextNode.getId() + "\"]; \n";
	 		}
	 	}
	
	 	Iterator<Edge> edges = graph.getEdges().iterator();   	
	 	while(edges.hasNext()){
	 		Edge nextEdge = edges.next();
	 		if(alwaysStandardAnnotations && printAnnotations){
	 			dotFormat += nextEdge.getSource().getId() + " -> " + nextEdge.getTarget().getId() + " [label=\"" + nextEdge.getLabel().getName() + " [1,1]\"]; \n"; 			
	 		}
	 		else if(printAnnotations){
	 			dotFormat += nextEdge.getSource().getId() + " -> " + nextEdge.getTarget().getId() + " [label=\"" + nextEdge.getLabel().getName() + " " + graph.getBoundLabel(nextEdge) + "\"]; \n"; 			
	 		}
	 		else {
	 			dotFormat += nextEdge.getSource().getId() + " -> " + nextEdge.getTarget().getId() + " [label=\"" + nextEdge.getLabel().getName() + "\"]; \n";
	 		}
	 	}
	 	
	 	return dotFormat;
	 }
	 
	 /**
	  * Returns the graph as an image in binary format.
	  * @param dot_source Source of the graph to be drawn.
	  * @param type Type of the output image to be produced, e.g.: gif, dot, fig, pdf, ps, svg, png.
	  * @return A byte array containing the image of the graph.
	  */
	 public byte[] getGraph(String dot_source, String type)
	 {
	     File dot;
	     byte[] img_stream = null;
	
	     try {
	         dot = writeDotSourceToFile(dot_source);
	         if (dot != null)
	         {
	             img_stream = get_img_stream(dot, type);
	             if (dot.delete() == false) 
	                 System.err.println("Warning: " + dot.getAbsolutePath() + " could not be deleted!");
	             return img_stream;
	         }
	         return null;
	     } catch (java.io.IOException ioe) { return null; }
	 }
	
	 /**
	  * Writes the graph's image in a file.
	  * @param img   A byte array containing the image of the graph.
	  * @param file  Name of the file to where we want to write.
	  * @return Success: 1, Failure: -1
	  */
	 public int writeGraphToFile(byte[] img, String file)
	 {
	     File to = new File(file);
	     return writeGraphToFile(img, to);
	 }
	
	 /**
	  * Writes the graph's image in a file.
	  * @param img   A byte array containing the image of the graph.
	  * @param to    A File object to where we want to write.
	  * @return Success: 1, Failure: -1
	  */
	 public int writeGraphToFile(byte[] img, File to)
	 {
	     try {
	         FileOutputStream fos = new FileOutputStream(to);
	         fos.write(img);
	         fos.close();
	     } catch (java.io.IOException ioe) { return -1; }
	     return 1;
	 }
	
	 /**
	  * It will call the external dot program, and return the image in
	  * binary format.
	  * @param dot Source of the graph (in dot language).
	  * @param type Type of the output image to be produced, e.g.: gif, dot, fig, pdf, ps, svg, png.
	  * @return The image of the graph in .gif format.
	  */
	 private byte[] get_img_stream(File dot, String type)
	 {
	     File img;
	     byte[] img_stream = null;
	
	     try {
	         img = File.createTempFile("graph_", "."+type, new File(LanguagePrinter.TEMP_DIR));
	         Runtime rt = Runtime.getRuntime();
	
	         String[] args = {DOT, "-T"+type, "-Gdpi="+72, dot.getAbsolutePath(), "-o", img.getAbsolutePath()};
	         Process p = rt.exec(args);
	
	         p.waitFor();
	
	         FileInputStream in = new FileInputStream(img.getAbsolutePath());
	         img_stream = new byte[in.available()];
	         in.read(img_stream);
	         // Close it if we need to
	         if( in != null ) in.close();
	
	         if (img.delete() == false) 
	             System.err.println("Warning: " + img.getAbsolutePath() + " could not be deleted!");
	     }
	     catch (java.io.IOException ioe) {
	         System.err.println("Error:    in I/O processing of tempfile in dir " + LanguagePrinter.TEMP_DIR+"\n");
	         System.err.println("       or in calling external command");
	         ioe.printStackTrace();
	     }
	     catch (java.lang.InterruptedException ie) {
	         System.err.println("Error: the execution of the external program was interrupted");
	         ie.printStackTrace();
	     }
	
	     return img_stream;
	 }
	
	 /**
	  * Writes the source of the graph in a file, and returns the written file
	  * as a File object.
	  * @param str Source of the graph (in dot language).
	  * @return The file (as a File object) that contains the source of the graph.
	  */
	 private File writeDotSourceToFile(String str) throws java.io.IOException
	 {
	     File temp;
	     try {
	         temp = File.createTempFile("dorrr",".dot", new File(LanguagePrinter.TEMP_DIR));
	         FileWriter fout = new FileWriter(temp);
	         fout.write(str);
	                    BufferedWriter br=new BufferedWriter(new FileWriter(LanguagePrinter.TEMP_DIR + "dotsource.dot"));
	                    br.write(str);
	                    br.flush();
	                    br.close();
	         fout.close();
	     }
	     catch (Exception e) {
	         System.err.println("Error: I/O error while writing the dot source to temp file!");
	         return null;
	     }
	     return temp;
	 }
	
	 /**
	  * Returns a string that is used to start a graph.
	 * @param name The name of the graph.
	  * @return A string to open a graph.
	  */
	 public String startDotSource(String name) {
	     return "digraph " + name + " {";
	 }
	
	 /**
	  * Returns a string that is used to end a graph.
	  * @return A string to close a graph.
	  */
	 public String endDotSource() {
	     return "}";
	 }
	
	 /**
	  * Takes the cluster or subgraph id as input parameter and returns a string
	  * that is used to start a subgraph.
	  * @param clusterid ID of the subgraph
	  * @return A string to open a subgraph.
	  */
	 public String start_subgraph(int clusterid) {
	     return "subgraph cluster_" + clusterid + " {";
	 }
	
	 /**
	  * Returns a string that is used to end a graph.
	  * @return A string to close a graph.
	  */
	 public String end_subgraph() {
	     return "}";
	 }
	
	 /**
	  * Read a DOT graph from a text file.
	  * 
	  * @param input Input text file containing the DOT graph
	  * source.
	  */
	 public void readSource(String input)
	 {
	     StringBuilder sb = new StringBuilder();
	
	     try
	     {
	         FileInputStream fis = new FileInputStream(input);
	         DataInputStream dis = new DataInputStream(fis);
	         BufferedReader br = new BufferedReader(new InputStreamReader(dis));
	         String line;
	         while ((line = br.readLine()) != null) {
	             sb.append(line);
	         }
	         dis.close();
	     } 
	     catch (Exception e) {
	         System.err.println("Error: " + e.getMessage());
	     }
	
	     this.dotSource = sb;
	 }
	 
	 /**
	  * Creates all needed folders to save files in their appropriate language
	  * @param languageName the folder name for the language (usually the type graphs name)
	  */
	 public void createFolderTree(String languageName)
	 {
		new File("output/languages/" + languageName + "/graphs").mkdirs();
	 }
	 
	 /**
	  * Creates all needed folders to save files in their appropriate language
	  */
	 public void createTMPFolder()
	 {
		new File("tmp").mkdirs();
	 }
	 
	 public void generateLanguageHTMLDoc(String languageName, int numberOfLanguageElements)
	 {
		String htmlSource = "";
		new File("output/languages/" + languageName).mkdirs();
		String htmlFileName = "output/languages/" + languageName + "/index.html";
		try {
			new File(htmlFileName).createNewFile();
		} catch (IOException e1) {
			e1.printStackTrace();
		}	 
		
		htmlSource += beginHTMLFile(languageName);
		htmlSource += addHTMLGraphs(numberOfLanguageElements);
		htmlSource += finishHTMLFile();
		
		PrintWriter writer = null;
		try {
			writer = new PrintWriter(htmlFileName, "UTF-8");
			writer.println(htmlSource);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} finally {
			writer.close();
		}
		
		writeCSSFile(languageName);
		 
	 }
	 
	 private String beginHTMLFile(String languageName)
	 {
		 String htmlBeginning = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<html>\n\t<head>\n\t\t<link rel=\"stylesheet\" type=\"text/css\" href=\"main.css\"/>\n\t</head>\n\t<body>\n" +
			"\t\t<div class=\"typeGraph\">\n\t\t\t<h2 class=\"left-heading\">\n\t\t\t\t Annotated Type Graph \n\t\t\t</h2>\n\t\t\t<img src=\"graphs/" + languageName + ".svg\" class=\"input-graph-img\"/>\n" +
			"\t\t</div>\n\t\t<div class=\"divider\"/>\n\t\t<div class=\"graphs\">\n\t\t\t<div class=\"level-div\">\n\t\t\t\t<h1>\n \t\t\t\t\tGraphs in the language \n\t\t\t\t</h1>\n";
		 return htmlBeginning;
	 }
	 
	 private String addHTMLGraphs(int max)
	 {
		 String htmlGraphs = "";
		 for(int i=0;i<=max;i++){
			 htmlGraphs += "\t\t\t\t<div class=\"containing-div\">\n\t\t\t\t\t<img src=\"graphs/" + i + ".svg\" class=\"graph\"/>\n\t\t\t\t</div>\n";
		 }
		 return htmlGraphs;
	 }
	 
	 private String finishHTMLFile()
	 {
		 String htmlEnd = "\t\t\t\t<div style=\"clear:both\"/>\n\t\t\t</div>\n\t\t</div>\n\t</body>\n</html>";
		 return htmlEnd;
	 }
	 
	 private void writeCSSFile(String languageName)
	 {
		String cssFileName = "output/languages/" + languageName + "/main.css";
		try {
			new File(cssFileName).createNewFile();
		} catch (IOException e1) {
			e1.printStackTrace();
		}
		
		PrintWriter writer = null;
		try {
			writer = new PrintWriter(cssFileName, "UTF-8");
			writer.println(
					"body,html{height:100%;width:99%}\n" + 
					".typeGraph{position:absolute;top:30px;left:30px;right:30%;bottom:20%;border:1px dashed grey;text-align:center}\n" +
					".typeGraph > .divider{margin-left:30px;margin-right:30px;margin-top:30px;bottom:65%;border:1px dashed grey}\n" +
					".typeGraph > img{width:80%;height:60%;margin:auto;display:block}\n" +
					".graphs{position:absolute;left:30px;top:80%;text-align:center}\n" +
					".level-div{margin-bottom:30px}\n" +
					".containing-div{float:left;width:250px;height:350px;margin-right:10px;margin-bottom:10px;display:block;border:1px dashed grey;padding:10px;padding-bottom:30px}\n" +
					".graph{max-width:100%;max-height:100%}"				
					);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} finally {
			writer.close();
		}
	 }

 
} 