/* 
 * Copyright (c) 2013-2014, H.J. Sander Bruggink, 
 * Copyright (c) 2013-2014, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   - Neither the names of its copyright holder(s) nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// 2014-09-26 - Created
// 2014-10-21 - Extended to implement Spliterator interface

package de.uni_due.inf.ti.graph.util;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Spliterator;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;

import org.checkerframework.checker.nullness.qual.Nullable;
import org.checkerframework.dataflow.qual.Pure;
import org.checkerframework.dataflow.qual.SideEffectFree;

/**
 * A class which enumerates (non-null) values. 
 * {@code Enumerator}s differ from {@code Iterator}s in that they typically
 * generate the elements on-the-fly.
 * <p>
 * The Enumerator class provides bridges between Java's Collection API and
 * Stream API (new in Java 8):
 * <ul>
 * <li>To obtain an iterator which iterates over the same elements as this
 *     Enumerator, one calls the {@link #iterator()} method.
 * <li>To obtain a list containing the elements enumerated over by this
 *     Enumerator, one calls the {@link #createList()} method.
 * <li>This class implements the {@link Spliterator} interface, so in order
 *     to create a stream, one can write
 *     {@code Stream<...> stream = StreamSupport.stream(enumerator, false);} 
 * </ul>
 * 
 * @author H.J. Sander Bruggink
 */
public abstract class Enumerator<T> implements Iterable<T>, Spliterator<T>
{
    private @Nullable Iterator<T> myIterator = null;
    
    /**
     * Returns the next element in the enumeration, or {@code null} if there
     * is no next element.
     * 
     * @return
     *      the next element of the enumeration
     */
    public abstract @Nullable T next();
    
    /**
     * Applies an action on the next element, if available.
     * 
     * @param action
     *      action which is applied to the next element of this Enumerator
     * @return 
     *      {@code true} if a next element was available and the action was
     *      applied to it, or {@code false} if a next element was not 
     *      available
     * @throws NullPointerException
     *      if {@code action} is {@code null}
     */
    @Override
    public boolean tryAdvance(Consumer<? super T> action)
    {
        if (action == null) throw new NullPointerException();
        
        T ob = next();
        if (ob != null) {
            action.accept(ob);
            return true;
        }
        else return false;
    }
    
    /**
     * Tries to split this Spliterator. This default method returns 
     * {@code null}.
     */
    @Override
    public @Nullable Spliterator<T> trySplit()
    {
        return null;
    }
    
    /**
     * Returns the characteristics of this Enumerator. This default method
     * returns {@code IMMUTABLE | NONNULL}; subclasses with other
     * characteristics should override this method.
     */
    @Override@Pure
    public int characteristics()
    {
        return IMMUTABLE | NONNULL;
    }
    
    /**
     * Returns the estimated number of elements in this Enumerator.
     * This default implementation returns {@link Long#MAX_VALUE}.
     */
    @Override@Pure
    public long estimateSize()
    {
        return Long.MAX_VALUE;
    }
    
    /**
     * Returns an Enumerator which applies a function to every element of
     * this Enumerator. The returned Enumerator is linked to this one:
     * calling the {@link #next()} method on either on will cause both
     * enumerators to proceed to the next element.
     * 
     * @param func
     *      function which is applied to each element
     * @return
     *      Enumerator which applies the function {@code func} to each element
     *      of this Enumerator
     * @throws NullPointerException
     *      if {@code func} is {@code null}
     */
    public final <S> Enumerator<S> map(final Function<? super T, ? extends S> func)
    {
        if (func == null) throw new NullPointerException();
        
        return new Enumerator<S>() {
            @Override
            public @Nullable S next()
            {
                T ob = Enumerator.this.next();
                if (ob == null) {
                    return null;
                }
                else {
                    return func.apply(ob);
                }
            }
        };
    }
    
    /**
     * Returns a Enumerator which enumerates all elements of this Enumerator
     * that satisfy a given condition. The returned Enumerator is linked to
     * this Enumerator: if {@link #next()} is called on the returned 
     * Enumerator, {@code next()} is also called on or more times on this
     * Enumerator and vice versa.
     * 
     * @param pred
     *      the predicate which tells which elements of this Enumerator
     *      are to be contained in the new Enumerator
     * @return
     *      Enumerator which returns all elements of this Enumerator for
     *      which the given predicate holds
     * @throws NullPointerException
     *      if {@code pred} is {@code null}
     */
    public final Enumerator<T> filter(final Predicate<? super T> pred)
    {
        if (pred == null) throw new NullPointerException();
        
        return new Enumerator<T>() {
            @Override
            public @Nullable T next()
            {
                T ob = Enumerator.this.next();
                while (ob != null && !pred.test(ob)) {
                    ob = Enumerator.this.next();
                }
                return ob;
            }
        };
    }

    /**
     * Returns an Iterator which returns the same elements in the same order
     * as this Enumeration.
     * <p>
     * It results in undefined behavior if both the returned Iterator and this
     * Enumerator's {@link #next()} methods are used. If this method is
     * called twice, the same Iterator object is returned.
     */
    @Override@Pure
    public final Iterator<T> iterator()
    {
        if (myIterator == null) myIterator = new EnumeratorIterator();
        return myIterator;
    }
    
    /**
     * Returns a list which contains the rest of elements enumerated by this
     * Enumerator.
     * This method keeps calling {@link #next()} until it returns {@code null},
     * and collects all returned objects in a list. Finally, this list is
     * returned.
     * <p>
     * After calling this method, subsequent calls to {@link next()} will
     * return {@code null}.
     * <p>
     * <em>Warning.</em> If this iterator produces an infinite stream of 
     * objects, this method will not return (or rather, it will cause an 
     * {@link OutOfMemoryError}).
     * 
     * @return
     *      a list containing the rest of the elements iterator by this 
     *      Enumerator
     */
    @SideEffectFree
    public final List<T> createList()
    {
        ArrayList<T> result = new ArrayList<>();
        
        @Nullable T el = next();
        while (el != null) {
            result.add(el);
            el = next();
        }
        
        result.trimToSize();
        return result;
    }
    
    /**
     * Bridge class between Enumerator and Iterator.
     */
    private class EnumeratorIterator implements Iterator<T>
    {
        private @Nullable T nextObject = null;

        @Override@Pure
        public boolean hasNext()
        {
            if (nextObject == null) {
                nextObject = Enumerator.this.next();
            }
            return (nextObject != null);
        }

        @Override
        public T next()
        {
            if (nextObject == null) {
                T ob = Enumerator.this.next();
                if (ob == null) throw new NoSuchElementException();
                return ob;
            }
            else {
                T ob = nextObject;
                nextObject = null;
                return ob;
            }
        }

        @Override
        public void remove()
        {
            throw new UnsupportedOperationException();
        }
    }
}
