/* 
 * Copyright (c) 2013-2014, H.J. Sander Bruggink, University of Duisburg-Essen
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   - Neither the names of its copyright holder(s) nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.graph.visual;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.uni_due.inf.ti.graph.Edge;
import de.uni_due.inf.ti.graph.ExplicitTransitionSystem;
import de.uni_due.inf.ti.graph.Graph;
import de.uni_due.inf.ti.graph.Morphism;
import de.uni_due.inf.ti.graph.Node;
import de.uni_due.inf.ti.graph.Transition;
import de.uni_due.inf.ti.visigraph.ArrowHead;
import de.uni_due.inf.ti.visigraph.Style;
import de.uni_due.inf.ti.visigraph.StyleMap;
import de.uni_due.inf.ti.visigraph.VxEdge;
import de.uni_due.inf.ti.visigraph.VxComponent;
import de.uni_due.inf.ti.visigraph.VxExplicitNode;
import de.uni_due.inf.ti.visigraph.VxGraph;
import de.uni_due.inf.ti.visigraph.VxGroup;
import de.uni_due.inf.ti.visigraph.VxLabelNode;
import de.uni_due.inf.ti.visigraph.VxNode;
import de.uni_due.inf.ti.visigraph.Shape;

/**
 * Instances of this class convert Graph instances to VxGraph instances.
 *
 * @author Dennis Nolte
 */
public class GraphVisualizer
{
    /**
     * Class used to return rule visualizations.
     */
    public static class RuleVisi 
    {
        private VxGraph visi;
        private VxGroup leftGroup;
        private VxGroup rightGroup;
        
        RuleVisi(VxGraph visi, VxGroup leftGroup, VxGroup rightGroup)
        {
            this.visi = visi;
            this.leftGroup = leftGroup;
            this.rightGroup = rightGroup;
        }
        
        public VxGraph getVxGraph()
        {
            return visi;
        }
        
        public VxGroup getLeftGroup()
        {
            return leftGroup;
        }
        
        public VxGroup getRightGroup()
        {
            return rightGroup;
        }
    }
    
    
    /** 
     * Represents whether added objects are created, deleted or neither. 
     * (This is only relevant for flat rule visualizations.)
     */
    public enum AddMode {
        NORMAL,
        ADD_CREATED,
        ADD_DELETED
    }
    
    
    public static String nodeStyleName                  = "node";
    public static String createdNodeStyleName           = "node.created";
    public static String deletedNodeStyleName           = "node.deleted";
    public static String hyperedgeStyleName             = "hyperedge";
    public static String createdHyperedgeStyleName      = "hyperedge.created";
    public static String deletedHyperedgeStyleName      = "hyperedge.deleted";
    public static String edgeStyleName                  = "edge";
    public static String createdEdgeStyleName           = "edge.created";
    public static String deletedEdgeStyleName           = "edge.deleted";
    public static String labelStyleName                 = "label";
    public static String createdLabelStyleName          = "label.created";
    public static String deletedLabelStyleName          = "label.deleted"; 
    public static String connectionStyleName            = "connection";
    public static String createdConnectionStyleName     = "connection.created";
    public static String deletedConnectionStyleName     = "connection.deleted";
    public static String arityStyleName                 = "arity";
    public static String createdArityStyleName          = "arity.created";
    public static String deletedArityStyleName          = "arity.deleted";
    public static String morphismStyleName              = "morphism";
    public static String stateStyleName                 = "state";
    public static String transitionStyleName            = "transition";
    
    
    // Properties //
    
    /** Whether or not graphs are always visualized using hypergraph style. */
    private boolean hyper;
    
    /** Whether or not nodes are labeled with id numbers. */
    private boolean nodeLabels;
    
    /** Whether or not edges are labeled. */
    private boolean edgeLabels;
    
    /** Bounds for the positions of new elements. */
    private Rectangle2D bounds;
    
    
    /**
     * Creates a new GraphVisualizer instance.
     */
    public GraphVisualizer()
    {
        hyper = false;
        nodeLabels = true;
        edgeLabels = true;
        
        bounds = new Rectangle2D.Double(0, 0, 400, 400);
    }
    
    
    /**
     * Adds default styles to the style names of this GraphVisualizer.
     * For each style name, this method determines if a style is already set
     * in the style map; if not, a suitable default is added to the style map.
     * 
     * @param styleMap
     *      style map to add default style to
     */
    public void addDefaultStyles(StyleMap styleMap)
    {
        Color white = Color.WHITE;
        Color gray = new Color(.9f, .9f, .9f);
        Color midGray = new Color(.6f, .6f, .6f);
        Color darkGray = new Color(.3f, .3f, .3f);
        Color red = new Color(1f, .6f, .6f);
        Color midRed = new Color(.75f, .45f, .45f);
        Color darkRed = new Color(.5f, .3f, .3f);
        Color blue = new Color(.6f, .6f, 1f);
        Color midBlue = new Color(.45f, .45f, .75f);
        Color darkBlue = new Color(.3f, .3f, .5f);
        
        Style generalNodeStyle = new Style();
        generalNodeStyle.setFilled(true);
        generalNodeStyle.setStroke(Style.getDefaultStroke(Style.LineWidth.MEDIUM, 
                Style.DashType.SOLID));

        Style nodeStyle = styleMap.get(nodeStyleName);
        if (nodeStyle == null) {
            nodeStyle = generalNodeStyle.createChild();
            nodeStyle.setLineColor(darkGray);
            nodeStyle.setFillColor(gray);
            nodeStyle.setFillColor(true, midGray);
            nodeStyle.setLabelColor(darkGray);
            nodeStyle.setLabelColor(true, white);
            styleMap.put(nodeStyleName, nodeStyle);
        }
        
        Style createdNodeStyle = styleMap.get(createdNodeStyleName);
        if (createdNodeStyle == null) {
            createdNodeStyle = nodeStyle.createChild();
            createdNodeStyle.setLineColor(darkBlue);
            createdNodeStyle.setLabelColor(darkBlue);
            createdNodeStyle.setFillColor(blue);
            createdNodeStyle.setFillColor(true, midBlue);
            styleMap.put(createdNodeStyleName, createdNodeStyle);
        }
        
        Style deletedNodeStyle = styleMap.get(deletedNodeStyleName);
        if (deletedNodeStyle == null) {
            deletedNodeStyle = nodeStyle.createChild();
            deletedNodeStyle.setLineColor(darkRed);
            deletedNodeStyle.setLabelColor(darkRed);
            deletedNodeStyle.setFillColor(red);
            deletedNodeStyle.setFillColor(true, midRed);
            styleMap.put(deletedNodeStyleName, deletedNodeStyle);
        }
        
        Style hyperedgeStyle = styleMap.get(hyperedgeStyleName);
        if (hyperedgeStyle == null) {
            hyperedgeStyle = generalNodeStyle.createChild();
            hyperedgeStyle.setLineColor(darkGray);
            hyperedgeStyle.setFillColor(gray);
            hyperedgeStyle.setFillColor(true, midGray);
            hyperedgeStyle.setShape(Shape.createRectangle(40, 20));
            hyperedgeStyle.setFlexibleShape(true);
            generalNodeStyle.setStroke(Style.getDefaultStroke(Style.LineWidth.MEDIUM, 
                    Style.DashType.SOLID));
            styleMap.put(hyperedgeStyleName, hyperedgeStyle);
        }
        
        Style createdHyperedgeStyle = styleMap.get(createdHyperedgeStyleName);
        if (createdHyperedgeStyle == null) {
            createdHyperedgeStyle = hyperedgeStyle.createChild();
            createdHyperedgeStyle.setLineColor(darkBlue);
            createdHyperedgeStyle.setLabelColor(darkBlue);
            createdHyperedgeStyle.setFillColor(blue);
            createdHyperedgeStyle.setFillColor(true, midBlue);
            styleMap.put(createdHyperedgeStyleName, createdHyperedgeStyle);
        }
        
        Style deletedHyperedgeStyle = styleMap.get(deletedHyperedgeStyleName);
        if (deletedHyperedgeStyle == null) {
            deletedHyperedgeStyle = hyperedgeStyle.createChild();
            deletedHyperedgeStyle.setLineColor(darkRed);
            deletedHyperedgeStyle.setLabelColor(darkRed);
            deletedHyperedgeStyle.setFillColor(red);
            deletedHyperedgeStyle.setFillColor(true, midRed);
            styleMap.put(deletedHyperedgeStyleName, deletedHyperedgeStyle);
        }
        
        Style edgeStyle = styleMap.get(edgeStyleName);
        if (edgeStyle == null) {
            edgeStyle = new Style();
            edgeStyle.setLineColor(midGray);
            edgeStyle.setLineColor(true, darkGray);
            edgeStyle.setFrontArrow(ArrowHead.newTriangleArrowHead(8, 8));
            edgeStyle.setStroke(Style.getDefaultStroke(Style.LineWidth.MEDIUM, 
                    Style.DashType.SOLID));
            styleMap.put(edgeStyleName, edgeStyle);
        }
        
        Style createdEdgeStyle = styleMap.get(createdEdgeStyleName);
        if (createdEdgeStyle == null) {
            createdEdgeStyle = edgeStyle.createChild();
            createdEdgeStyle.setLineColor(blue);
            createdEdgeStyle.setLineColor(true, midBlue);
            styleMap.put(createdEdgeStyleName, createdEdgeStyle);
        }

        Style deletedEdgeStyle = styleMap.get(deletedEdgeStyleName);
        if (deletedEdgeStyle == null) {
            deletedEdgeStyle = edgeStyle.createChild();
            deletedEdgeStyle.setLineColor(red);
            deletedEdgeStyle.setLineColor(true, midRed);
            styleMap.put(deletedEdgeStyleName, deletedEdgeStyle);
        }
        
        Style labelStyle = styleMap.get(labelStyleName);
        if (labelStyle == null) {
            labelStyle = new Style();
            labelStyle.setShape(Shape.createRectangle(10, 5));
            labelStyle.setFlexibleShape(true);
            labelStyle.setDrawn(false);
            labelStyle.setFilled(false);
            labelStyle.setLabelColor(midGray);
            labelStyle.setLabelColor(true, darkGray);
            styleMap.put(labelStyleName, labelStyle);
        }
        
        Style createdLabelStyle = styleMap.get(createdLabelStyleName);
        if (createdLabelStyle == null) {
            createdLabelStyle = labelStyle.createChild();
            createdLabelStyle.setLabelColor(midBlue);
            createdLabelStyle.setLabelColor(true, darkBlue);
            styleMap.put(createdLabelStyleName, createdLabelStyle);
        }

        Style deletedLabelStyle = styleMap.get(deletedLabelStyleName);
        if (deletedLabelStyle == null) {
            deletedLabelStyle = labelStyle.createChild();
            deletedLabelStyle.setLabelColor(midRed);
            deletedLabelStyle.setLabelColor(true, darkRed);
            styleMap.put(deletedLabelStyleName, deletedLabelStyle);
        }

        Style connectionStyle = styleMap.get(connectionStyleName);
        if (connectionStyle == null) {
            connectionStyle = new Style();
            connectionStyle.setLineColor(midGray);
            connectionStyle.setLineColor(true, darkGray);
            styleMap.put(connectionStyleName, connectionStyle);
        }
        
        Style createdConnectionStyle = styleMap.get(createdConnectionStyleName);
        if (createdConnectionStyle == null) {
            createdConnectionStyle = connectionStyle.createChild();
            createdConnectionStyle.setLineColor(midBlue);
            createdConnectionStyle.setLineColor(true, darkBlue);
            styleMap.put(createdConnectionStyleName, createdConnectionStyle);
        }

        Style deletedConnectionStyle = styleMap.get(deletedConnectionStyleName);
        if (deletedConnectionStyle == null) {
            deletedConnectionStyle = connectionStyle.createChild();
            deletedConnectionStyle.setLineColor(midRed);
            deletedConnectionStyle.setLineColor(true, darkRed);
            styleMap.put(deletedConnectionStyleName, deletedConnectionStyle);
        }
        
        Style morphismStyle = styleMap.get(morphismStyleName);
        if (morphismStyle == null) {
            morphismStyle = new Style();
            morphismStyle.setStroke(Style.getDefaultStroke(
                    Style.LineWidth.NORMAL, Style.DashType.DASHED));
            morphismStyle.setLineColor(Color.GREEN);
            morphismStyle.setFrontArrow(ArrowHead.newLineArrowHead(6, 6));
            styleMap.put(morphismStyleName, morphismStyle);
        }
        
        Style stateStyle = styleMap.get(stateStyleName);
        if (stateStyle == null) {
            stateStyle = new Style();
            stateStyle.setFilled(true);
            stateStyle.setFillColor(false, gray);
            stateStyle.setFillColor(true, midGray);
            stateStyle.setLabelColor(false, gray);
            stateStyle.setLabelColor(true, white);
            stateStyle.setShape(Shape.createRoundedRectangle(15, 15, 5));
            stateStyle.setFlexibleShape(true);
            styleMap.put(stateStyleName, stateStyle);
        }
        
        Style transitionStyle = styleMap.get(transitionStyleName);
        if (transitionStyle == null) {
            transitionStyle = new Style();
            transitionStyle.setFrontArrow(ArrowHead.newLineArrowHead(6, 6));
            styleMap.put(transitionStyleName, transitionStyle);
        }
    }

    
    /**
     * Sets the bounds of the generated graphs. The generated graphs will have
     * random positions within these bounds.
     * 
     * @param x
     *      x-coordinate of the top-left corner of the bounding rectangle
     * @param y
     *      y-coordinate of the top-left corner of the bounding rectangle
     * @param width
     *      width of the bounding rectangle
     * @param height
     *      height of the bounding rectangle
     */
    public void setBounds(double x, double y, double width, double height)
    {
        bounds.setRect(x, y, width, height);
    }

    
    /**
     * Sets the bounds of the generated graphs to a rectangle. The generated 
     * graphs will have random positions within these bounds.
     * 
     * @param rect
     *      the new bounding rectangle
     */
    public void setBounds(Rectangle2D rect)
    {
        bounds.setRect(rect);
    }
    
    
    /**
     * Returns the bounds of the generated graphs. If the second parameter is
     * non-null, it is modified to hold the bounds and returned. If it is
     * null, a new {@link Rectangle2D} object holding the bounds is created and
     * returned. 
     * 
     * @param result
     *      buffer to hold the return value
     * @return
     *      rectangle in which new elements are created by this
     *      GraphVisualizer
     */
    public Rectangle2D getBounds(Rectangle2D result)
    {
        if (result == null) {
            result = new Rectangle2D.Double();
        }
        result.setRect(bounds);
        return result;
    }

    
    /**
     * Returns the bounds of the generated graphs. 
     * 
     * @param result
     *      buffer to hold the return value
     * @return
     *      rectangle in which new elements are created by this
     *      GraphVisualizer
     */
    public Rectangle2D getBounds()
    {
        return getBounds(null);
    }

    
    /**
     * Sets whether this GraphVisualizer will always generated hypergraph
     * representations. If {@code true}, the GraphVisualizer will generate
     * hypergraph representations of graphs which contain only binary edges.
     * 
     * @param h
     *      whether this GraphVisualizer will always generate hypergraph
     *      representations
     */
    public void setHyperStyle(boolean h)
    {
        hyper = h;
    }
    
    
    /**
     * Returns whether this GraphVisualizer will always generate hypergraph
     * representations.
     * 
     * @see #setHyperStyle(boolean)
     */
    public boolean hasHyperStyle()
    {
        return hyper;
    }
    
    
    /**
     * Sets whether graph visualizations have node labels. If set to true,
     * graph visualizations created afterwards will have the nodes' id numbers
     * as labels.
     * <p>
     * The default value is {@code true}.
     * 
     * @param nodeLabels
     *      whether or not nodes will have their id numbers as labels
     * @see #hasNodeLabels()
     */
    public void setNodeLabels(boolean nodeLabels)
    {
        this.nodeLabels = nodeLabels;
    }
    

    /**
     * Determines whether graph visualizations have node labels.
     *
     * @return 
     *      whether graph visualizations created by this GraphVisualizer have
     *      node labels
     * @see #setNodeLabels(boolean)
     */
    public boolean hasNodeLabels()
    {
        return nodeLabels;
    }
    
    
    /**
     * Sets whether graph visualizations have edge labels. If set to 
     * {@code true}, edges in graph visualization created with this 
     * GraphVisualizer afterwards will have labels. Since edge labels are 
     * semantically significant in the graphs modelled by {@link Graph Graph} 
     * objects, setting this property to {@code false} is mainly useful if it
     * is known that only one label can occur in a graph. 
     * <p>
     * The default value of this property is {@code true}.
     *  
     * @param edgeLabels
     *      whether or not edges will have labels
     * @see #hasEdgeLabels()
     */
    public void setEdgeLabels(boolean edgeLabels)
    {
        this.edgeLabels = edgeLabels;
    }
    
    
    /**
     * Determines whether graph visualizations have edge labels.
     * 
     * @return
     *      whether graph visualizations created by this GraphVisualizer have
     *      edge labels
     * @see #setEdgeLabels(boolean)
     */
    public boolean hasEdgeLabels()
    {
        return edgeLabels;
    }
    
    
    /**
     * Worker method to visualize a graph. This method is called by various
     * other methods which need to visualize a graph in different situations.
     * 
     * @param input
     *      input graph
     * @param output
     *      output VxGraph; the visual representation of graph is added to 
     *      this
     * @param hyper
     *      whether to use hypergraph style
     * @param nodeMap
     *      map where node connections are stored; should be a newly created
     *      empty map or {@code null}
     * @param edgeMap
     *      map where edge connections are stored; should be a newly created
     *      empty map or {@code null}
     * @param bounds
     *      bounds of the graph; if {@code null}, the bounds from the
     *      bounds attribute are used
     * @param group
     *      group where newly added nodes and edges are added to; 
     *      can be {@code null}
     */
    private void visualizeGraph(
            Graph input, 
            VxGraph output, 
            boolean hyper,
            Map<Node,VxNode> nodeMap,
            Map<Edge,VxComponent> edgeMap,
            Rectangle2D bounds,
            VxGroup group)
    {
        if (bounds == null) bounds = this.bounds;
        if (nodeMap == null) nodeMap = new HashMap<>();
        
        for (Node node : input.getNodes()) {
            Point2D pos = randomPos(bounds);
            VxNode vxn = output.addNode(
                    pos.getX(), 
                    pos.getY(), 
                    nodeLabels ? Integer.toString(node.getId()) : null,
                    nodeStyleName);
            vxn.setReference(node);
            nodeMap.put(node, vxn);
            if (group != null) group.add(vxn);
        }
        
        for (Edge edge : input.getEdges()) {
            List<Node> nodes = edge.getNodes();
            if (hyper) {
                VxNode[] vxs = new VxNode[nodes.size()];
                for (int i = 0; i < nodes.size(); i++) {
                    vxs[i] = nodeMap.get(nodes.get(i));
                }
                VxNode hyperEdge = output.addNode(
                        getBarycenter(vxs),
                        edgeLabels ? edge.getLabel().getName() : null,
                        hyperedgeStyleName);
                hyperEdge.setReference(edge);
                if (edgeMap != null) edgeMap.put(edge, hyperEdge);
                if (group != null) group.add(hyperEdge);
                for (int i = 0; i < vxs.length; i++) {
                    VxEdge con = output.addEdge(
                                    hyperEdge,
                                    vxs[i],
                                    connectionStyleName);
                    VxLabelNode label = con.addLabel(Integer.toString(i));
                    label.setStyle(labelStyleName);
                    label.setPositionOnEdge(.3);
                    if (group != null) group.add(con);
                }
            }
            else {
                VxNode source = nodeMap.get(nodes.get(0));
                VxNode target = nodeMap.get(nodes.get(1));
                assert (source != null && target != null);
                VxEdge vxe = output.addEdge(
                        source, 
                        target, 
                        edgeStyleName);
                if (edgeLabels) {
                	VxLabelNode label = vxe.addLabel(edge.getLabel().getName());
                    label.setStyle(labelStyleName);    
                }
                vxe.setReference(edge);
                if (edgeMap != null) edgeMap.put(edge, vxe);
                if (group != null) group.add(vxe);
            }
        }
    }
    
    
    /**
     * Creates a VxGraph which visualizes a graph.
     * <p>
     * If either {@code getHyperStyle() == true}, or this graph contains
     * non-binary edges, the graph is visualized using a hypergraph 
     * representation; otherwise an ordinary graph representation is used.
     * <p>
     * The nodes and hyperedges of the graphical representation are put on
     * arbitrary positions, whereas all edges will be straight lines from the 
     * source to the target node. In most cases, it will be necessary to layout
     * the resulting graphical representation, either manually or with one
     * of the graph layout algorithm provided with the VisiGraph library.
     * <p>
     * The result visualization uses the following style names:
     * <ul>
     * <li><tt>nodeStyleName</tt>:
     *     the style for the nodes of the graph
     * <li><tt>edgeStyleName</tt>:
     *     the style for the edges in the graph (only in ordinary graph
     *     representations)
     * <li><tt>hyperedgeStyleName</tt>:
     *     the style for hyperedges in the graph (only in hypergraph
     *     representations)
     * <li><tt>connectionStyleName</tt>:
     *     the style for connections between hyperedges and nodes (only in 
     *     hypergraph representations)
     * <li><tt>labelStyleName</tt>:
     *     the style of labels of edges and connections 
     * </ul>  
     * 
     * @param graph
     *      the graph of which a visualization is generated
     * @return
     *      new VxGraph instance which visualizes the graph
     * @throws NullPointerException
     *      if {@code graph} is {@code null}
     */
    public VxGraph visualizeGraph(Graph graph)
    {
        if (graph == null) {
            throw new NullPointerException();
        }
        
        VxGraph result = new VxGraph();
        visualizeGraph(
                graph, 
                result, 
                hyper || !graph.isBinary(), 
                null, null, null, null);
        return result;
    }
    
    
    /**
     * Creates a VxGraph which visualizes a rule. The visual representation
     * will consist of the left-hand side and the right-hand side with 
     * explicit morphism arrows between nodes which are mapped to one another
     * by the correspondence morphism. The parts of the representation 
     * belonging to the left-hand side and right-hand side are added to the
     * {@code leftGroup} and {@code rightGroup} groups, respectively.
     * <p>
     * If either {@code getHyperStyle() == true}, this rule contains
     * non-binary edges, or the correspondence morphism maps at least one edge
     * of the left-hand side to one edge in the right-hand side, the graph is 
     * visualized using a hypergraph representation; otherwise an ordinary
     * graph representation is used.
     * <p>
     * The nodes and hyperedges of the graphical representation are put on
     * random positions, whereas all edges will be straight lines from the 
     * source to the target node. In most cases, it will be necessary to layout
     * the resulting graphical representation, either manually or with one
     * of the graph layout algorithm provided with the VisiGraph library.
     * <p>
     * The resulting visualization uses the following style names:
     * <ul>
     * <li><tt>nodeStyleName</tt>:
     *     the style for the nodes of the graph
     * <li><tt>edgeStyleName</tt>:
     *     the style for the edges in the graph (only in ordinary graph
     *     representations)
     * <li><tt>hyperedgeStyleName</tt>:
     *     the style for hyperedges in the graph (only in hypergraph
     *     representations)
     * <li><tt>connectionStyleName</tt>:
     *     the style for connections between hyperedges and nodes (only in 
     *     hypergraph representations)
     * <li><tt>labelStyleName</tt>:
     *     the style of labels of edges and connections 
     * <li><tt>morphismStyleName</tt> (edge style):
     *     the style for the morphism arrows
     * </ul>  
     * 
     * @param morph
     *      the morphism of which a visualization is generated
     * @return
     *      RuleVisi which contains a new VxGraph instance visualizing the 
     *      morphism, and the groups which contain the nodes and edges of the 
     *      domain and codomain respectively
     * @throws NullPointerException
     *      if {@code morph} is {@code null}
     */
    public RuleVisi visualizeMorphism(Morphism morph)
    {
        if (morph == null) {
            throw new NullPointerException();
        }
        
        Graph left = morph.getDomain();
        Graph right = morph.getCodomain();
        boolean edgesMapped = !morph.getEdgeMap().isEmpty();
        
        boolean hyper = this.hyper || !left.isBinary() || !right.isBinary()
                || edgesMapped;
        
        double halfWidth = bounds.getWidth() / 2;
        Rectangle2D leftBnds = new Rectangle2D.Double(
                bounds.getX(), bounds.getY(), halfWidth, bounds.getHeight());
        Rectangle2D rightBnds = new Rectangle2D.Double(
                bounds.getX() + halfWidth, bounds.getY(), halfWidth, bounds.getHeight());
        
        Map<Node,VxNode> leftNodeMap = new HashMap<>();
        Map<Node,VxNode> rightNodeMap = new HashMap<>();
        Map<Edge,VxComponent> leftEdgeMap = null;
        Map<Edge,VxComponent> rightEdgeMap = null;
        if (edgesMapped) {
            leftEdgeMap = new HashMap<>();
            rightEdgeMap = new HashMap<>();
        }
        
        VxGraph result = new VxGraph();
        VxGroup leftGroup = result.createGroup();
        VxGroup rightGroup = result.createGroup();
        
        visualizeGraph(left, result, hyper, leftNodeMap, leftEdgeMap, leftBnds, leftGroup);
        visualizeGraph(right, result, hyper, rightNodeMap, rightEdgeMap, rightBnds, rightGroup);
        
        for (Map.Entry<Node,Node> entry : morph.getNodeMap().entrySet()) {
            VxNode src = leftNodeMap.get(entry.getKey());
            VxNode tgt = rightNodeMap.get(entry.getValue());
            result.addEdge(src, tgt, morphismStyleName);
        }
        
        if (edgesMapped) {
            for (Map.Entry<Edge,Edge> entry : morph.getEdgeMap().entrySet()) {
                VxNode src = (VxNode)leftEdgeMap.get(entry.getKey());
                VxNode tgt = (VxNode)rightEdgeMap.get(entry.getValue());
                result.addEdge(src, tgt, morphismStyleName);
            }
        }
        
        return new RuleVisi(result, leftGroup, rightGroup);
    }

    
    /**
     * Creates a VxGraph which visualizes a rule, where the rule is displayed
     * as a single graph with different style for deleted and created parts.
     * For this visualization to work, the rule must have an injective
     * correspondence morphism.
     * <p>
     * If either {@code getHyperStyle() == true}, or this rule contains
     * non-binary edges, the rule is visualized using a hypergraph 
     * representation; otherwise an ordinary graph representation is used.
     * <p>
     * The nodes and hyperedges of the graphical representation are put on
     * random positions, whereas all edges will be straight lines from the 
     * source to the target node. In most cases, it will be necessary to layout
     * the resulting graphical representation, either manually or with one
     * of the graph layout algorithms provided with the VisiGraph library.
     * <p>
     * The result visualization uses the styles associated with the following
     * style identifiers:
     * <ul>
     * <li><tt>createNodeStyleName</tt>:
     *     for nodes which are in the right-hand side of the rule, but have no
     *     corresponding node in the left-hand side, that is, nodes 
     *     <em>created</em> by this rule
     * <li><tt>deletedNodeStyleName</tt>:
     *     for nodes which are in the left-hand side of the rule, but have no
     *     corresponding node in the right-hand side, that is, nodes 
     *     <em>deleted</em> by this rule
     * <li><tt>nodeStyleName</tt>:
     *     for nodes which are neither created nor deleted by the rule
     * <li><tt>createdEdgeStyleName</tt>:
     *     for edges which are created by this rule (only in ordinary
     *     graph representation)
     * <li><tt>deletedEdgeStyleName</tt>:
     *     for edge which are deleted by this rule (only in ordinary 
     *     graph representation)
     * <li><tt>edgeStyleName</tt>:
     *     for edges which are neither deleted nor created by this rule
     * <li><tt>createdHyperedgeStyleName</tt>:
     *     for hyperedges which are created by this rule (only in hypergraph
     *     representation)
     * <li><tt>deletedHyperedgeStyleName</tt>:
     *     for hyperedges which are deleted by this rule (only in hypergraph
     *     representation)
     * <li><tt>hyperedgeStyle</tt>:
     *     for hyperedges which are neither created nor deleted by this rule
     *     (only in hypergraph representation)
     * <li><tt>createdConnectionStyleName</tt>:
     *     for connections between hyperedge and node, for hyperedges which
     *     are created by the rule (only in hypergraph representation)
     * <li><tt>deletedConnectionStyleName</tt> (edge style):
     *     for connections between hyperedge and node, for hyperedges which
     *     are deleted by the rule (only in hypergraph representation)
     * <li><tt>connectionStyleName</tt> (edge style):
     *     for connections between hyperedge and node, for hyperedges which
     *     are neither deleted nor created by this rule (only in hypergraph
     *     representation)
     * <li><tt>createdLabelStyleName</tt>:
     *     for labels of edges and connections which are created by the rule
     * <li><tt>deletedLabelStyleName</tt>:
     *     for labels of edges and connections which are deleted by the rule
     * <li><tt>labelStyleName</tt>:
     *     for labels of edges and connections which are neither created nor
     *     deleted by the rule
     * </ul>  
     * 
     * @param rule
     *      the rule to be visualized
     * @return
     *      a visual representation of the rule
     * @throws NullPointerException
     *      if {@code rule} is {@code null}
     * @throws IllegalArgumentException
     *      if {@code rule} does not have an injective correspondence
     *      morphism
     */
    public VxGraph visualizeFlatMorphism(Morphism morph) 
    {
        Graph left = morph.getDomain();
        Graph right = morph.getCodomain();
        
        if (!morph.isInjective()) {
            throw new IllegalArgumentException("Morphism must be injective.");
        }
        
        boolean hyper = this.hyper || !left.isBinary() || !right.isBinary();

        Map<Node,VxNode> nodeMap = new HashMap<>();
        
        VxGraph result = new VxGraph();
        
        for (Node node : left.getNodes()) {
            Point2D pos = randomPos();
            VxNode vxn = result.addNode(
                    pos.getX(), 
                    pos.getY(), 
                    nodeLabels ? Integer.toString(node.getId()) : null,
                    nodeStyleName);
            if (!morph.getNodeMap().containsKey(node)) {
                vxn.setStyle(deletedNodeStyleName);
            }
            vxn.setReference(node);
            nodeMap.put(node, vxn);
        }
        
        for (Edge edge : left.getEdges()) {
            List<Node> nodes = edge.getNodes();
            
            String esn = edgeStyleName;
            String hsn = hyperedgeStyleName;
            String csn = connectionStyleName;
            String lsn = labelStyleName;
            if (!morph.getEdgeMap().containsKey(edge)) {
                esn = deletedEdgeStyleName;
                hsn = deletedHyperedgeStyleName;
                csn = deletedConnectionStyleName;
                lsn = deletedLabelStyleName;
            }

            if (hyper) {
                VxNode[] vxs = new VxNode[nodes.size()];
                for (int i = 0; i < nodes.size(); i++) {
                    vxs[i] = nodeMap.get(nodes.get(i));
                }
                VxNode hyperEdge = result.addNode(
                        getBarycenter(vxs),
                        edgeLabels ? edge.getLabel().getName() : null,
                        hsn);
                hyperEdge.setReference(edge);
                for (int i = 0; i < vxs.length; i++) {
                    VxEdge newEdge = result.addEdge(
                            hyperEdge,
                            vxs[i],
                            csn);
                    VxLabelNode label = newEdge.addLabel(Integer.toString(i));
                    label.setStyle(lsn);
                    label.setPositionOnEdge(.3);
                }
            }
            else {
                VxNode source = nodeMap.get(nodes.get(0));
                VxNode target = nodeMap.get(nodes.get(1));
                assert (source != null && target != null);
                VxEdge vxe = result.addEdge(source, target, esn);
                if (edgeLabels) {
                    VxLabelNode label = vxe.addLabel(edge.getLabel().getName());
                    label.setStyle(lsn);
                }
                vxe.setReference(edge);
            }
        }
        
        for (Node node : right.getNodes()) {
            if (!morph.inRange(node)) {
                Point2D pos = randomPos();
                VxNode vxn = result.addNode(
                        pos.getX(), 
                        pos.getY(), 
                        nodeLabels ? Integer.toString(node.getId()) : null,
                        createdNodeStyleName);
                vxn.setReference(node);
                nodeMap.put(node, vxn);
            }
        }
        
        Morphism corrInv = morph.getInverse();
        
        for (Edge edge : right.getEdges()) {
            if (!morph.inRange(edge)) {
                List<Node> nodes = edge.getNodes();
                if (hyper) {
                    VxNode[] vxs = new VxNode[nodes.size()];
                    for (int i = 0; i < nodes.size(); i++) {
                        Node n = nodes.get(i);
                        if (morph.inRange(n)) {
                            n = corrInv.get(n);
                        }
                        assert n != null;
                        vxs[i] = nodeMap.get(n);
                    }
                    VxNode hyperEdge = result.addNode(
                            edgeLabels ? edge.getLabel().getName() : null,
                            createdHyperedgeStyleName);
                    hyperEdge.setReference(edge);
                    for (int i = 0; i < vxs.length; i++) {
                        VxEdge newEdge = result.addEdge(
                                hyperEdge,
                                vxs[i],
                                createdConnectionStyleName);
                        VxLabelNode label = newEdge.addLabel(Integer.toString(i));
                        label.setStyle(createdLabelStyleName);
                        label.setPositionOnEdge(.3);
                    }
                }
                else {
                    Node src = edge.getNodes().get(0);
                    if (morph.inRange(src)) {
                        src = corrInv.get(src);
                    }
                    Node tgt = edge.getNodes().get(1);
                    if (morph.inRange(tgt)) {
                        tgt = corrInv.get(tgt);
                    }
                    VxNode vxSrc = nodeMap.get(src);
                    VxNode vxTgt = nodeMap.get(tgt);
                    assert (vxSrc != null && vxTgt != null);
                    VxEdge vxe = result.addEdge(
                            vxSrc, 
                            vxTgt, 
                            createdEdgeStyleName);
                    if (edgeLabels) {
                        VxLabelNode label = vxe.addLabel(edge.getLabel().getName());
                        label.setStyle(createdLabelStyleName);
                    }
                    vxe.setReference(edge);
                }
            }
        }
        
        return result;
    }
    
    
    /**
     * Visualizes a transition system. This method generates a VxGraph
     * representing the labeled transition system of the given system. It does
     * not visualize the objects of the transition system.
     * <p>
     * This method ignores the property set by {@link #setHyperStyle(boolean)}
     * and will always produce binary, directed graphs.
     * <p>
     * The result visualization uses the styles associated with the following
     * style identifiers:
     * <ul>
     * <li><tt>stateStyleName</tt> (node style):
     *     The style used for the objects in the transition system.
     * <li><tt>transitionStyleName</tt> (edge style):
     *     The style used for the transitions in the transition system.
     * </ul>
     * 
     * @param system
     *      system for which a visual representation is to be generated
     * @return
     *      VxGraph which represents the given transition system
     */
    public VxGraph visualizeSystem(ExplicitTransitionSystem system)
    {
        Map<Graph,VxNode> nodes = new HashMap<>();
        VxGraph visi = new VxGraph();
        
        for (Transition t : system.getTransitions()) {
            Graph src = t.getSource();
            VxNode srcNode = nodes.get(src);
            if (srcNode == null) {
                Point2D pos = randomPos();
                srcNode = visi.addNode(pos, src.getName(), stateStyleName);
                srcNode.setReference(src);
                nodes.put(src, srcNode);
            }
            Graph tgt = t.getTarget();
            VxNode tgtNode = nodes.get(tgt);
            if (tgtNode == null) {
                Point2D pos = randomPos();
                tgtNode = visi.addNode(pos, tgt.getName(), stateStyleName);
                tgtNode.setReference(tgt);
                nodes.put(tgt, tgtNode);
            }
            VxEdge te = visi.addEdge(srcNode, tgtNode, transitionStyleName);
            te.addLabel(t.getLabel());
            te.setReference(t);
        }
        
        return visi;
    }
    
    
    /**
     * Returns a GraphEditSynchronizer for Graph instances which uses the 
     * style properties of this visualizer.
     */
    public VisualSynchronizer getSynchronizer()
    {
        return new VisualSynchronizer(); 
    }
    
    
    /**
     * GraphEditSynchronizer implementation which uses the style properties of
     * this GraphVisualizer. 
     * The property <tt>addMode</tt> (manipulated with the methods 
     * {@link GraphVisualizer$VisualSynchronizer#setAddMode(AddMode) setAddMode} and
     * {@link GraphVisualizer$VisualSynchronizer#getAddMode() getAddMode})
     * controls which of the style names are used.
     */
    public class VisualSynchronizer extends AbstractGraphSynchronizer
    {
        private AddMode addMode = AddMode.NORMAL;
        
        public void setAddMode(AddMode addMode)
        {
            this.addMode = addMode;
        }
        
        public AddMode getAddMode()
        {
            return addMode;
        }

        @Override
        protected boolean getBinaryArrows()
        {
            return !hyper;
        }

        @Override
        protected String getNodeStyleName()
        {
            switch (addMode) {
                case NORMAL:
                    return nodeStyleName;
                case ADD_CREATED:
                    return createdNodeStyleName;
                case ADD_DELETED:
                    return deletedNodeStyleName;
            }
            assert false; // not reachable
            return null;
        }

        @Override
        protected String getEdgeStyleName()
        {
            switch (addMode) {
                case NORMAL:
                    return edgeStyleName;
                case ADD_CREATED:
                    return createdEdgeStyleName;
                case ADD_DELETED:
                    return deletedEdgeStyleName;
            }
            assert false; // not reachable
            return null;
        }

        @Override
        protected String getHyperedgeStyleName()
        {
            switch (addMode) {
                case NORMAL:
                    return hyperedgeStyleName;
                case ADD_CREATED:
                    return createdHyperedgeStyleName;
                case ADD_DELETED:
                    return deletedHyperedgeStyleName;
            }
            assert false; // not reachable
            return null;
        }

        @Override
        protected String getConnectionStyleName()
        {
            switch (addMode) {
                case NORMAL:
                    return connectionStyleName;
                case ADD_CREATED:
                    return createdConnectionStyleName;
                case ADD_DELETED:
                    return deletedConnectionStyleName;
            }
            assert false; // not reachable
            return null;
        }

        @Override
        protected String getLabelStyleName()
        {
            switch (addMode) {
                case NORMAL:
                    return labelStyleName;
                case ADD_CREATED:
                    return createdLabelStyleName;
                case ADD_DELETED:
                    return deletedLabelStyleName;
            }
            assert false; // not reachable
            return null;
        }
    }
    
        
    // ---------------------------------------------------------------------- //
    //                       Private utility methods                          //
    // ---------------------------------------------------------------------- //
    
    /**
     * Returns a random position within this GraphVisualizer's bounds.
     */
    private Point2D randomPos()
    {
        return randomPos(bounds);
    }

    
    /**
     * Returns a random position within user-specified bounds.
     */
    private static Point2D randomPos(Rectangle2D bounds)
    {
        return new Point2D.Double(
                bounds.getX() + Math.random() * bounds.getWidth(),
                bounds.getY() + Math.random() * bounds.getHeight());
    }
    
    
    /**
     * Returns the barycenter of an array of nodes, or a random position if
     * one of the nodes has no explicit position.
     */
    private Point2D getBarycenter(VxNode[] nodes) 
    {
        double x = 0;
        double y = 0;
        int c = 0;
        
        for (VxNode node : nodes) {
            if (node instanceof VxExplicitNode) {
                VxExplicitNode enode = (VxExplicitNode)node;
                Point2D pos = enode.position();
                x += pos.getX();
                y += pos.getY();
                c++;
            }
            else {
                return randomPos();
            }
        }
        
        return new Point2D.Double(x / (double)c, y / (double)c);
    }
}
